# This is a makemhr HRIR definition file.  It is used to define the layout and
# source data to be processed into an OpenAL Soft compatible HRTF.
#
# This definition is used to transform the left and right ear HRIRs from a
# data set used in several papers and articles by Fumitada Itakura, Kazuya
# Takeda, Mikio Ikeda, Shoji Kajita, and Takanori Nishino.
#
# The data (data02.tgz) can be obtained from The Database of Head Related
# Transfer Functions hosted by the Takeda Laboratory at Nagoya University:
#
#  http://www.sp.m.is.nagoya-u.ac.jp/HRTF/database.html
#
# It is copyright 1999 by Itakura Laboratory and the Center for Integrated
# Acoustic Information Research (CIAIR) of Nagoya University and provided
# free of charge with no restrictions on use so long as the authors (above)
# are cited.

rate     = 44100

# The CIAIR set is stereo because it provides both ear HRIRs.
type     = stereo

points   = 512

# No head radius was provided.  Just use the average radius of 9 cm.
radius   = 0.09

# The CIAIR set is composed of a single field with an unknown distance
# between the source and the listener, so a guess of 1.5 meters is used.
distance = 1.5

# This set has a uniform number of azimuths for all but the poles (-90 and 90
# degree elevation).
azimuths = 1, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 1

# The CIAIR source azimuth is counter-clockwise, so it needs to be flipped.
# The extension of the source data may be misleading, they're ASCII text
# lists of floating point values (one per line).  Left and right ear HRIRs
# (from the respective files) are used to create a stereo HRTF.
[  9,  0 ] = ascii (fp) : "./hrtfs/elev-45/L-45e000a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e000a.dat" right
[  9,  1 ] = ascii (fp) : "./hrtfs/elev-45/L-45e355a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e355a.dat" right
[  9,  2 ] = ascii (fp) : "./hrtfs/elev-45/L-45e350a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e350a.dat" right
[  9,  3 ] = ascii (fp) : "./hrtfs/elev-45/L-45e345a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e345a.dat" right
[  9,  4 ] = ascii (fp) : "./hrtfs/elev-45/L-45e340a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e340a.dat" right
[  9,  5 ] = ascii (fp) : "./hrtfs/elev-45/L-45e335a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e335a.dat" right
[  9,  6 ] = ascii (fp) : "./hrtfs/elev-45/L-45e330a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e330a.dat" right
[  9,  7 ] = ascii (fp) : "./hrtfs/elev-45/L-45e325a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e325a.dat" right
[  9,  8 ] = ascii (fp) : "./hrtfs/elev-45/L-45e320a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e320a.dat" right
[  9,  9 ] = ascii (fp) : "./hrtfs/elev-45/L-45e315a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e315a.dat" right
[  9, 10 ] = ascii (fp) : "./hrtfs/elev-45/L-45e310a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e310a.dat" right
[  9, 11 ] = ascii (fp) : "./hrtfs/elev-45/L-45e305a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e305a.dat" right
[  9, 12 ] = ascii (fp) : "./hrtfs/elev-45/L-45e300a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e300a.dat" right
[  9, 13 ] = ascii (fp) : "./hrtfs/elev-45/L-45e295a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e295a.dat" right
[  9, 14 ] = ascii (fp) : "./hrtfs/elev-45/L-45e290a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e290a.dat" right
[  9, 15 ] = ascii (fp) : "./hrtfs/elev-45/L-45e285a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e285a.dat" right
[  9, 16 ] = ascii (fp) : "./hrtfs/elev-45/L-45e280a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e280a.dat" right
[  9, 17 ] = ascii (fp) : "./hrtfs/elev-45/L-45e275a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e275a.dat" right
[  9, 18 ] = ascii (fp) : "./hrtfs/elev-45/L-45e270a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e270a.dat" right
[  9, 19 ] = ascii (fp) : "./hrtfs/elev-45/L-45e265a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e265a.dat" right
[  9, 20 ] = ascii (fp) : "./hrtfs/elev-45/L-45e260a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e260a.dat" right
[  9, 21 ] = ascii (fp) : "./hrtfs/elev-45/L-45e255a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e255a.dat" right
[  9, 22 ] = ascii (fp) : "./hrtfs/elev-45/L-45e250a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e250a.dat" right
[  9, 23 ] = ascii (fp) : "./hrtfs/elev-45/L-45e245a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e245a.dat" right
[  9, 24 ] = ascii (fp) : "./hrtfs/elev-45/L-45e240a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e240a.dat" right
[  9, 25 ] = ascii (fp) : "./hrtfs/elev-45/L-45e235a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e235a.dat" right
[  9, 26 ] = ascii (fp) : "./hrtfs/elev-45/L-45e230a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e230a.dat" right
[  9, 27 ] = ascii (fp) : "./hrtfs/elev-45/L-45e225a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e225a.dat" right
[  9, 28 ] = ascii (fp) : "./hrtfs/elev-45/L-45e220a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e220a.dat" right
[  9, 29 ] = ascii (fp) : "./hrtfs/elev-45/L-45e215a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e215a.dat" right
[  9, 30 ] = ascii (fp) : "./hrtfs/elev-45/L-45e210a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e210a.dat" right
[  9, 31 ] = ascii (fp) : "./hrtfs/elev-45/L-45e205a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e205a.dat" right
[  9, 32 ] = ascii (fp) : "./hrtfs/elev-45/L-45e200a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e200a.dat" right
[  9, 33 ] = ascii (fp) : "./hrtfs/elev-45/L-45e195a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e195a.dat" right
[  9, 34 ] = ascii (fp) : "./hrtfs/elev-45/L-45e190a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e190a.dat" right
[  9, 35 ] = ascii (fp) : "./hrtfs/elev-45/L-45e185a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e185a.dat" right
[  9, 36 ] = ascii (fp) : "./hrtfs/elev-45/L-45e180a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e180a.dat" right
[  9, 37 ] = ascii (fp) : "./hrtfs/elev-45/L-45e175a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e175a.dat" right
[  9, 38 ] = ascii (fp) : "./hrtfs/elev-45/L-45e170a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e170a.dat" right
[  9, 39 ] = ascii (fp) : "./hrtfs/elev-45/L-45e165a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e165a.dat" right
[  9, 40 ] = ascii (fp) : "./hrtfs/elev-45/L-45e160a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e160a.dat" right
[  9, 41 ] = ascii (fp) : "./hrtfs/elev-45/L-45e155a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e155a.dat" right
[  9, 42 ] = ascii (fp) : "./hrtfs/elev-45/L-45e150a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e150a.dat" right
[  9, 43 ] = ascii (fp) : "./hrtfs/elev-45/L-45e145a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e145a.dat" right
[  9, 44 ] = ascii (fp) : "./hrtfs/elev-45/L-45e140a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e140a.dat" right
[  9, 45 ] = ascii (fp) : "./hrtfs/elev-45/L-45e135a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e135a.dat" right
[  9, 46 ] = ascii (fp) : "./hrtfs/elev-45/L-45e130a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e130a.dat" right
[  9, 47 ] = ascii (fp) : "./hrtfs/elev-45/L-45e125a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e125a.dat" right
[  9, 48 ] = ascii (fp) : "./hrtfs/elev-45/L-45e120a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e120a.dat" right
[  9, 49 ] = ascii (fp) : "./hrtfs/elev-45/L-45e115a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e115a.dat" right
[  9, 50 ] = ascii (fp) : "./hrtfs/elev-45/L-45e110a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e110a.dat" right
[  9, 51 ] = ascii (fp) : "./hrtfs/elev-45/L-45e105a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e105a.dat" right
[  9, 52 ] = ascii (fp) : "./hrtfs/elev-45/L-45e100a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e100a.dat" right
[  9, 53 ] = ascii (fp) : "./hrtfs/elev-45/L-45e095a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e095a.dat" right
[  9, 54 ] = ascii (fp) : "./hrtfs/elev-45/L-45e090a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e090a.dat" right
[  9, 55 ] = ascii (fp) : "./hrtfs/elev-45/L-45e085a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e085a.dat" right
[  9, 56 ] = ascii (fp) : "./hrtfs/elev-45/L-45e080a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e080a.dat" right
[  9, 57 ] = ascii (fp) : "./hrtfs/elev-45/L-45e075a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e075a.dat" right
[  9, 58 ] = ascii (fp) : "./hrtfs/elev-45/L-45e070a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e070a.dat" right
[  9, 59 ] = ascii (fp) : "./hrtfs/elev-45/L-45e065a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e065a.dat" right
[  9, 60 ] = ascii (fp) : "./hrtfs/elev-45/L-45e060a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e060a.dat" right
[  9, 61 ] = ascii (fp) : "./hrtfs/elev-45/L-45e055a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e055a.dat" right
[  9, 62 ] = ascii (fp) : "./hrtfs/elev-45/L-45e050a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e050a.dat" right
[  9, 63 ] = ascii (fp) : "./hrtfs/elev-45/L-45e045a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e045a.dat" right
[  9, 64 ] = ascii (fp) : "./hrtfs/elev-45/L-45e040a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e040a.dat" right
[  9, 65 ] = ascii (fp) : "./hrtfs/elev-45/L-45e035a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e035a.dat" right
[  9, 66 ] = ascii (fp) : "./hrtfs/elev-45/L-45e030a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e030a.dat" right
[  9, 67 ] = ascii (fp) : "./hrtfs/elev-45/L-45e025a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e025a.dat" right
[  9, 68 ] = ascii (fp) : "./hrtfs/elev-45/L-45e020a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e020a.dat" right
[  9, 69 ] = ascii (fp) : "./hrtfs/elev-45/L-45e015a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e015a.dat" right
[  9, 70 ] = ascii (fp) : "./hrtfs/elev-45/L-45e010a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e010a.dat" right
[  9, 71 ] = ascii (fp) : "./hrtfs/elev-45/L-45e005a.dat" left
           + ascii (fp) : "./hrtfs/elev-45/R-45e005a.dat" right

[ 10,  0 ] = ascii (fp) : "./hrtfs/elev-40/L-40e000a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e000a.dat" right
[ 10,  1 ] = ascii (fp) : "./hrtfs/elev-40/L-40e355a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e355a.dat" right
[ 10,  2 ] = ascii (fp) : "./hrtfs/elev-40/L-40e350a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e350a.dat" right
[ 10,  3 ] = ascii (fp) : "./hrtfs/elev-40/L-40e345a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e345a.dat" right
[ 10,  4 ] = ascii (fp) : "./hrtfs/elev-40/L-40e340a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e340a.dat" right
[ 10,  5 ] = ascii (fp) : "./hrtfs/elev-40/L-40e335a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e335a.dat" right
[ 10,  6 ] = ascii (fp) : "./hrtfs/elev-40/L-40e330a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e330a.dat" right
[ 10,  7 ] = ascii (fp) : "./hrtfs/elev-40/L-40e325a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e325a.dat" right
[ 10,  8 ] = ascii (fp) : "./hrtfs/elev-40/L-40e320a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e320a.dat" right
[ 10,  9 ] = ascii (fp) : "./hrtfs/elev-40/L-40e315a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e315a.dat" right
[ 10, 10 ] = ascii (fp) : "./hrtfs/elev-40/L-40e310a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e310a.dat" right
[ 10, 11 ] = ascii (fp) : "./hrtfs/elev-40/L-40e305a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e305a.dat" right
[ 10, 12 ] = ascii (fp) : "./hrtfs/elev-40/L-40e300a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e300a.dat" right
[ 10, 13 ] = ascii (fp) : "./hrtfs/elev-40/L-40e295a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e295a.dat" right
[ 10, 14 ] = ascii (fp) : "./hrtfs/elev-40/L-40e290a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e290a.dat" right
[ 10, 15 ] = ascii (fp) : "./hrtfs/elev-40/L-40e285a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e285a.dat" right
[ 10, 16 ] = ascii (fp) : "./hrtfs/elev-40/L-40e280a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e280a.dat" right
[ 10, 17 ] = ascii (fp) : "./hrtfs/elev-40/L-40e275a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e275a.dat" right
[ 10, 18 ] = ascii (fp) : "./hrtfs/elev-40/L-40e270a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e270a.dat" right
[ 10, 19 ] = ascii (fp) : "./hrtfs/elev-40/L-40e265a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e265a.dat" right
[ 10, 20 ] = ascii (fp) : "./hrtfs/elev-40/L-40e260a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e260a.dat" right
[ 10, 21 ] = ascii (fp) : "./hrtfs/elev-40/L-40e255a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e255a.dat" right
[ 10, 22 ] = ascii (fp) : "./hrtfs/elev-40/L-40e250a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e250a.dat" right
[ 10, 23 ] = ascii (fp) : "./hrtfs/elev-40/L-40e245a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e245a.dat" right
[ 10, 24 ] = ascii (fp) : "./hrtfs/elev-40/L-40e240a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e240a.dat" right
[ 10, 25 ] = ascii (fp) : "./hrtfs/elev-40/L-40e235a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e235a.dat" right
[ 10, 26 ] = ascii (fp) : "./hrtfs/elev-40/L-40e230a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e230a.dat" right
[ 10, 27 ] = ascii (fp) : "./hrtfs/elev-40/L-40e225a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e225a.dat" right
[ 10, 28 ] = ascii (fp) : "./hrtfs/elev-40/L-40e220a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e220a.dat" right
[ 10, 29 ] = ascii (fp) : "./hrtfs/elev-40/L-40e215a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e215a.dat" right
[ 10, 30 ] = ascii (fp) : "./hrtfs/elev-40/L-40e210a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e210a.dat" right
[ 10, 31 ] = ascii (fp) : "./hrtfs/elev-40/L-40e205a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e205a.dat" right
[ 10, 32 ] = ascii (fp) : "./hrtfs/elev-40/L-40e200a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e200a.dat" right
[ 10, 33 ] = ascii (fp) : "./hrtfs/elev-40/L-40e195a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e195a.dat" right
[ 10, 34 ] = ascii (fp) : "./hrtfs/elev-40/L-40e190a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e190a.dat" right
[ 10, 35 ] = ascii (fp) : "./hrtfs/elev-40/L-40e185a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e185a.dat" right
[ 10, 36 ] = ascii (fp) : "./hrtfs/elev-40/L-40e180a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e180a.dat" right
[ 10, 37 ] = ascii (fp) : "./hrtfs/elev-40/L-40e175a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e175a.dat" right
[ 10, 38 ] = ascii (fp) : "./hrtfs/elev-40/L-40e170a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e170a.dat" right
[ 10, 39 ] = ascii (fp) : "./hrtfs/elev-40/L-40e165a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e165a.dat" right
[ 10, 40 ] = ascii (fp) : "./hrtfs/elev-40/L-40e160a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e160a.dat" right
[ 10, 41 ] = ascii (fp) : "./hrtfs/elev-40/L-40e155a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e155a.dat" right
[ 10, 42 ] = ascii (fp) : "./hrtfs/elev-40/L-40e150a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e150a.dat" right
[ 10, 43 ] = ascii (fp) : "./hrtfs/elev-40/L-40e145a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e145a.dat" right
[ 10, 44 ] = ascii (fp) : "./hrtfs/elev-40/L-40e140a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e140a.dat" right
[ 10, 45 ] = ascii (fp) : "./hrtfs/elev-40/L-40e135a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e135a.dat" right
[ 10, 46 ] = ascii (fp) : "./hrtfs/elev-40/L-40e130a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e130a.dat" right
[ 10, 47 ] = ascii (fp) : "./hrtfs/elev-40/L-40e125a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e125a.dat" right
[ 10, 48 ] = ascii (fp) : "./hrtfs/elev-40/L-40e120a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e120a.dat" right
[ 10, 49 ] = ascii (fp) : "./hrtfs/elev-40/L-40e115a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e115a.dat" right
[ 10, 50 ] = ascii (fp) : "./hrtfs/elev-40/L-40e110a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e110a.dat" right
[ 10, 51 ] = ascii (fp) : "./hrtfs/elev-40/L-40e105a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e105a.dat" right
[ 10, 52 ] = ascii (fp) : "./hrtfs/elev-40/L-40e100a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e100a.dat" right
[ 10, 53 ] = ascii (fp) : "./hrtfs/elev-40/L-40e095a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e095a.dat" right
[ 10, 54 ] = ascii (fp) : "./hrtfs/elev-40/L-40e090a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e090a.dat" right
[ 10, 55 ] = ascii (fp) : "./hrtfs/elev-40/L-40e085a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e085a.dat" right
[ 10, 56 ] = ascii (fp) : "./hrtfs/elev-40/L-40e080a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e080a.dat" right
[ 10, 57 ] = ascii (fp) : "./hrtfs/elev-40/L-40e075a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e075a.dat" right
[ 10, 58 ] = ascii (fp) : "./hrtfs/elev-40/L-40e070a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e070a.dat" right
[ 10, 59 ] = ascii (fp) : "./hrtfs/elev-40/L-40e065a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e065a.dat" right
[ 10, 60 ] = ascii (fp) : "./hrtfs/elev-40/L-40e060a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e060a.dat" right
[ 10, 61 ] = ascii (fp) : "./hrtfs/elev-40/L-40e055a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e055a.dat" right
[ 10, 62 ] = ascii (fp) : "./hrtfs/elev-40/L-40e050a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e050a.dat" right
[ 10, 63 ] = ascii (fp) : "./hrtfs/elev-40/L-40e045a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e045a.dat" right
[ 10, 64 ] = ascii (fp) : "./hrtfs/elev-40/L-40e040a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e040a.dat" right
[ 10, 65 ] = ascii (fp) : "./hrtfs/elev-40/L-40e035a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e035a.dat" right
[ 10, 66 ] = ascii (fp) : "./hrtfs/elev-40/L-40e030a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e030a.dat" right
[ 10, 67 ] = ascii (fp) : "./hrtfs/elev-40/L-40e025a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e025a.dat" right
[ 10, 68 ] = ascii (fp) : "./hrtfs/elev-40/L-40e020a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e020a.dat" right
[ 10, 69 ] = ascii (fp) : "./hrtfs/elev-40/L-40e015a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e015a.dat" right
[ 10, 70 ] = ascii (fp) : "./hrtfs/elev-40/L-40e010a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e010a.dat" right
[ 10, 71 ] = ascii (fp) : "./hrtfs/elev-40/L-40e005a.dat" left
           + ascii (fp) : "./hrtfs/elev-40/R-40e005a.dat" right

[ 11,  0 ] = ascii (fp) : "./hrtfs/elev-35/L-35e000a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e000a.dat" right
[ 11,  1 ] = ascii (fp) : "./hrtfs/elev-35/L-35e355a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e355a.dat" right
[ 11,  2 ] = ascii (fp) : "./hrtfs/elev-35/L-35e350a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e350a.dat" right
[ 11,  3 ] = ascii (fp) : "./hrtfs/elev-35/L-35e345a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e345a.dat" right
[ 11,  4 ] = ascii (fp) : "./hrtfs/elev-35/L-35e340a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e340a.dat" right
[ 11,  5 ] = ascii (fp) : "./hrtfs/elev-35/L-35e335a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e335a.dat" right
[ 11,  6 ] = ascii (fp) : "./hrtfs/elev-35/L-35e330a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e330a.dat" right
[ 11,  7 ] = ascii (fp) : "./hrtfs/elev-35/L-35e325a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e325a.dat" right
[ 11,  8 ] = ascii (fp) : "./hrtfs/elev-35/L-35e320a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e320a.dat" right
[ 11,  9 ] = ascii (fp) : "./hrtfs/elev-35/L-35e315a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e315a.dat" right
[ 11, 10 ] = ascii (fp) : "./hrtfs/elev-35/L-35e310a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e310a.dat" right
[ 11, 11 ] = ascii (fp) : "./hrtfs/elev-35/L-35e305a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e305a.dat" right
[ 11, 12 ] = ascii (fp) : "./hrtfs/elev-35/L-35e300a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e300a.dat" right
[ 11, 13 ] = ascii (fp) : "./hrtfs/elev-35/L-35e295a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e295a.dat" right
[ 11, 14 ] = ascii (fp) : "./hrtfs/elev-35/L-35e290a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e290a.dat" right
[ 11, 15 ] = ascii (fp) : "./hrtfs/elev-35/L-35e285a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e285a.dat" right
[ 11, 16 ] = ascii (fp) : "./hrtfs/elev-35/L-35e280a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e280a.dat" right
[ 11, 17 ] = ascii (fp) : "./hrtfs/elev-35/L-35e275a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e275a.dat" right
[ 11, 18 ] = ascii (fp) : "./hrtfs/elev-35/L-35e270a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e270a.dat" right
[ 11, 19 ] = ascii (fp) : "./hrtfs/elev-35/L-35e265a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e265a.dat" right
[ 11, 20 ] = ascii (fp) : "./hrtfs/elev-35/L-35e260a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e260a.dat" right
[ 11, 21 ] = ascii (fp) : "./hrtfs/elev-35/L-35e255a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e255a.dat" right
[ 11, 22 ] = ascii (fp) : "./hrtfs/elev-35/L-35e250a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e250a.dat" right
[ 11, 23 ] = ascii (fp) : "./hrtfs/elev-35/L-35e245a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e245a.dat" right
[ 11, 24 ] = ascii (fp) : "./hrtfs/elev-35/L-35e240a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e240a.dat" right
[ 11, 25 ] = ascii (fp) : "./hrtfs/elev-35/L-35e235a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e235a.dat" right
[ 11, 26 ] = ascii (fp) : "./hrtfs/elev-35/L-35e230a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e230a.dat" right
[ 11, 27 ] = ascii (fp) : "./hrtfs/elev-35/L-35e225a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e225a.dat" right
[ 11, 28 ] = ascii (fp) : "./hrtfs/elev-35/L-35e220a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e220a.dat" right
[ 11, 29 ] = ascii (fp) : "./hrtfs/elev-35/L-35e215a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e215a.dat" right
[ 11, 30 ] = ascii (fp) : "./hrtfs/elev-35/L-35e210a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e210a.dat" right
[ 11, 31 ] = ascii (fp) : "./hrtfs/elev-35/L-35e205a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e205a.dat" right
[ 11, 32 ] = ascii (fp) : "./hrtfs/elev-35/L-35e200a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e200a.dat" right
[ 11, 33 ] = ascii (fp) : "./hrtfs/elev-35/L-35e195a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e195a.dat" right
[ 11, 34 ] = ascii (fp) : "./hrtfs/elev-35/L-35e190a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e190a.dat" right
[ 11, 35 ] = ascii (fp) : "./hrtfs/elev-35/L-35e185a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e185a.dat" right
[ 11, 36 ] = ascii (fp) : "./hrtfs/elev-35/L-35e180a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e180a.dat" right
[ 11, 37 ] = ascii (fp) : "./hrtfs/elev-35/L-35e175a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e175a.dat" right
[ 11, 38 ] = ascii (fp) : "./hrtfs/elev-35/L-35e170a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e170a.dat" right
[ 11, 39 ] = ascii (fp) : "./hrtfs/elev-35/L-35e165a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e165a.dat" right
[ 11, 40 ] = ascii (fp) : "./hrtfs/elev-35/L-35e160a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e160a.dat" right
[ 11, 41 ] = ascii (fp) : "./hrtfs/elev-35/L-35e155a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e155a.dat" right
[ 11, 42 ] = ascii (fp) : "./hrtfs/elev-35/L-35e150a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e150a.dat" right
[ 11, 43 ] = ascii (fp) : "./hrtfs/elev-35/L-35e145a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e145a.dat" right
[ 11, 44 ] = ascii (fp) : "./hrtfs/elev-35/L-35e140a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e140a.dat" right
[ 11, 45 ] = ascii (fp) : "./hrtfs/elev-35/L-35e135a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e135a.dat" right
[ 11, 46 ] = ascii (fp) : "./hrtfs/elev-35/L-35e130a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e130a.dat" right
[ 11, 47 ] = ascii (fp) : "./hrtfs/elev-35/L-35e125a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e125a.dat" right
[ 11, 48 ] = ascii (fp) : "./hrtfs/elev-35/L-35e120a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e120a.dat" right
[ 11, 49 ] = ascii (fp) : "./hrtfs/elev-35/L-35e115a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e115a.dat" right
[ 11, 50 ] = ascii (fp) : "./hrtfs/elev-35/L-35e110a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e110a.dat" right
[ 11, 51 ] = ascii (fp) : "./hrtfs/elev-35/L-35e105a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e105a.dat" right
[ 11, 52 ] = ascii (fp) : "./hrtfs/elev-35/L-35e100a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e100a.dat" right
[ 11, 53 ] = ascii (fp) : "./hrtfs/elev-35/L-35e095a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e095a.dat" right
[ 11, 54 ] = ascii (fp) : "./hrtfs/elev-35/L-35e090a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e090a.dat" right
[ 11, 55 ] = ascii (fp) : "./hrtfs/elev-35/L-35e085a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e085a.dat" right
[ 11, 56 ] = ascii (fp) : "./hrtfs/elev-35/L-35e080a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e080a.dat" right
[ 11, 57 ] = ascii (fp) : "./hrtfs/elev-35/L-35e075a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e075a.dat" right
[ 11, 58 ] = ascii (fp) : "./hrtfs/elev-35/L-35e070a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e070a.dat" right
[ 11, 59 ] = ascii (fp) : "./hrtfs/elev-35/L-35e065a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e065a.dat" right
[ 11, 60 ] = ascii (fp) : "./hrtfs/elev-35/L-35e060a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e060a.dat" right
[ 11, 61 ] = ascii (fp) : "./hrtfs/elev-35/L-35e055a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e055a.dat" right
[ 11, 62 ] = ascii (fp) : "./hrtfs/elev-35/L-35e050a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e050a.dat" right
[ 11, 63 ] = ascii (fp) : "./hrtfs/elev-35/L-35e045a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e045a.dat" right
[ 11, 64 ] = ascii (fp) : "./hrtfs/elev-35/L-35e040a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e040a.dat" right
[ 11, 65 ] = ascii (fp) : "./hrtfs/elev-35/L-35e035a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e035a.dat" right
[ 11, 66 ] = ascii (fp) : "./hrtfs/elev-35/L-35e030a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e030a.dat" right
[ 11, 67 ] = ascii (fp) : "./hrtfs/elev-35/L-35e025a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e025a.dat" right
[ 11, 68 ] = ascii (fp) : "./hrtfs/elev-35/L-35e020a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e020a.dat" right
[ 11, 69 ] = ascii (fp) : "./hrtfs/elev-35/L-35e015a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e015a.dat" right
[ 11, 70 ] = ascii (fp) : "./hrtfs/elev-35/L-35e010a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e010a.dat" right
[ 11, 71 ] = ascii (fp) : "./hrtfs/elev-35/L-35e005a.dat" left
           + ascii (fp) : "./hrtfs/elev-35/R-35e005a.dat" right

[ 12,  0 ] = ascii (fp) : "./hrtfs/elev-30/L-30e000a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e000a.dat" right
[ 12,  1 ] = ascii (fp) : "./hrtfs/elev-30/L-30e355a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e355a.dat" right
[ 12,  2 ] = ascii (fp) : "./hrtfs/elev-30/L-30e350a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e350a.dat" right
[ 12,  3 ] = ascii (fp) : "./hrtfs/elev-30/L-30e345a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e345a.dat" right
[ 12,  4 ] = ascii (fp) : "./hrtfs/elev-30/L-30e340a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e340a.dat" right
[ 12,  5 ] = ascii (fp) : "./hrtfs/elev-30/L-30e335a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e335a.dat" right
[ 12,  6 ] = ascii (fp) : "./hrtfs/elev-30/L-30e330a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e330a.dat" right
[ 12,  7 ] = ascii (fp) : "./hrtfs/elev-30/L-30e325a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e325a.dat" right
[ 12,  8 ] = ascii (fp) : "./hrtfs/elev-30/L-30e320a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e320a.dat" right
[ 12,  9 ] = ascii (fp) : "./hrtfs/elev-30/L-30e315a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e315a.dat" right
[ 12, 10 ] = ascii (fp) : "./hrtfs/elev-30/L-30e310a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e310a.dat" right
[ 12, 11 ] = ascii (fp) : "./hrtfs/elev-30/L-30e305a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e305a.dat" right
[ 12, 12 ] = ascii (fp) : "./hrtfs/elev-30/L-30e300a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e300a.dat" right
[ 12, 13 ] = ascii (fp) : "./hrtfs/elev-30/L-30e295a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e295a.dat" right
[ 12, 14 ] = ascii (fp) : "./hrtfs/elev-30/L-30e290a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e290a.dat" right
[ 12, 15 ] = ascii (fp) : "./hrtfs/elev-30/L-30e285a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e285a.dat" right
[ 12, 16 ] = ascii (fp) : "./hrtfs/elev-30/L-30e280a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e280a.dat" right
[ 12, 17 ] = ascii (fp) : "./hrtfs/elev-30/L-30e275a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e275a.dat" right
[ 12, 18 ] = ascii (fp) : "./hrtfs/elev-30/L-30e270a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e270a.dat" right
[ 12, 19 ] = ascii (fp) : "./hrtfs/elev-30/L-30e265a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e265a.dat" right
[ 12, 20 ] = ascii (fp) : "./hrtfs/elev-30/L-30e260a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e260a.dat" right
[ 12, 21 ] = ascii (fp) : "./hrtfs/elev-30/L-30e255a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e255a.dat" right
[ 12, 22 ] = ascii (fp) : "./hrtfs/elev-30/L-30e250a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e250a.dat" right
[ 12, 23 ] = ascii (fp) : "./hrtfs/elev-30/L-30e245a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e245a.dat" right
[ 12, 24 ] = ascii (fp) : "./hrtfs/elev-30/L-30e240a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e240a.dat" right
[ 12, 25 ] = ascii (fp) : "./hrtfs/elev-30/L-30e235a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e235a.dat" right
[ 12, 26 ] = ascii (fp) : "./hrtfs/elev-30/L-30e230a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e230a.dat" right
[ 12, 27 ] = ascii (fp) : "./hrtfs/elev-30/L-30e225a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e225a.dat" right
[ 12, 28 ] = ascii (fp) : "./hrtfs/elev-30/L-30e220a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e220a.dat" right
[ 12, 29 ] = ascii (fp) : "./hrtfs/elev-30/L-30e215a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e215a.dat" right
[ 12, 30 ] = ascii (fp) : "./hrtfs/elev-30/L-30e210a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e210a.dat" right
[ 12, 31 ] = ascii (fp) : "./hrtfs/elev-30/L-30e205a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e205a.dat" right
[ 12, 32 ] = ascii (fp) : "./hrtfs/elev-30/L-30e200a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e200a.dat" right
[ 12, 33 ] = ascii (fp) : "./hrtfs/elev-30/L-30e195a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e195a.dat" right
[ 12, 34 ] = ascii (fp) : "./hrtfs/elev-30/L-30e190a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e190a.dat" right
[ 12, 35 ] = ascii (fp) : "./hrtfs/elev-30/L-30e185a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e185a.dat" right
[ 12, 36 ] = ascii (fp) : "./hrtfs/elev-30/L-30e180a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e180a.dat" right
[ 12, 37 ] = ascii (fp) : "./hrtfs/elev-30/L-30e175a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e175a.dat" right
[ 12, 38 ] = ascii (fp) : "./hrtfs/elev-30/L-30e170a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e170a.dat" right
[ 12, 39 ] = ascii (fp) : "./hrtfs/elev-30/L-30e165a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e165a.dat" right
[ 12, 40 ] = ascii (fp) : "./hrtfs/elev-30/L-30e160a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e160a.dat" right
[ 12, 41 ] = ascii (fp) : "./hrtfs/elev-30/L-30e155a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e155a.dat" right
[ 12, 42 ] = ascii (fp) : "./hrtfs/elev-30/L-30e150a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e150a.dat" right
[ 12, 43 ] = ascii (fp) : "./hrtfs/elev-30/L-30e145a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e145a.dat" right
[ 12, 44 ] = ascii (fp) : "./hrtfs/elev-30/L-30e140a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e140a.dat" right
[ 12, 45 ] = ascii (fp) : "./hrtfs/elev-30/L-30e135a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e135a.dat" right
[ 12, 46 ] = ascii (fp) : "./hrtfs/elev-30/L-30e130a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e130a.dat" right
[ 12, 47 ] = ascii (fp) : "./hrtfs/elev-30/L-30e125a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e125a.dat" right
[ 12, 48 ] = ascii (fp) : "./hrtfs/elev-30/L-30e120a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e120a.dat" right
[ 12, 49 ] = ascii (fp) : "./hrtfs/elev-30/L-30e115a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e115a.dat" right
[ 12, 50 ] = ascii (fp) : "./hrtfs/elev-30/L-30e110a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e110a.dat" right
[ 12, 51 ] = ascii (fp) : "./hrtfs/elev-30/L-30e105a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e105a.dat" right
[ 12, 52 ] = ascii (fp) : "./hrtfs/elev-30/L-30e100a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e100a.dat" right
[ 12, 53 ] = ascii (fp) : "./hrtfs/elev-30/L-30e095a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e095a.dat" right
[ 12, 54 ] = ascii (fp) : "./hrtfs/elev-30/L-30e090a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e090a.dat" right
[ 12, 55 ] = ascii (fp) : "./hrtfs/elev-30/L-30e085a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e085a.dat" right
[ 12, 56 ] = ascii (fp) : "./hrtfs/elev-30/L-30e080a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e080a.dat" right
[ 12, 57 ] = ascii (fp) : "./hrtfs/elev-30/L-30e075a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e075a.dat" right
[ 12, 58 ] = ascii (fp) : "./hrtfs/elev-30/L-30e070a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e070a.dat" right
[ 12, 59 ] = ascii (fp) : "./hrtfs/elev-30/L-30e065a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e065a.dat" right
[ 12, 60 ] = ascii (fp) : "./hrtfs/elev-30/L-30e060a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e060a.dat" right
[ 12, 61 ] = ascii (fp) : "./hrtfs/elev-30/L-30e055a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e055a.dat" right
[ 12, 62 ] = ascii (fp) : "./hrtfs/elev-30/L-30e050a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e050a.dat" right
[ 12, 63 ] = ascii (fp) : "./hrtfs/elev-30/L-30e045a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e045a.dat" right
[ 12, 64 ] = ascii (fp) : "./hrtfs/elev-30/L-30e040a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e040a.dat" right
[ 12, 65 ] = ascii (fp) : "./hrtfs/elev-30/L-30e035a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e035a.dat" right
[ 12, 66 ] = ascii (fp) : "./hrtfs/elev-30/L-30e030a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e030a.dat" right
[ 12, 67 ] = ascii (fp) : "./hrtfs/elev-30/L-30e025a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e025a.dat" right
[ 12, 68 ] = ascii (fp) : "./hrtfs/elev-30/L-30e020a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e020a.dat" right
[ 12, 69 ] = ascii (fp) : "./hrtfs/elev-30/L-30e015a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e015a.dat" right
[ 12, 70 ] = ascii (fp) : "./hrtfs/elev-30/L-30e010a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e010a.dat" right
[ 12, 71 ] = ascii (fp) : "./hrtfs/elev-30/L-30e005a.dat" left
           + ascii (fp) : "./hrtfs/elev-30/R-30e005a.dat" right

[ 13,  0 ] = ascii (fp) : "./hrtfs/elev-25/L-25e000a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e000a.dat" right
[ 13,  1 ] = ascii (fp) : "./hrtfs/elev-25/L-25e355a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e355a.dat" right
[ 13,  2 ] = ascii (fp) : "./hrtfs/elev-25/L-25e350a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e350a.dat" right
[ 13,  3 ] = ascii (fp) : "./hrtfs/elev-25/L-25e345a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e345a.dat" right
[ 13,  4 ] = ascii (fp) : "./hrtfs/elev-25/L-25e340a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e340a.dat" right
[ 13,  5 ] = ascii (fp) : "./hrtfs/elev-25/L-25e335a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e335a.dat" right
[ 13,  6 ] = ascii (fp) : "./hrtfs/elev-25/L-25e330a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e330a.dat" right
[ 13,  7 ] = ascii (fp) : "./hrtfs/elev-25/L-25e325a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e325a.dat" right
[ 13,  8 ] = ascii (fp) : "./hrtfs/elev-25/L-25e320a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e320a.dat" right
[ 13,  9 ] = ascii (fp) : "./hrtfs/elev-25/L-25e315a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e315a.dat" right
[ 13, 10 ] = ascii (fp) : "./hrtfs/elev-25/L-25e310a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e310a.dat" right
[ 13, 11 ] = ascii (fp) : "./hrtfs/elev-25/L-25e305a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e305a.dat" right
[ 13, 12 ] = ascii (fp) : "./hrtfs/elev-25/L-25e300a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e300a.dat" right
[ 13, 13 ] = ascii (fp) : "./hrtfs/elev-25/L-25e295a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e295a.dat" right
[ 13, 14 ] = ascii (fp) : "./hrtfs/elev-25/L-25e290a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e290a.dat" right
[ 13, 15 ] = ascii (fp) : "./hrtfs/elev-25/L-25e285a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e285a.dat" right
[ 13, 16 ] = ascii (fp) : "./hrtfs/elev-25/L-25e280a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e280a.dat" right
[ 13, 17 ] = ascii (fp) : "./hrtfs/elev-25/L-25e275a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e275a.dat" right
[ 13, 18 ] = ascii (fp) : "./hrtfs/elev-25/L-25e270a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e270a.dat" right
[ 13, 19 ] = ascii (fp) : "./hrtfs/elev-25/L-25e265a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e265a.dat" right
[ 13, 20 ] = ascii (fp) : "./hrtfs/elev-25/L-25e260a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e260a.dat" right
[ 13, 21 ] = ascii (fp) : "./hrtfs/elev-25/L-25e255a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e255a.dat" right
[ 13, 22 ] = ascii (fp) : "./hrtfs/elev-25/L-25e250a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e250a.dat" right
[ 13, 23 ] = ascii (fp) : "./hrtfs/elev-25/L-25e245a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e245a.dat" right
[ 13, 24 ] = ascii (fp) : "./hrtfs/elev-25/L-25e240a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e240a.dat" right
[ 13, 25 ] = ascii (fp) : "./hrtfs/elev-25/L-25e235a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e235a.dat" right
[ 13, 26 ] = ascii (fp) : "./hrtfs/elev-25/L-25e230a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e230a.dat" right
[ 13, 27 ] = ascii (fp) : "./hrtfs/elev-25/L-25e225a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e225a.dat" right
[ 13, 28 ] = ascii (fp) : "./hrtfs/elev-25/L-25e220a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e220a.dat" right
[ 13, 29 ] = ascii (fp) : "./hrtfs/elev-25/L-25e215a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e215a.dat" right
[ 13, 30 ] = ascii (fp) : "./hrtfs/elev-25/L-25e210a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e210a.dat" right
[ 13, 31 ] = ascii (fp) : "./hrtfs/elev-25/L-25e205a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e205a.dat" right
[ 13, 32 ] = ascii (fp) : "./hrtfs/elev-25/L-25e200a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e200a.dat" right
[ 13, 33 ] = ascii (fp) : "./hrtfs/elev-25/L-25e195a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e195a.dat" right
[ 13, 34 ] = ascii (fp) : "./hrtfs/elev-25/L-25e190a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e190a.dat" right
[ 13, 35 ] = ascii (fp) : "./hrtfs/elev-25/L-25e185a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e185a.dat" right
[ 13, 36 ] = ascii (fp) : "./hrtfs/elev-25/L-25e180a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e180a.dat" right
[ 13, 37 ] = ascii (fp) : "./hrtfs/elev-25/L-25e175a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e175a.dat" right
[ 13, 38 ] = ascii (fp) : "./hrtfs/elev-25/L-25e170a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e170a.dat" right
[ 13, 39 ] = ascii (fp) : "./hrtfs/elev-25/L-25e165a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e165a.dat" right
[ 13, 40 ] = ascii (fp) : "./hrtfs/elev-25/L-25e160a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e160a.dat" right
[ 13, 41 ] = ascii (fp) : "./hrtfs/elev-25/L-25e155a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e155a.dat" right
[ 13, 42 ] = ascii (fp) : "./hrtfs/elev-25/L-25e150a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e150a.dat" right
[ 13, 43 ] = ascii (fp) : "./hrtfs/elev-25/L-25e145a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e145a.dat" right
[ 13, 44 ] = ascii (fp) : "./hrtfs/elev-25/L-25e140a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e140a.dat" right
[ 13, 45 ] = ascii (fp) : "./hrtfs/elev-25/L-25e135a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e135a.dat" right
[ 13, 46 ] = ascii (fp) : "./hrtfs/elev-25/L-25e130a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e130a.dat" right
[ 13, 47 ] = ascii (fp) : "./hrtfs/elev-25/L-25e125a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e125a.dat" right
[ 13, 48 ] = ascii (fp) : "./hrtfs/elev-25/L-25e120a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e120a.dat" right
[ 13, 49 ] = ascii (fp) : "./hrtfs/elev-25/L-25e115a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e115a.dat" right
[ 13, 50 ] = ascii (fp) : "./hrtfs/elev-25/L-25e110a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e110a.dat" right
[ 13, 51 ] = ascii (fp) : "./hrtfs/elev-25/L-25e105a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e105a.dat" right
[ 13, 52 ] = ascii (fp) : "./hrtfs/elev-25/L-25e100a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e100a.dat" right
[ 13, 53 ] = ascii (fp) : "./hrtfs/elev-25/L-25e095a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e095a.dat" right
[ 13, 54 ] = ascii (fp) : "./hrtfs/elev-25/L-25e090a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e090a.dat" right
[ 13, 55 ] = ascii (fp) : "./hrtfs/elev-25/L-25e085a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e085a.dat" right
[ 13, 56 ] = ascii (fp) : "./hrtfs/elev-25/L-25e080a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e080a.dat" right
[ 13, 57 ] = ascii (fp) : "./hrtfs/elev-25/L-25e075a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e075a.dat" right
[ 13, 58 ] = ascii (fp) : "./hrtfs/elev-25/L-25e070a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e070a.dat" right
[ 13, 59 ] = ascii (fp) : "./hrtfs/elev-25/L-25e065a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e065a.dat" right
[ 13, 60 ] = ascii (fp) : "./hrtfs/elev-25/L-25e060a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e060a.dat" right
[ 13, 61 ] = ascii (fp) : "./hrtfs/elev-25/L-25e055a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e055a.dat" right
[ 13, 62 ] = ascii (fp) : "./hrtfs/elev-25/L-25e050a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e050a.dat" right
[ 13, 63 ] = ascii (fp) : "./hrtfs/elev-25/L-25e045a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e045a.dat" right
[ 13, 64 ] = ascii (fp) : "./hrtfs/elev-25/L-25e040a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e040a.dat" right
[ 13, 65 ] = ascii (fp) : "./hrtfs/elev-25/L-25e035a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e035a.dat" right
[ 13, 66 ] = ascii (fp) : "./hrtfs/elev-25/L-25e030a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e030a.dat" right
[ 13, 67 ] = ascii (fp) : "./hrtfs/elev-25/L-25e025a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e025a.dat" right
[ 13, 68 ] = ascii (fp) : "./hrtfs/elev-25/L-25e020a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e020a.dat" right
[ 13, 69 ] = ascii (fp) : "./hrtfs/elev-25/L-25e015a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e015a.dat" right
[ 13, 70 ] = ascii (fp) : "./hrtfs/elev-25/L-25e010a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e010a.dat" right
[ 13, 71 ] = ascii (fp) : "./hrtfs/elev-25/L-25e005a.dat" left
           + ascii (fp) : "./hrtfs/elev-25/R-25e005a.dat" right

[ 14,  0 ] = ascii (fp) : "./hrtfs/elev-20/L-20e000a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e000a.dat" right
[ 14,  1 ] = ascii (fp) : "./hrtfs/elev-20/L-20e355a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e355a.dat" right
[ 14,  2 ] = ascii (fp) : "./hrtfs/elev-20/L-20e350a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e350a.dat" right
[ 14,  3 ] = ascii (fp) : "./hrtfs/elev-20/L-20e345a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e345a.dat" right
[ 14,  4 ] = ascii (fp) : "./hrtfs/elev-20/L-20e340a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e340a.dat" right
[ 14,  5 ] = ascii (fp) : "./hrtfs/elev-20/L-20e335a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e335a.dat" right
[ 14,  6 ] = ascii (fp) : "./hrtfs/elev-20/L-20e330a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e330a.dat" right
[ 14,  7 ] = ascii (fp) : "./hrtfs/elev-20/L-20e325a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e325a.dat" right
[ 14,  8 ] = ascii (fp) : "./hrtfs/elev-20/L-20e320a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e320a.dat" right
[ 14,  9 ] = ascii (fp) : "./hrtfs/elev-20/L-20e315a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e315a.dat" right
[ 14, 10 ] = ascii (fp) : "./hrtfs/elev-20/L-20e310a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e310a.dat" right
[ 14, 11 ] = ascii (fp) : "./hrtfs/elev-20/L-20e305a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e305a.dat" right
[ 14, 12 ] = ascii (fp) : "./hrtfs/elev-20/L-20e300a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e300a.dat" right
[ 14, 13 ] = ascii (fp) : "./hrtfs/elev-20/L-20e295a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e295a.dat" right
[ 14, 14 ] = ascii (fp) : "./hrtfs/elev-20/L-20e290a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e290a.dat" right
[ 14, 15 ] = ascii (fp) : "./hrtfs/elev-20/L-20e285a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e285a.dat" right
[ 14, 16 ] = ascii (fp) : "./hrtfs/elev-20/L-20e280a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e280a.dat" right
[ 14, 17 ] = ascii (fp) : "./hrtfs/elev-20/L-20e275a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e275a.dat" right
[ 14, 18 ] = ascii (fp) : "./hrtfs/elev-20/L-20e270a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e270a.dat" right
[ 14, 19 ] = ascii (fp) : "./hrtfs/elev-20/L-20e265a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e265a.dat" right
[ 14, 20 ] = ascii (fp) : "./hrtfs/elev-20/L-20e260a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e260a.dat" right
[ 14, 21 ] = ascii (fp) : "./hrtfs/elev-20/L-20e255a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e255a.dat" right
[ 14, 22 ] = ascii (fp) : "./hrtfs/elev-20/L-20e250a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e250a.dat" right
[ 14, 23 ] = ascii (fp) : "./hrtfs/elev-20/L-20e245a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e245a.dat" right
[ 14, 24 ] = ascii (fp) : "./hrtfs/elev-20/L-20e240a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e240a.dat" right
[ 14, 25 ] = ascii (fp) : "./hrtfs/elev-20/L-20e235a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e235a.dat" right
[ 14, 26 ] = ascii (fp) : "./hrtfs/elev-20/L-20e230a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e230a.dat" right
[ 14, 27 ] = ascii (fp) : "./hrtfs/elev-20/L-20e225a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e225a.dat" right
[ 14, 28 ] = ascii (fp) : "./hrtfs/elev-20/L-20e220a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e220a.dat" right
[ 14, 29 ] = ascii (fp) : "./hrtfs/elev-20/L-20e215a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e215a.dat" right
[ 14, 30 ] = ascii (fp) : "./hrtfs/elev-20/L-20e210a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e210a.dat" right
[ 14, 31 ] = ascii (fp) : "./hrtfs/elev-20/L-20e205a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e205a.dat" right
[ 14, 32 ] = ascii (fp) : "./hrtfs/elev-20/L-20e200a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e200a.dat" right
[ 14, 33 ] = ascii (fp) : "./hrtfs/elev-20/L-20e195a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e195a.dat" right
[ 14, 34 ] = ascii (fp) : "./hrtfs/elev-20/L-20e190a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e190a.dat" right
[ 14, 35 ] = ascii (fp) : "./hrtfs/elev-20/L-20e185a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e185a.dat" right
[ 14, 36 ] = ascii (fp) : "./hrtfs/elev-20/L-20e180a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e180a.dat" right
[ 14, 37 ] = ascii (fp) : "./hrtfs/elev-20/L-20e175a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e175a.dat" right
[ 14, 38 ] = ascii (fp) : "./hrtfs/elev-20/L-20e170a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e170a.dat" right
[ 14, 39 ] = ascii (fp) : "./hrtfs/elev-20/L-20e165a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e165a.dat" right
[ 14, 40 ] = ascii (fp) : "./hrtfs/elev-20/L-20e160a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e160a.dat" right
[ 14, 41 ] = ascii (fp) : "./hrtfs/elev-20/L-20e155a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e155a.dat" right
[ 14, 42 ] = ascii (fp) : "./hrtfs/elev-20/L-20e150a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e150a.dat" right
[ 14, 43 ] = ascii (fp) : "./hrtfs/elev-20/L-20e145a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e145a.dat" right
[ 14, 44 ] = ascii (fp) : "./hrtfs/elev-20/L-20e140a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e140a.dat" right
[ 14, 45 ] = ascii (fp) : "./hrtfs/elev-20/L-20e135a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e135a.dat" right
[ 14, 46 ] = ascii (fp) : "./hrtfs/elev-20/L-20e130a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e130a.dat" right
[ 14, 47 ] = ascii (fp) : "./hrtfs/elev-20/L-20e125a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e125a.dat" right
[ 14, 48 ] = ascii (fp) : "./hrtfs/elev-20/L-20e120a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e120a.dat" right
[ 14, 49 ] = ascii (fp) : "./hrtfs/elev-20/L-20e115a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e115a.dat" right
[ 14, 50 ] = ascii (fp) : "./hrtfs/elev-20/L-20e110a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e110a.dat" right
[ 14, 51 ] = ascii (fp) : "./hrtfs/elev-20/L-20e105a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e105a.dat" right
[ 14, 52 ] = ascii (fp) : "./hrtfs/elev-20/L-20e100a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e100a.dat" right
[ 14, 53 ] = ascii (fp) : "./hrtfs/elev-20/L-20e095a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e095a.dat" right
[ 14, 54 ] = ascii (fp) : "./hrtfs/elev-20/L-20e090a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e090a.dat" right
[ 14, 55 ] = ascii (fp) : "./hrtfs/elev-20/L-20e085a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e085a.dat" right
[ 14, 56 ] = ascii (fp) : "./hrtfs/elev-20/L-20e080a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e080a.dat" right
[ 14, 57 ] = ascii (fp) : "./hrtfs/elev-20/L-20e075a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e075a.dat" right
[ 14, 58 ] = ascii (fp) : "./hrtfs/elev-20/L-20e070a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e070a.dat" right
[ 14, 59 ] = ascii (fp) : "./hrtfs/elev-20/L-20e065a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e065a.dat" right
[ 14, 60 ] = ascii (fp) : "./hrtfs/elev-20/L-20e060a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e060a.dat" right
[ 14, 61 ] = ascii (fp) : "./hrtfs/elev-20/L-20e055a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e055a.dat" right
[ 14, 62 ] = ascii (fp) : "./hrtfs/elev-20/L-20e050a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e050a.dat" right
[ 14, 63 ] = ascii (fp) : "./hrtfs/elev-20/L-20e045a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e045a.dat" right
[ 14, 64 ] = ascii (fp) : "./hrtfs/elev-20/L-20e040a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e040a.dat" right
[ 14, 65 ] = ascii (fp) : "./hrtfs/elev-20/L-20e035a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e035a.dat" right
[ 14, 66 ] = ascii (fp) : "./hrtfs/elev-20/L-20e030a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e030a.dat" right
[ 14, 67 ] = ascii (fp) : "./hrtfs/elev-20/L-20e025a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e025a.dat" right
[ 14, 68 ] = ascii (fp) : "./hrtfs/elev-20/L-20e020a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e020a.dat" right
[ 14, 69 ] = ascii (fp) : "./hrtfs/elev-20/L-20e015a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e015a.dat" right
[ 14, 70 ] = ascii (fp) : "./hrtfs/elev-20/L-20e010a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e010a.dat" right
[ 14, 71 ] = ascii (fp) : "./hrtfs/elev-20/L-20e005a.dat" left
           + ascii (fp) : "./hrtfs/elev-20/R-20e005a.dat" right

[ 15,  0 ] = ascii (fp) : "./hrtfs/elev-15/L-15e000a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e000a.dat" right
[ 15,  1 ] = ascii (fp) : "./hrtfs/elev-15/L-15e355a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e355a.dat" right
[ 15,  2 ] = ascii (fp) : "./hrtfs/elev-15/L-15e350a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e350a.dat" right
[ 15,  3 ] = ascii (fp) : "./hrtfs/elev-15/L-15e345a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e345a.dat" right
[ 15,  4 ] = ascii (fp) : "./hrtfs/elev-15/L-15e340a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e340a.dat" right
[ 15,  5 ] = ascii (fp) : "./hrtfs/elev-15/L-15e335a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e335a.dat" right
[ 15,  6 ] = ascii (fp) : "./hrtfs/elev-15/L-15e330a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e330a.dat" right
[ 15,  7 ] = ascii (fp) : "./hrtfs/elev-15/L-15e325a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e325a.dat" right
[ 15,  8 ] = ascii (fp) : "./hrtfs/elev-15/L-15e320a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e320a.dat" right
[ 15,  9 ] = ascii (fp) : "./hrtfs/elev-15/L-15e315a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e315a.dat" right
[ 15, 10 ] = ascii (fp) : "./hrtfs/elev-15/L-15e310a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e310a.dat" right
[ 15, 11 ] = ascii (fp) : "./hrtfs/elev-15/L-15e305a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e305a.dat" right
[ 15, 12 ] = ascii (fp) : "./hrtfs/elev-15/L-15e300a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e300a.dat" right
[ 15, 13 ] = ascii (fp) : "./hrtfs/elev-15/L-15e295a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e295a.dat" right
[ 15, 14 ] = ascii (fp) : "./hrtfs/elev-15/L-15e290a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e290a.dat" right
[ 15, 15 ] = ascii (fp) : "./hrtfs/elev-15/L-15e285a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e285a.dat" right
[ 15, 16 ] = ascii (fp) : "./hrtfs/elev-15/L-15e280a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e280a.dat" right
[ 15, 17 ] = ascii (fp) : "./hrtfs/elev-15/L-15e275a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e275a.dat" right
[ 15, 18 ] = ascii (fp) : "./hrtfs/elev-15/L-15e270a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e270a.dat" right
[ 15, 19 ] = ascii (fp) : "./hrtfs/elev-15/L-15e265a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e265a.dat" right
[ 15, 20 ] = ascii (fp) : "./hrtfs/elev-15/L-15e260a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e260a.dat" right
[ 15, 21 ] = ascii (fp) : "./hrtfs/elev-15/L-15e255a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e255a.dat" right
[ 15, 22 ] = ascii (fp) : "./hrtfs/elev-15/L-15e250a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e250a.dat" right
[ 15, 23 ] = ascii (fp) : "./hrtfs/elev-15/L-15e245a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e245a.dat" right
[ 15, 24 ] = ascii (fp) : "./hrtfs/elev-15/L-15e240a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e240a.dat" right
[ 15, 25 ] = ascii (fp) : "./hrtfs/elev-15/L-15e235a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e235a.dat" right
[ 15, 26 ] = ascii (fp) : "./hrtfs/elev-15/L-15e230a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e230a.dat" right
[ 15, 27 ] = ascii (fp) : "./hrtfs/elev-15/L-15e225a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e225a.dat" right
[ 15, 28 ] = ascii (fp) : "./hrtfs/elev-15/L-15e220a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e220a.dat" right
[ 15, 29 ] = ascii (fp) : "./hrtfs/elev-15/L-15e215a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e215a.dat" right
[ 15, 30 ] = ascii (fp) : "./hrtfs/elev-15/L-15e210a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e210a.dat" right
[ 15, 31 ] = ascii (fp) : "./hrtfs/elev-15/L-15e205a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e205a.dat" right
[ 15, 32 ] = ascii (fp) : "./hrtfs/elev-15/L-15e200a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e200a.dat" right
[ 15, 33 ] = ascii (fp) : "./hrtfs/elev-15/L-15e195a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e195a.dat" right
[ 15, 34 ] = ascii (fp) : "./hrtfs/elev-15/L-15e190a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e190a.dat" right
[ 15, 35 ] = ascii (fp) : "./hrtfs/elev-15/L-15e185a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e185a.dat" right
[ 15, 36 ] = ascii (fp) : "./hrtfs/elev-15/L-15e180a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e180a.dat" right
[ 15, 37 ] = ascii (fp) : "./hrtfs/elev-15/L-15e175a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e175a.dat" right
[ 15, 38 ] = ascii (fp) : "./hrtfs/elev-15/L-15e170a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e170a.dat" right
[ 15, 39 ] = ascii (fp) : "./hrtfs/elev-15/L-15e165a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e165a.dat" right
[ 15, 40 ] = ascii (fp) : "./hrtfs/elev-15/L-15e160a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e160a.dat" right
[ 15, 41 ] = ascii (fp) : "./hrtfs/elev-15/L-15e155a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e155a.dat" right
[ 15, 42 ] = ascii (fp) : "./hrtfs/elev-15/L-15e150a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e150a.dat" right
[ 15, 43 ] = ascii (fp) : "./hrtfs/elev-15/L-15e145a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e145a.dat" right
[ 15, 44 ] = ascii (fp) : "./hrtfs/elev-15/L-15e140a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e140a.dat" right
[ 15, 45 ] = ascii (fp) : "./hrtfs/elev-15/L-15e135a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e135a.dat" right
[ 15, 46 ] = ascii (fp) : "./hrtfs/elev-15/L-15e130a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e130a.dat" right
[ 15, 47 ] = ascii (fp) : "./hrtfs/elev-15/L-15e125a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e125a.dat" right
[ 15, 48 ] = ascii (fp) : "./hrtfs/elev-15/L-15e120a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e120a.dat" right
[ 15, 49 ] = ascii (fp) : "./hrtfs/elev-15/L-15e115a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e115a.dat" right
[ 15, 50 ] = ascii (fp) : "./hrtfs/elev-15/L-15e110a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e110a.dat" right
[ 15, 51 ] = ascii (fp) : "./hrtfs/elev-15/L-15e105a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e105a.dat" right
[ 15, 52 ] = ascii (fp) : "./hrtfs/elev-15/L-15e100a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e100a.dat" right
[ 15, 53 ] = ascii (fp) : "./hrtfs/elev-15/L-15e095a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e095a.dat" right
[ 15, 54 ] = ascii (fp) : "./hrtfs/elev-15/L-15e090a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e090a.dat" right
[ 15, 55 ] = ascii (fp) : "./hrtfs/elev-15/L-15e085a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e085a.dat" right
[ 15, 56 ] = ascii (fp) : "./hrtfs/elev-15/L-15e080a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e080a.dat" right
[ 15, 57 ] = ascii (fp) : "./hrtfs/elev-15/L-15e075a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e075a.dat" right
[ 15, 58 ] = ascii (fp) : "./hrtfs/elev-15/L-15e070a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e070a.dat" right
[ 15, 59 ] = ascii (fp) : "./hrtfs/elev-15/L-15e065a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e065a.dat" right
[ 15, 60 ] = ascii (fp) : "./hrtfs/elev-15/L-15e060a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e060a.dat" right
[ 15, 61 ] = ascii (fp) : "./hrtfs/elev-15/L-15e055a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e055a.dat" right
[ 15, 62 ] = ascii (fp) : "./hrtfs/elev-15/L-15e050a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e050a.dat" right
[ 15, 63 ] = ascii (fp) : "./hrtfs/elev-15/L-15e045a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e045a.dat" right
[ 15, 64 ] = ascii (fp) : "./hrtfs/elev-15/L-15e040a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e040a.dat" right
[ 15, 65 ] = ascii (fp) : "./hrtfs/elev-15/L-15e035a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e035a.dat" right
[ 15, 66 ] = ascii (fp) : "./hrtfs/elev-15/L-15e030a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e030a.dat" right
[ 15, 67 ] = ascii (fp) : "./hrtfs/elev-15/L-15e025a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e025a.dat" right
[ 15, 68 ] = ascii (fp) : "./hrtfs/elev-15/L-15e020a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e020a.dat" right
[ 15, 69 ] = ascii (fp) : "./hrtfs/elev-15/L-15e015a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e015a.dat" right
[ 15, 70 ] = ascii (fp) : "./hrtfs/elev-15/L-15e010a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e010a.dat" right
[ 15, 71 ] = ascii (fp) : "./hrtfs/elev-15/L-15e005a.dat" left
           + ascii (fp) : "./hrtfs/elev-15/R-15e005a.dat" right

[ 16,  0 ] = ascii (fp) : "./hrtfs/elev-10/L-10e000a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e000a.dat" right
[ 16,  1 ] = ascii (fp) : "./hrtfs/elev-10/L-10e355a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e355a.dat" right
[ 16,  2 ] = ascii (fp) : "./hrtfs/elev-10/L-10e350a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e350a.dat" right
[ 16,  3 ] = ascii (fp) : "./hrtfs/elev-10/L-10e345a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e345a.dat" right
[ 16,  4 ] = ascii (fp) : "./hrtfs/elev-10/L-10e340a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e340a.dat" right
[ 16,  5 ] = ascii (fp) : "./hrtfs/elev-10/L-10e335a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e335a.dat" right
[ 16,  6 ] = ascii (fp) : "./hrtfs/elev-10/L-10e330a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e330a.dat" right
[ 16,  7 ] = ascii (fp) : "./hrtfs/elev-10/L-10e325a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e325a.dat" right
[ 16,  8 ] = ascii (fp) : "./hrtfs/elev-10/L-10e320a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e320a.dat" right
[ 16,  9 ] = ascii (fp) : "./hrtfs/elev-10/L-10e315a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e315a.dat" right
[ 16, 10 ] = ascii (fp) : "./hrtfs/elev-10/L-10e310a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e310a.dat" right
[ 16, 11 ] = ascii (fp) : "./hrtfs/elev-10/L-10e305a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e305a.dat" right
[ 16, 12 ] = ascii (fp) : "./hrtfs/elev-10/L-10e300a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e300a.dat" right
[ 16, 13 ] = ascii (fp) : "./hrtfs/elev-10/L-10e295a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e295a.dat" right
[ 16, 14 ] = ascii (fp) : "./hrtfs/elev-10/L-10e290a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e290a.dat" right
[ 16, 15 ] = ascii (fp) : "./hrtfs/elev-10/L-10e285a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e285a.dat" right
[ 16, 16 ] = ascii (fp) : "./hrtfs/elev-10/L-10e280a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e280a.dat" right
[ 16, 17 ] = ascii (fp) : "./hrtfs/elev-10/L-10e275a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e275a.dat" right
[ 16, 18 ] = ascii (fp) : "./hrtfs/elev-10/L-10e270a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e270a.dat" right
[ 16, 19 ] = ascii (fp) : "./hrtfs/elev-10/L-10e265a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e265a.dat" right
[ 16, 20 ] = ascii (fp) : "./hrtfs/elev-10/L-10e260a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e260a.dat" right
[ 16, 21 ] = ascii (fp) : "./hrtfs/elev-10/L-10e255a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e255a.dat" right
[ 16, 22 ] = ascii (fp) : "./hrtfs/elev-10/L-10e250a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e250a.dat" right
[ 16, 23 ] = ascii (fp) : "./hrtfs/elev-10/L-10e245a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e245a.dat" right
[ 16, 24 ] = ascii (fp) : "./hrtfs/elev-10/L-10e240a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e240a.dat" right
[ 16, 25 ] = ascii (fp) : "./hrtfs/elev-10/L-10e235a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e235a.dat" right
[ 16, 26 ] = ascii (fp) : "./hrtfs/elev-10/L-10e230a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e230a.dat" right
[ 16, 27 ] = ascii (fp) : "./hrtfs/elev-10/L-10e225a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e225a.dat" right
[ 16, 28 ] = ascii (fp) : "./hrtfs/elev-10/L-10e220a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e220a.dat" right
[ 16, 29 ] = ascii (fp) : "./hrtfs/elev-10/L-10e215a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e215a.dat" right
[ 16, 30 ] = ascii (fp) : "./hrtfs/elev-10/L-10e210a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e210a.dat" right
[ 16, 31 ] = ascii (fp) : "./hrtfs/elev-10/L-10e205a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e205a.dat" right
[ 16, 32 ] = ascii (fp) : "./hrtfs/elev-10/L-10e200a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e200a.dat" right
[ 16, 33 ] = ascii (fp) : "./hrtfs/elev-10/L-10e195a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e195a.dat" right
[ 16, 34 ] = ascii (fp) : "./hrtfs/elev-10/L-10e190a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e190a.dat" right
[ 16, 35 ] = ascii (fp) : "./hrtfs/elev-10/L-10e185a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e185a.dat" right
[ 16, 36 ] = ascii (fp) : "./hrtfs/elev-10/L-10e180a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e180a.dat" right
[ 16, 37 ] = ascii (fp) : "./hrtfs/elev-10/L-10e175a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e175a.dat" right
[ 16, 38 ] = ascii (fp) : "./hrtfs/elev-10/L-10e170a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e170a.dat" right
[ 16, 39 ] = ascii (fp) : "./hrtfs/elev-10/L-10e165a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e165a.dat" right
[ 16, 40 ] = ascii (fp) : "./hrtfs/elev-10/L-10e160a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e160a.dat" right
[ 16, 41 ] = ascii (fp) : "./hrtfs/elev-10/L-10e155a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e155a.dat" right
[ 16, 42 ] = ascii (fp) : "./hrtfs/elev-10/L-10e150a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e150a.dat" right
[ 16, 43 ] = ascii (fp) : "./hrtfs/elev-10/L-10e145a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e145a.dat" right
[ 16, 44 ] = ascii (fp) : "./hrtfs/elev-10/L-10e140a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e140a.dat" right
[ 16, 45 ] = ascii (fp) : "./hrtfs/elev-10/L-10e135a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e135a.dat" right
[ 16, 46 ] = ascii (fp) : "./hrtfs/elev-10/L-10e130a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e130a.dat" right
[ 16, 47 ] = ascii (fp) : "./hrtfs/elev-10/L-10e125a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e125a.dat" right
[ 16, 48 ] = ascii (fp) : "./hrtfs/elev-10/L-10e120a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e120a.dat" right
[ 16, 49 ] = ascii (fp) : "./hrtfs/elev-10/L-10e115a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e115a.dat" right
[ 16, 50 ] = ascii (fp) : "./hrtfs/elev-10/L-10e110a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e110a.dat" right
[ 16, 51 ] = ascii (fp) : "./hrtfs/elev-10/L-10e105a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e105a.dat" right
[ 16, 52 ] = ascii (fp) : "./hrtfs/elev-10/L-10e100a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e100a.dat" right
[ 16, 53 ] = ascii (fp) : "./hrtfs/elev-10/L-10e095a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e095a.dat" right
[ 16, 54 ] = ascii (fp) : "./hrtfs/elev-10/L-10e090a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e090a.dat" right
[ 16, 55 ] = ascii (fp) : "./hrtfs/elev-10/L-10e085a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e085a.dat" right
[ 16, 56 ] = ascii (fp) : "./hrtfs/elev-10/L-10e080a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e080a.dat" right
[ 16, 57 ] = ascii (fp) : "./hrtfs/elev-10/L-10e075a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e075a.dat" right
[ 16, 58 ] = ascii (fp) : "./hrtfs/elev-10/L-10e070a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e070a.dat" right
[ 16, 59 ] = ascii (fp) : "./hrtfs/elev-10/L-10e065a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e065a.dat" right
[ 16, 60 ] = ascii (fp) : "./hrtfs/elev-10/L-10e060a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e060a.dat" right
[ 16, 61 ] = ascii (fp) : "./hrtfs/elev-10/L-10e055a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e055a.dat" right
[ 16, 62 ] = ascii (fp) : "./hrtfs/elev-10/L-10e050a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e050a.dat" right
[ 16, 63 ] = ascii (fp) : "./hrtfs/elev-10/L-10e045a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e045a.dat" right
[ 16, 64 ] = ascii (fp) : "./hrtfs/elev-10/L-10e040a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e040a.dat" right
[ 16, 65 ] = ascii (fp) : "./hrtfs/elev-10/L-10e035a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e035a.dat" right
[ 16, 66 ] = ascii (fp) : "./hrtfs/elev-10/L-10e030a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e030a.dat" right
[ 16, 67 ] = ascii (fp) : "./hrtfs/elev-10/L-10e025a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e025a.dat" right
[ 16, 68 ] = ascii (fp) : "./hrtfs/elev-10/L-10e020a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e020a.dat" right
[ 16, 69 ] = ascii (fp) : "./hrtfs/elev-10/L-10e015a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e015a.dat" right
[ 16, 70 ] = ascii (fp) : "./hrtfs/elev-10/L-10e010a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e010a.dat" right
[ 16, 71 ] = ascii (fp) : "./hrtfs/elev-10/L-10e005a.dat" left
           + ascii (fp) : "./hrtfs/elev-10/R-10e005a.dat" right

[ 17,  0 ] = ascii (fp) : "./hrtfs/elev-5/L-5e000a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e000a.dat" right
[ 17,  1 ] = ascii (fp) : "./hrtfs/elev-5/L-5e355a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e355a.dat" right
[ 17,  2 ] = ascii (fp) : "./hrtfs/elev-5/L-5e350a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e350a.dat" right
[ 17,  3 ] = ascii (fp) : "./hrtfs/elev-5/L-5e345a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e345a.dat" right
[ 17,  4 ] = ascii (fp) : "./hrtfs/elev-5/L-5e340a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e340a.dat" right
[ 17,  5 ] = ascii (fp) : "./hrtfs/elev-5/L-5e335a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e335a.dat" right
[ 17,  6 ] = ascii (fp) : "./hrtfs/elev-5/L-5e330a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e330a.dat" right
[ 17,  7 ] = ascii (fp) : "./hrtfs/elev-5/L-5e325a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e325a.dat" right
[ 17,  8 ] = ascii (fp) : "./hrtfs/elev-5/L-5e320a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e320a.dat" right
[ 17,  9 ] = ascii (fp) : "./hrtfs/elev-5/L-5e315a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e315a.dat" right
[ 17, 10 ] = ascii (fp) : "./hrtfs/elev-5/L-5e310a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e310a.dat" right
[ 17, 11 ] = ascii (fp) : "./hrtfs/elev-5/L-5e305a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e305a.dat" right
[ 17, 12 ] = ascii (fp) : "./hrtfs/elev-5/L-5e300a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e300a.dat" right
[ 17, 13 ] = ascii (fp) : "./hrtfs/elev-5/L-5e295a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e295a.dat" right
[ 17, 14 ] = ascii (fp) : "./hrtfs/elev-5/L-5e290a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e290a.dat" right
[ 17, 15 ] = ascii (fp) : "./hrtfs/elev-5/L-5e285a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e285a.dat" right
[ 17, 16 ] = ascii (fp) : "./hrtfs/elev-5/L-5e280a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e280a.dat" right
[ 17, 17 ] = ascii (fp) : "./hrtfs/elev-5/L-5e275a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e275a.dat" right
[ 17, 18 ] = ascii (fp) : "./hrtfs/elev-5/L-5e270a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e270a.dat" right
[ 17, 19 ] = ascii (fp) : "./hrtfs/elev-5/L-5e265a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e265a.dat" right
[ 17, 20 ] = ascii (fp) : "./hrtfs/elev-5/L-5e260a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e260a.dat" right
[ 17, 21 ] = ascii (fp) : "./hrtfs/elev-5/L-5e255a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e255a.dat" right
[ 17, 22 ] = ascii (fp) : "./hrtfs/elev-5/L-5e250a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e250a.dat" right
[ 17, 23 ] = ascii (fp) : "./hrtfs/elev-5/L-5e245a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e245a.dat" right
[ 17, 24 ] = ascii (fp) : "./hrtfs/elev-5/L-5e240a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e240a.dat" right
[ 17, 25 ] = ascii (fp) : "./hrtfs/elev-5/L-5e235a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e235a.dat" right
[ 17, 26 ] = ascii (fp) : "./hrtfs/elev-5/L-5e230a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e230a.dat" right
[ 17, 27 ] = ascii (fp) : "./hrtfs/elev-5/L-5e225a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e225a.dat" right
[ 17, 28 ] = ascii (fp) : "./hrtfs/elev-5/L-5e220a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e220a.dat" right
[ 17, 29 ] = ascii (fp) : "./hrtfs/elev-5/L-5e215a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e215a.dat" right
[ 17, 30 ] = ascii (fp) : "./hrtfs/elev-5/L-5e210a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e210a.dat" right
[ 17, 31 ] = ascii (fp) : "./hrtfs/elev-5/L-5e205a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e205a.dat" right
[ 17, 32 ] = ascii (fp) : "./hrtfs/elev-5/L-5e200a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e200a.dat" right
[ 17, 33 ] = ascii (fp) : "./hrtfs/elev-5/L-5e195a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e195a.dat" right
[ 17, 34 ] = ascii (fp) : "./hrtfs/elev-5/L-5e190a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e190a.dat" right
[ 17, 35 ] = ascii (fp) : "./hrtfs/elev-5/L-5e185a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e185a.dat" right
[ 17, 36 ] = ascii (fp) : "./hrtfs/elev-5/L-5e180a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e180a.dat" right
[ 17, 37 ] = ascii (fp) : "./hrtfs/elev-5/L-5e175a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e175a.dat" right
[ 17, 38 ] = ascii (fp) : "./hrtfs/elev-5/L-5e170a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e170a.dat" right
[ 17, 39 ] = ascii (fp) : "./hrtfs/elev-5/L-5e165a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e165a.dat" right
[ 17, 40 ] = ascii (fp) : "./hrtfs/elev-5/L-5e160a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e160a.dat" right
[ 17, 41 ] = ascii (fp) : "./hrtfs/elev-5/L-5e155a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e155a.dat" right
[ 17, 42 ] = ascii (fp) : "./hrtfs/elev-5/L-5e150a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e150a.dat" right
[ 17, 43 ] = ascii (fp) : "./hrtfs/elev-5/L-5e145a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e145a.dat" right
[ 17, 44 ] = ascii (fp) : "./hrtfs/elev-5/L-5e140a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e140a.dat" right
[ 17, 45 ] = ascii (fp) : "./hrtfs/elev-5/L-5e135a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e135a.dat" right
[ 17, 46 ] = ascii (fp) : "./hrtfs/elev-5/L-5e130a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e130a.dat" right
[ 17, 47 ] = ascii (fp) : "./hrtfs/elev-5/L-5e125a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e125a.dat" right
[ 17, 48 ] = ascii (fp) : "./hrtfs/elev-5/L-5e120a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e120a.dat" right
[ 17, 49 ] = ascii (fp) : "./hrtfs/elev-5/L-5e115a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e115a.dat" right
[ 17, 50 ] = ascii (fp) : "./hrtfs/elev-5/L-5e110a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e110a.dat" right
[ 17, 51 ] = ascii (fp) : "./hrtfs/elev-5/L-5e105a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e105a.dat" right
[ 17, 52 ] = ascii (fp) : "./hrtfs/elev-5/L-5e100a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e100a.dat" right
[ 17, 53 ] = ascii (fp) : "./hrtfs/elev-5/L-5e095a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e095a.dat" right
[ 17, 54 ] = ascii (fp) : "./hrtfs/elev-5/L-5e090a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e090a.dat" right
[ 17, 55 ] = ascii (fp) : "./hrtfs/elev-5/L-5e085a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e085a.dat" right
[ 17, 56 ] = ascii (fp) : "./hrtfs/elev-5/L-5e080a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e080a.dat" right
[ 17, 57 ] = ascii (fp) : "./hrtfs/elev-5/L-5e075a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e075a.dat" right
[ 17, 58 ] = ascii (fp) : "./hrtfs/elev-5/L-5e070a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e070a.dat" right
[ 17, 59 ] = ascii (fp) : "./hrtfs/elev-5/L-5e065a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e065a.dat" right
[ 17, 60 ] = ascii (fp) : "./hrtfs/elev-5/L-5e060a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e060a.dat" right
[ 17, 61 ] = ascii (fp) : "./hrtfs/elev-5/L-5e055a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e055a.dat" right
[ 17, 62 ] = ascii (fp) : "./hrtfs/elev-5/L-5e050a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e050a.dat" right
[ 17, 63 ] = ascii (fp) : "./hrtfs/elev-5/L-5e045a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e045a.dat" right
[ 17, 64 ] = ascii (fp) : "./hrtfs/elev-5/L-5e040a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e040a.dat" right
[ 17, 65 ] = ascii (fp) : "./hrtfs/elev-5/L-5e035a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e035a.dat" right
[ 17, 66 ] = ascii (fp) : "./hrtfs/elev-5/L-5e030a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e030a.dat" right
[ 17, 67 ] = ascii (fp) : "./hrtfs/elev-5/L-5e025a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e025a.dat" right
[ 17, 68 ] = ascii (fp) : "./hrtfs/elev-5/L-5e020a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e020a.dat" right
[ 17, 69 ] = ascii (fp) : "./hrtfs/elev-5/L-5e015a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e015a.dat" right
[ 17, 70 ] = ascii (fp) : "./hrtfs/elev-5/L-5e010a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e010a.dat" right
[ 17, 71 ] = ascii (fp) : "./hrtfs/elev-5/L-5e005a.dat" left
           + ascii (fp) : "./hrtfs/elev-5/R-5e005a.dat" right

[ 18,  0 ] = ascii (fp) : "./hrtfs/elev0/L0e000a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e000a.dat" right
[ 18,  1 ] = ascii (fp) : "./hrtfs/elev0/L0e355a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e355a.dat" right
[ 18,  2 ] = ascii (fp) : "./hrtfs/elev0/L0e350a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e350a.dat" right
[ 18,  3 ] = ascii (fp) : "./hrtfs/elev0/L0e345a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e345a.dat" right
[ 18,  4 ] = ascii (fp) : "./hrtfs/elev0/L0e340a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e340a.dat" right
[ 18,  5 ] = ascii (fp) : "./hrtfs/elev0/L0e335a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e335a.dat" right
[ 18,  6 ] = ascii (fp) : "./hrtfs/elev0/L0e330a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e330a.dat" right
[ 18,  7 ] = ascii (fp) : "./hrtfs/elev0/L0e325a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e325a.dat" right
[ 18,  8 ] = ascii (fp) : "./hrtfs/elev0/L0e320a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e320a.dat" right
[ 18,  9 ] = ascii (fp) : "./hrtfs/elev0/L0e315a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e315a.dat" right
[ 18, 10 ] = ascii (fp) : "./hrtfs/elev0/L0e310a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e310a.dat" right
[ 18, 11 ] = ascii (fp) : "./hrtfs/elev0/L0e305a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e305a.dat" right
[ 18, 12 ] = ascii (fp) : "./hrtfs/elev0/L0e300a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e300a.dat" right
[ 18, 13 ] = ascii (fp) : "./hrtfs/elev0/L0e295a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e295a.dat" right
[ 18, 14 ] = ascii (fp) : "./hrtfs/elev0/L0e290a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e290a.dat" right
[ 18, 15 ] = ascii (fp) : "./hrtfs/elev0/L0e285a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e285a.dat" right
[ 18, 16 ] = ascii (fp) : "./hrtfs/elev0/L0e280a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e280a.dat" right
[ 18, 17 ] = ascii (fp) : "./hrtfs/elev0/L0e275a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e275a.dat" right
[ 18, 18 ] = ascii (fp) : "./hrtfs/elev0/L0e270a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e270a.dat" right
[ 18, 19 ] = ascii (fp) : "./hrtfs/elev0/L0e265a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e265a.dat" right
[ 18, 20 ] = ascii (fp) : "./hrtfs/elev0/L0e260a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e260a.dat" right
[ 18, 21 ] = ascii (fp) : "./hrtfs/elev0/L0e255a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e255a.dat" right
[ 18, 22 ] = ascii (fp) : "./hrtfs/elev0/L0e250a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e250a.dat" right
[ 18, 23 ] = ascii (fp) : "./hrtfs/elev0/L0e245a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e245a.dat" right
[ 18, 24 ] = ascii (fp) : "./hrtfs/elev0/L0e240a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e240a.dat" right
[ 18, 25 ] = ascii (fp) : "./hrtfs/elev0/L0e235a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e235a.dat" right
[ 18, 26 ] = ascii (fp) : "./hrtfs/elev0/L0e230a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e230a.dat" right
[ 18, 27 ] = ascii (fp) : "./hrtfs/elev0/L0e225a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e225a.dat" right
[ 18, 28 ] = ascii (fp) : "./hrtfs/elev0/L0e220a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e220a.dat" right
[ 18, 29 ] = ascii (fp) : "./hrtfs/elev0/L0e215a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e215a.dat" right
[ 18, 30 ] = ascii (fp) : "./hrtfs/elev0/L0e210a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e210a.dat" right
[ 18, 31 ] = ascii (fp) : "./hrtfs/elev0/L0e205a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e205a.dat" right
[ 18, 32 ] = ascii (fp) : "./hrtfs/elev0/L0e200a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e200a.dat" right
[ 18, 33 ] = ascii (fp) : "./hrtfs/elev0/L0e195a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e195a.dat" right
[ 18, 34 ] = ascii (fp) : "./hrtfs/elev0/L0e190a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e190a.dat" right
[ 18, 35 ] = ascii (fp) : "./hrtfs/elev0/L0e185a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e185a.dat" right
[ 18, 36 ] = ascii (fp) : "./hrtfs/elev0/L0e180a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e180a.dat" right
[ 18, 37 ] = ascii (fp) : "./hrtfs/elev0/L0e175a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e175a.dat" right
[ 18, 38 ] = ascii (fp) : "./hrtfs/elev0/L0e170a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e170a.dat" right
[ 18, 39 ] = ascii (fp) : "./hrtfs/elev0/L0e165a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e165a.dat" right
[ 18, 40 ] = ascii (fp) : "./hrtfs/elev0/L0e160a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e160a.dat" right
[ 18, 41 ] = ascii (fp) : "./hrtfs/elev0/L0e155a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e155a.dat" right
[ 18, 42 ] = ascii (fp) : "./hrtfs/elev0/L0e150a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e150a.dat" right
[ 18, 43 ] = ascii (fp) : "./hrtfs/elev0/L0e145a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e145a.dat" right
[ 18, 44 ] = ascii (fp) : "./hrtfs/elev0/L0e140a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e140a.dat" right
[ 18, 45 ] = ascii (fp) : "./hrtfs/elev0/L0e135a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e135a.dat" right
[ 18, 46 ] = ascii (fp) : "./hrtfs/elev0/L0e130a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e130a.dat" right
[ 18, 47 ] = ascii (fp) : "./hrtfs/elev0/L0e125a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e125a.dat" right
[ 18, 48 ] = ascii (fp) : "./hrtfs/elev0/L0e120a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e120a.dat" right
[ 18, 49 ] = ascii (fp) : "./hrtfs/elev0/L0e115a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e115a.dat" right
[ 18, 50 ] = ascii (fp) : "./hrtfs/elev0/L0e110a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e110a.dat" right
[ 18, 51 ] = ascii (fp) : "./hrtfs/elev0/L0e105a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e105a.dat" right
[ 18, 52 ] = ascii (fp) : "./hrtfs/elev0/L0e100a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e100a.dat" right
[ 18, 53 ] = ascii (fp) : "./hrtfs/elev0/L0e095a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e095a.dat" right
[ 18, 54 ] = ascii (fp) : "./hrtfs/elev0/L0e090a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e090a.dat" right
[ 18, 55 ] = ascii (fp) : "./hrtfs/elev0/L0e085a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e085a.dat" right
[ 18, 56 ] = ascii (fp) : "./hrtfs/elev0/L0e080a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e080a.dat" right
[ 18, 57 ] = ascii (fp) : "./hrtfs/elev0/L0e075a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e075a.dat" right
[ 18, 58 ] = ascii (fp) : "./hrtfs/elev0/L0e070a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e070a.dat" right
[ 18, 59 ] = ascii (fp) : "./hrtfs/elev0/L0e065a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e065a.dat" right
[ 18, 60 ] = ascii (fp) : "./hrtfs/elev0/L0e060a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e060a.dat" right
[ 18, 61 ] = ascii (fp) : "./hrtfs/elev0/L0e055a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e055a.dat" right
[ 18, 62 ] = ascii (fp) : "./hrtfs/elev0/L0e050a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e050a.dat" right
[ 18, 63 ] = ascii (fp) : "./hrtfs/elev0/L0e045a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e045a.dat" right
[ 18, 64 ] = ascii (fp) : "./hrtfs/elev0/L0e040a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e040a.dat" right
[ 18, 65 ] = ascii (fp) : "./hrtfs/elev0/L0e035a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e035a.dat" right
[ 18, 66 ] = ascii (fp) : "./hrtfs/elev0/L0e030a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e030a.dat" right
[ 18, 67 ] = ascii (fp) : "./hrtfs/elev0/L0e025a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e025a.dat" right
[ 18, 68 ] = ascii (fp) : "./hrtfs/elev0/L0e020a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e020a.dat" right
[ 18, 69 ] = ascii (fp) : "./hrtfs/elev0/L0e015a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e015a.dat" right
[ 18, 70 ] = ascii (fp) : "./hrtfs/elev0/L0e010a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e010a.dat" right
[ 18, 71 ] = ascii (fp) : "./hrtfs/elev0/L0e005a.dat" left
           + ascii (fp) : "./hrtfs/elev0/R0e005a.dat" right

[ 19,  0 ] = ascii (fp) : "./hrtfs/elev5/L5e000a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e000a.dat" right
[ 19,  1 ] = ascii (fp) : "./hrtfs/elev5/L5e355a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e355a.dat" right
[ 19,  2 ] = ascii (fp) : "./hrtfs/elev5/L5e350a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e350a.dat" right
[ 19,  3 ] = ascii (fp) : "./hrtfs/elev5/L5e345a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e345a.dat" right
[ 19,  4 ] = ascii (fp) : "./hrtfs/elev5/L5e340a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e340a.dat" right
[ 19,  5 ] = ascii (fp) : "./hrtfs/elev5/L5e335a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e335a.dat" right
[ 19,  6 ] = ascii (fp) : "./hrtfs/elev5/L5e330a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e330a.dat" right
[ 19,  7 ] = ascii (fp) : "./hrtfs/elev5/L5e325a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e325a.dat" right
[ 19,  8 ] = ascii (fp) : "./hrtfs/elev5/L5e320a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e320a.dat" right
[ 19,  9 ] = ascii (fp) : "./hrtfs/elev5/L5e315a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e315a.dat" right
[ 19, 10 ] = ascii (fp) : "./hrtfs/elev5/L5e310a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e310a.dat" right
[ 19, 11 ] = ascii (fp) : "./hrtfs/elev5/L5e305a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e305a.dat" right
[ 19, 12 ] = ascii (fp) : "./hrtfs/elev5/L5e300a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e300a.dat" right
[ 19, 13 ] = ascii (fp) : "./hrtfs/elev5/L5e295a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e295a.dat" right
[ 19, 14 ] = ascii (fp) : "./hrtfs/elev5/L5e290a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e290a.dat" right
[ 19, 15 ] = ascii (fp) : "./hrtfs/elev5/L5e285a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e285a.dat" right
[ 19, 16 ] = ascii (fp) : "./hrtfs/elev5/L5e280a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e280a.dat" right
[ 19, 17 ] = ascii (fp) : "./hrtfs/elev5/L5e275a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e275a.dat" right
[ 19, 18 ] = ascii (fp) : "./hrtfs/elev5/L5e270a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e270a.dat" right
[ 19, 19 ] = ascii (fp) : "./hrtfs/elev5/L5e265a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e265a.dat" right
[ 19, 20 ] = ascii (fp) : "./hrtfs/elev5/L5e260a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e260a.dat" right
[ 19, 21 ] = ascii (fp) : "./hrtfs/elev5/L5e255a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e255a.dat" right
[ 19, 22 ] = ascii (fp) : "./hrtfs/elev5/L5e250a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e250a.dat" right
[ 19, 23 ] = ascii (fp) : "./hrtfs/elev5/L5e245a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e245a.dat" right
[ 19, 24 ] = ascii (fp) : "./hrtfs/elev5/L5e240a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e240a.dat" right
[ 19, 25 ] = ascii (fp) : "./hrtfs/elev5/L5e235a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e235a.dat" right
[ 19, 26 ] = ascii (fp) : "./hrtfs/elev5/L5e230a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e230a.dat" right
[ 19, 27 ] = ascii (fp) : "./hrtfs/elev5/L5e225a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e225a.dat" right
[ 19, 28 ] = ascii (fp) : "./hrtfs/elev5/L5e220a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e220a.dat" right
[ 19, 29 ] = ascii (fp) : "./hrtfs/elev5/L5e215a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e215a.dat" right
[ 19, 30 ] = ascii (fp) : "./hrtfs/elev5/L5e210a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e210a.dat" right
[ 19, 31 ] = ascii (fp) : "./hrtfs/elev5/L5e205a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e205a.dat" right
[ 19, 32 ] = ascii (fp) : "./hrtfs/elev5/L5e200a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e200a.dat" right
[ 19, 33 ] = ascii (fp) : "./hrtfs/elev5/L5e195a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e195a.dat" right
[ 19, 34 ] = ascii (fp) : "./hrtfs/elev5/L5e190a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e190a.dat" right
[ 19, 35 ] = ascii (fp) : "./hrtfs/elev5/L5e185a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e185a.dat" right
[ 19, 36 ] = ascii (fp) : "./hrtfs/elev5/L5e180a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e180a.dat" right
[ 19, 37 ] = ascii (fp) : "./hrtfs/elev5/L5e175a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e175a.dat" right
[ 19, 38 ] = ascii (fp) : "./hrtfs/elev5/L5e170a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e170a.dat" right
[ 19, 39 ] = ascii (fp) : "./hrtfs/elev5/L5e165a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e165a.dat" right
[ 19, 40 ] = ascii (fp) : "./hrtfs/elev5/L5e160a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e160a.dat" right
[ 19, 41 ] = ascii (fp) : "./hrtfs/elev5/L5e155a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e155a.dat" right
[ 19, 42 ] = ascii (fp) : "./hrtfs/elev5/L5e150a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e150a.dat" right
[ 19, 43 ] = ascii (fp) : "./hrtfs/elev5/L5e145a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e145a.dat" right
[ 19, 44 ] = ascii (fp) : "./hrtfs/elev5/L5e140a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e140a.dat" right
[ 19, 45 ] = ascii (fp) : "./hrtfs/elev5/L5e135a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e135a.dat" right
[ 19, 46 ] = ascii (fp) : "./hrtfs/elev5/L5e130a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e130a.dat" right
[ 19, 47 ] = ascii (fp) : "./hrtfs/elev5/L5e125a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e125a.dat" right
[ 19, 48 ] = ascii (fp) : "./hrtfs/elev5/L5e120a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e120a.dat" right
[ 19, 49 ] = ascii (fp) : "./hrtfs/elev5/L5e115a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e115a.dat" right
[ 19, 50 ] = ascii (fp) : "./hrtfs/elev5/L5e110a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e110a.dat" right
[ 19, 51 ] = ascii (fp) : "./hrtfs/elev5/L5e105a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e105a.dat" right
[ 19, 52 ] = ascii (fp) : "./hrtfs/elev5/L5e100a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e100a.dat" right
[ 19, 53 ] = ascii (fp) : "./hrtfs/elev5/L5e095a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e095a.dat" right
[ 19, 54 ] = ascii (fp) : "./hrtfs/elev5/L5e090a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e090a.dat" right
[ 19, 55 ] = ascii (fp) : "./hrtfs/elev5/L5e085a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e085a.dat" right
[ 19, 56 ] = ascii (fp) : "./hrtfs/elev5/L5e080a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e080a.dat" right
[ 19, 57 ] = ascii (fp) : "./hrtfs/elev5/L5e075a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e075a.dat" right
[ 19, 58 ] = ascii (fp) : "./hrtfs/elev5/L5e070a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e070a.dat" right
[ 19, 59 ] = ascii (fp) : "./hrtfs/elev5/L5e065a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e065a.dat" right
[ 19, 60 ] = ascii (fp) : "./hrtfs/elev5/L5e060a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e060a.dat" right
[ 19, 61 ] = ascii (fp) : "./hrtfs/elev5/L5e055a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e055a.dat" right
[ 19, 62 ] = ascii (fp) : "./hrtfs/elev5/L5e050a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e050a.dat" right
[ 19, 63 ] = ascii (fp) : "./hrtfs/elev5/L5e045a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e045a.dat" right
[ 19, 64 ] = ascii (fp) : "./hrtfs/elev5/L5e040a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e040a.dat" right
[ 19, 65 ] = ascii (fp) : "./hrtfs/elev5/L5e035a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e035a.dat" right
[ 19, 66 ] = ascii (fp) : "./hrtfs/elev5/L5e030a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e030a.dat" right
[ 19, 67 ] = ascii (fp) : "./hrtfs/elev5/L5e025a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e025a.dat" right
[ 19, 68 ] = ascii (fp) : "./hrtfs/elev5/L5e020a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e020a.dat" right
[ 19, 69 ] = ascii (fp) : "./hrtfs/elev5/L5e015a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e015a.dat" right
[ 19, 70 ] = ascii (fp) : "./hrtfs/elev5/L5e010a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e010a.dat" right
[ 19, 71 ] = ascii (fp) : "./hrtfs/elev5/L5e005a.dat" left
           + ascii (fp) : "./hrtfs/elev5/R5e005a.dat" right

[ 20,  0 ] = ascii (fp) : "./hrtfs/elev10/L10e000a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e000a.dat" right
[ 20,  1 ] = ascii (fp) : "./hrtfs/elev10/L10e355a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e355a.dat" right
[ 20,  2 ] = ascii (fp) : "./hrtfs/elev10/L10e350a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e350a.dat" right
[ 20,  3 ] = ascii (fp) : "./hrtfs/elev10/L10e345a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e345a.dat" right
[ 20,  4 ] = ascii (fp) : "./hrtfs/elev10/L10e340a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e340a.dat" right
[ 20,  5 ] = ascii (fp) : "./hrtfs/elev10/L10e335a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e335a.dat" right
[ 20,  6 ] = ascii (fp) : "./hrtfs/elev10/L10e330a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e330a.dat" right
[ 20,  7 ] = ascii (fp) : "./hrtfs/elev10/L10e325a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e325a.dat" right
[ 20,  8 ] = ascii (fp) : "./hrtfs/elev10/L10e320a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e320a.dat" right
[ 20,  9 ] = ascii (fp) : "./hrtfs/elev10/L10e315a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e315a.dat" right
[ 20, 10 ] = ascii (fp) : "./hrtfs/elev10/L10e310a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e310a.dat" right
[ 20, 11 ] = ascii (fp) : "./hrtfs/elev10/L10e305a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e305a.dat" right
[ 20, 12 ] = ascii (fp) : "./hrtfs/elev10/L10e300a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e300a.dat" right
[ 20, 13 ] = ascii (fp) : "./hrtfs/elev10/L10e295a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e295a.dat" right
[ 20, 14 ] = ascii (fp) : "./hrtfs/elev10/L10e290a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e290a.dat" right
[ 20, 15 ] = ascii (fp) : "./hrtfs/elev10/L10e285a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e285a.dat" right
[ 20, 16 ] = ascii (fp) : "./hrtfs/elev10/L10e280a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e280a.dat" right
[ 20, 17 ] = ascii (fp) : "./hrtfs/elev10/L10e275a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e275a.dat" right
[ 20, 18 ] = ascii (fp) : "./hrtfs/elev10/L10e270a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e270a.dat" right
[ 20, 19 ] = ascii (fp) : "./hrtfs/elev10/L10e265a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e265a.dat" right
[ 20, 20 ] = ascii (fp) : "./hrtfs/elev10/L10e260a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e260a.dat" right
[ 20, 21 ] = ascii (fp) : "./hrtfs/elev10/L10e255a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e255a.dat" right
[ 20, 22 ] = ascii (fp) : "./hrtfs/elev10/L10e250a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e250a.dat" right
[ 20, 23 ] = ascii (fp) : "./hrtfs/elev10/L10e245a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e245a.dat" right
[ 20, 24 ] = ascii (fp) : "./hrtfs/elev10/L10e240a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e240a.dat" right
[ 20, 25 ] = ascii (fp) : "./hrtfs/elev10/L10e235a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e235a.dat" right
[ 20, 26 ] = ascii (fp) : "./hrtfs/elev10/L10e230a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e230a.dat" right
[ 20, 27 ] = ascii (fp) : "./hrtfs/elev10/L10e225a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e225a.dat" right
[ 20, 28 ] = ascii (fp) : "./hrtfs/elev10/L10e220a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e220a.dat" right
[ 20, 29 ] = ascii (fp) : "./hrtfs/elev10/L10e215a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e215a.dat" right
[ 20, 30 ] = ascii (fp) : "./hrtfs/elev10/L10e210a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e210a.dat" right
[ 20, 31 ] = ascii (fp) : "./hrtfs/elev10/L10e205a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e205a.dat" right
[ 20, 32 ] = ascii (fp) : "./hrtfs/elev10/L10e200a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e200a.dat" right
[ 20, 33 ] = ascii (fp) : "./hrtfs/elev10/L10e195a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e195a.dat" right
[ 20, 34 ] = ascii (fp) : "./hrtfs/elev10/L10e190a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e190a.dat" right
[ 20, 35 ] = ascii (fp) : "./hrtfs/elev10/L10e185a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e185a.dat" right
[ 20, 36 ] = ascii (fp) : "./hrtfs/elev10/L10e180a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e180a.dat" right
[ 20, 37 ] = ascii (fp) : "./hrtfs/elev10/L10e175a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e175a.dat" right
[ 20, 38 ] = ascii (fp) : "./hrtfs/elev10/L10e170a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e170a.dat" right
[ 20, 39 ] = ascii (fp) : "./hrtfs/elev10/L10e165a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e165a.dat" right
[ 20, 40 ] = ascii (fp) : "./hrtfs/elev10/L10e160a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e160a.dat" right
[ 20, 41 ] = ascii (fp) : "./hrtfs/elev10/L10e155a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e155a.dat" right
[ 20, 42 ] = ascii (fp) : "./hrtfs/elev10/L10e150a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e150a.dat" right
[ 20, 43 ] = ascii (fp) : "./hrtfs/elev10/L10e145a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e145a.dat" right
[ 20, 44 ] = ascii (fp) : "./hrtfs/elev10/L10e140a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e140a.dat" right
[ 20, 45 ] = ascii (fp) : "./hrtfs/elev10/L10e135a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e135a.dat" right
[ 20, 46 ] = ascii (fp) : "./hrtfs/elev10/L10e130a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e130a.dat" right
[ 20, 47 ] = ascii (fp) : "./hrtfs/elev10/L10e125a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e125a.dat" right
[ 20, 48 ] = ascii (fp) : "./hrtfs/elev10/L10e120a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e120a.dat" right
[ 20, 49 ] = ascii (fp) : "./hrtfs/elev10/L10e115a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e115a.dat" right
[ 20, 50 ] = ascii (fp) : "./hrtfs/elev10/L10e110a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e110a.dat" right
[ 20, 51 ] = ascii (fp) : "./hrtfs/elev10/L10e105a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e105a.dat" right
[ 20, 52 ] = ascii (fp) : "./hrtfs/elev10/L10e100a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e100a.dat" right
[ 20, 53 ] = ascii (fp) : "./hrtfs/elev10/L10e095a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e095a.dat" right
[ 20, 54 ] = ascii (fp) : "./hrtfs/elev10/L10e090a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e090a.dat" right
[ 20, 55 ] = ascii (fp) : "./hrtfs/elev10/L10e085a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e085a.dat" right
[ 20, 56 ] = ascii (fp) : "./hrtfs/elev10/L10e080a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e080a.dat" right
[ 20, 57 ] = ascii (fp) : "./hrtfs/elev10/L10e075a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e075a.dat" right
[ 20, 58 ] = ascii (fp) : "./hrtfs/elev10/L10e070a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e070a.dat" right
[ 20, 59 ] = ascii (fp) : "./hrtfs/elev10/L10e065a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e065a.dat" right
[ 20, 60 ] = ascii (fp) : "./hrtfs/elev10/L10e060a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e060a.dat" right
[ 20, 61 ] = ascii (fp) : "./hrtfs/elev10/L10e055a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e055a.dat" right
[ 20, 62 ] = ascii (fp) : "./hrtfs/elev10/L10e050a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e050a.dat" right
[ 20, 63 ] = ascii (fp) : "./hrtfs/elev10/L10e045a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e045a.dat" right
[ 20, 64 ] = ascii (fp) : "./hrtfs/elev10/L10e040a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e040a.dat" right
[ 20, 65 ] = ascii (fp) : "./hrtfs/elev10/L10e035a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e035a.dat" right
[ 20, 66 ] = ascii (fp) : "./hrtfs/elev10/L10e030a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e030a.dat" right
[ 20, 67 ] = ascii (fp) : "./hrtfs/elev10/L10e025a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e025a.dat" right
[ 20, 68 ] = ascii (fp) : "./hrtfs/elev10/L10e020a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e020a.dat" right
[ 20, 69 ] = ascii (fp) : "./hrtfs/elev10/L10e015a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e015a.dat" right
[ 20, 70 ] = ascii (fp) : "./hrtfs/elev10/L10e010a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e010a.dat" right
[ 20, 71 ] = ascii (fp) : "./hrtfs/elev10/L10e005a.dat" left
           + ascii (fp) : "./hrtfs/elev10/R10e005a.dat" right

[ 21,  0 ] = ascii (fp) : "./hrtfs/elev15/L15e000a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e000a.dat" right
[ 21,  1 ] = ascii (fp) : "./hrtfs/elev15/L15e355a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e355a.dat" right
[ 21,  2 ] = ascii (fp) : "./hrtfs/elev15/L15e350a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e350a.dat" right
[ 21,  3 ] = ascii (fp) : "./hrtfs/elev15/L15e345a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e345a.dat" right
[ 21,  4 ] = ascii (fp) : "./hrtfs/elev15/L15e340a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e340a.dat" right
[ 21,  5 ] = ascii (fp) : "./hrtfs/elev15/L15e335a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e335a.dat" right
[ 21,  6 ] = ascii (fp) : "./hrtfs/elev15/L15e330a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e330a.dat" right
[ 21,  7 ] = ascii (fp) : "./hrtfs/elev15/L15e325a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e325a.dat" right
[ 21,  8 ] = ascii (fp) : "./hrtfs/elev15/L15e320a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e320a.dat" right
[ 21,  9 ] = ascii (fp) : "./hrtfs/elev15/L15e315a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e315a.dat" right
[ 21, 10 ] = ascii (fp) : "./hrtfs/elev15/L15e310a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e310a.dat" right
[ 21, 11 ] = ascii (fp) : "./hrtfs/elev15/L15e305a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e305a.dat" right
[ 21, 12 ] = ascii (fp) : "./hrtfs/elev15/L15e300a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e300a.dat" right
[ 21, 13 ] = ascii (fp) : "./hrtfs/elev15/L15e295a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e295a.dat" right
[ 21, 14 ] = ascii (fp) : "./hrtfs/elev15/L15e290a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e290a.dat" right
[ 21, 15 ] = ascii (fp) : "./hrtfs/elev15/L15e285a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e285a.dat" right
[ 21, 16 ] = ascii (fp) : "./hrtfs/elev15/L15e280a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e280a.dat" right
[ 21, 17 ] = ascii (fp) : "./hrtfs/elev15/L15e275a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e275a.dat" right
[ 21, 18 ] = ascii (fp) : "./hrtfs/elev15/L15e270a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e270a.dat" right
[ 21, 19 ] = ascii (fp) : "./hrtfs/elev15/L15e265a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e265a.dat" right
[ 21, 20 ] = ascii (fp) : "./hrtfs/elev15/L15e260a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e260a.dat" right
[ 21, 21 ] = ascii (fp) : "./hrtfs/elev15/L15e255a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e255a.dat" right
[ 21, 22 ] = ascii (fp) : "./hrtfs/elev15/L15e250a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e250a.dat" right
[ 21, 23 ] = ascii (fp) : "./hrtfs/elev15/L15e245a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e245a.dat" right
[ 21, 24 ] = ascii (fp) : "./hrtfs/elev15/L15e240a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e240a.dat" right
[ 21, 25 ] = ascii (fp) : "./hrtfs/elev15/L15e235a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e235a.dat" right
[ 21, 26 ] = ascii (fp) : "./hrtfs/elev15/L15e230a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e230a.dat" right
[ 21, 27 ] = ascii (fp) : "./hrtfs/elev15/L15e225a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e225a.dat" right
[ 21, 28 ] = ascii (fp) : "./hrtfs/elev15/L15e220a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e220a.dat" right
[ 21, 29 ] = ascii (fp) : "./hrtfs/elev15/L15e215a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e215a.dat" right
[ 21, 30 ] = ascii (fp) : "./hrtfs/elev15/L15e210a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e210a.dat" right
[ 21, 31 ] = ascii (fp) : "./hrtfs/elev15/L15e205a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e205a.dat" right
[ 21, 32 ] = ascii (fp) : "./hrtfs/elev15/L15e200a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e200a.dat" right
[ 21, 33 ] = ascii (fp) : "./hrtfs/elev15/L15e195a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e195a.dat" right
[ 21, 34 ] = ascii (fp) : "./hrtfs/elev15/L15e190a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e190a.dat" right
[ 21, 35 ] = ascii (fp) : "./hrtfs/elev15/L15e185a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e185a.dat" right
[ 21, 36 ] = ascii (fp) : "./hrtfs/elev15/L15e180a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e180a.dat" right
[ 21, 37 ] = ascii (fp) : "./hrtfs/elev15/L15e175a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e175a.dat" right
[ 21, 38 ] = ascii (fp) : "./hrtfs/elev15/L15e170a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e170a.dat" right
[ 21, 39 ] = ascii (fp) : "./hrtfs/elev15/L15e165a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e165a.dat" right
[ 21, 40 ] = ascii (fp) : "./hrtfs/elev15/L15e160a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e160a.dat" right
[ 21, 41 ] = ascii (fp) : "./hrtfs/elev15/L15e155a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e155a.dat" right
[ 21, 42 ] = ascii (fp) : "./hrtfs/elev15/L15e150a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e150a.dat" right
[ 21, 43 ] = ascii (fp) : "./hrtfs/elev15/L15e145a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e145a.dat" right
[ 21, 44 ] = ascii (fp) : "./hrtfs/elev15/L15e140a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e140a.dat" right
[ 21, 45 ] = ascii (fp) : "./hrtfs/elev15/L15e135a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e135a.dat" right
[ 21, 46 ] = ascii (fp) : "./hrtfs/elev15/L15e130a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e130a.dat" right
[ 21, 47 ] = ascii (fp) : "./hrtfs/elev15/L15e125a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e125a.dat" right
[ 21, 48 ] = ascii (fp) : "./hrtfs/elev15/L15e120a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e120a.dat" right
[ 21, 49 ] = ascii (fp) : "./hrtfs/elev15/L15e115a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e115a.dat" right
[ 21, 50 ] = ascii (fp) : "./hrtfs/elev15/L15e110a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e110a.dat" right
[ 21, 51 ] = ascii (fp) : "./hrtfs/elev15/L15e105a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e105a.dat" right
[ 21, 52 ] = ascii (fp) : "./hrtfs/elev15/L15e100a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e100a.dat" right
[ 21, 53 ] = ascii (fp) : "./hrtfs/elev15/L15e095a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e095a.dat" right
[ 21, 54 ] = ascii (fp) : "./hrtfs/elev15/L15e090a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e090a.dat" right
[ 21, 55 ] = ascii (fp) : "./hrtfs/elev15/L15e085a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e085a.dat" right
[ 21, 56 ] = ascii (fp) : "./hrtfs/elev15/L15e080a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e080a.dat" right
[ 21, 57 ] = ascii (fp) : "./hrtfs/elev15/L15e075a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e075a.dat" right
[ 21, 58 ] = ascii (fp) : "./hrtfs/elev15/L15e070a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e070a.dat" right
[ 21, 59 ] = ascii (fp) : "./hrtfs/elev15/L15e065a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e065a.dat" right
[ 21, 60 ] = ascii (fp) : "./hrtfs/elev15/L15e060a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e060a.dat" right
[ 21, 61 ] = ascii (fp) : "./hrtfs/elev15/L15e055a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e055a.dat" right
[ 21, 62 ] = ascii (fp) : "./hrtfs/elev15/L15e050a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e050a.dat" right
[ 21, 63 ] = ascii (fp) : "./hrtfs/elev15/L15e045a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e045a.dat" right
[ 21, 64 ] = ascii (fp) : "./hrtfs/elev15/L15e040a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e040a.dat" right
[ 21, 65 ] = ascii (fp) : "./hrtfs/elev15/L15e035a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e035a.dat" right
[ 21, 66 ] = ascii (fp) : "./hrtfs/elev15/L15e030a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e030a.dat" right
[ 21, 67 ] = ascii (fp) : "./hrtfs/elev15/L15e025a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e025a.dat" right
[ 21, 68 ] = ascii (fp) : "./hrtfs/elev15/L15e020a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e020a.dat" right
[ 21, 69 ] = ascii (fp) : "./hrtfs/elev15/L15e015a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e015a.dat" right
[ 21, 70 ] = ascii (fp) : "./hrtfs/elev15/L15e010a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e010a.dat" right
[ 21, 71 ] = ascii (fp) : "./hrtfs/elev15/L15e005a.dat" left
           + ascii (fp) : "./hrtfs/elev15/R15e005a.dat" right

[ 22,  0 ] = ascii (fp) : "./hrtfs/elev20/L20e000a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e000a.dat" right
[ 22,  1 ] = ascii (fp) : "./hrtfs/elev20/L20e355a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e355a.dat" right
[ 22,  2 ] = ascii (fp) : "./hrtfs/elev20/L20e350a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e350a.dat" right
[ 22,  3 ] = ascii (fp) : "./hrtfs/elev20/L20e345a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e345a.dat" right
[ 22,  4 ] = ascii (fp) : "./hrtfs/elev20/L20e340a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e340a.dat" right
[ 22,  5 ] = ascii (fp) : "./hrtfs/elev20/L20e335a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e335a.dat" right
[ 22,  6 ] = ascii (fp) : "./hrtfs/elev20/L20e330a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e330a.dat" right
[ 22,  7 ] = ascii (fp) : "./hrtfs/elev20/L20e325a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e325a.dat" right
[ 22,  8 ] = ascii (fp) : "./hrtfs/elev20/L20e320a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e320a.dat" right
[ 22,  9 ] = ascii (fp) : "./hrtfs/elev20/L20e315a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e315a.dat" right
[ 22, 10 ] = ascii (fp) : "./hrtfs/elev20/L20e310a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e310a.dat" right
[ 22, 11 ] = ascii (fp) : "./hrtfs/elev20/L20e305a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e305a.dat" right
[ 22, 12 ] = ascii (fp) : "./hrtfs/elev20/L20e300a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e300a.dat" right
[ 22, 13 ] = ascii (fp) : "./hrtfs/elev20/L20e295a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e295a.dat" right
[ 22, 14 ] = ascii (fp) : "./hrtfs/elev20/L20e290a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e290a.dat" right
[ 22, 15 ] = ascii (fp) : "./hrtfs/elev20/L20e285a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e285a.dat" right
[ 22, 16 ] = ascii (fp) : "./hrtfs/elev20/L20e280a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e280a.dat" right
[ 22, 17 ] = ascii (fp) : "./hrtfs/elev20/L20e275a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e275a.dat" right
[ 22, 18 ] = ascii (fp) : "./hrtfs/elev20/L20e270a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e270a.dat" right
[ 22, 19 ] = ascii (fp) : "./hrtfs/elev20/L20e265a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e265a.dat" right
[ 22, 20 ] = ascii (fp) : "./hrtfs/elev20/L20e260a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e260a.dat" right
[ 22, 21 ] = ascii (fp) : "./hrtfs/elev20/L20e255a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e255a.dat" right
[ 22, 22 ] = ascii (fp) : "./hrtfs/elev20/L20e250a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e250a.dat" right
[ 22, 23 ] = ascii (fp) : "./hrtfs/elev20/L20e245a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e245a.dat" right
[ 22, 24 ] = ascii (fp) : "./hrtfs/elev20/L20e240a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e240a.dat" right
[ 22, 25 ] = ascii (fp) : "./hrtfs/elev20/L20e235a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e235a.dat" right
[ 22, 26 ] = ascii (fp) : "./hrtfs/elev20/L20e230a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e230a.dat" right
[ 22, 27 ] = ascii (fp) : "./hrtfs/elev20/L20e225a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e225a.dat" right
[ 22, 28 ] = ascii (fp) : "./hrtfs/elev20/L20e220a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e220a.dat" right
[ 22, 29 ] = ascii (fp) : "./hrtfs/elev20/L20e215a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e215a.dat" right
[ 22, 30 ] = ascii (fp) : "./hrtfs/elev20/L20e210a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e210a.dat" right
[ 22, 31 ] = ascii (fp) : "./hrtfs/elev20/L20e205a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e205a.dat" right
[ 22, 32 ] = ascii (fp) : "./hrtfs/elev20/L20e200a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e200a.dat" right
[ 22, 33 ] = ascii (fp) : "./hrtfs/elev20/L20e195a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e195a.dat" right
[ 22, 34 ] = ascii (fp) : "./hrtfs/elev20/L20e190a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e190a.dat" right
[ 22, 35 ] = ascii (fp) : "./hrtfs/elev20/L20e185a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e185a.dat" right
[ 22, 36 ] = ascii (fp) : "./hrtfs/elev20/L20e180a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e180a.dat" right
[ 22, 37 ] = ascii (fp) : "./hrtfs/elev20/L20e175a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e175a.dat" right
[ 22, 38 ] = ascii (fp) : "./hrtfs/elev20/L20e170a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e170a.dat" right
[ 22, 39 ] = ascii (fp) : "./hrtfs/elev20/L20e165a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e165a.dat" right
[ 22, 40 ] = ascii (fp) : "./hrtfs/elev20/L20e160a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e160a.dat" right
[ 22, 41 ] = ascii (fp) : "./hrtfs/elev20/L20e155a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e155a.dat" right
[ 22, 42 ] = ascii (fp) : "./hrtfs/elev20/L20e150a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e150a.dat" right
[ 22, 43 ] = ascii (fp) : "./hrtfs/elev20/L20e145a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e145a.dat" right
[ 22, 44 ] = ascii (fp) : "./hrtfs/elev20/L20e140a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e140a.dat" right
[ 22, 45 ] = ascii (fp) : "./hrtfs/elev20/L20e135a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e135a.dat" right
[ 22, 46 ] = ascii (fp) : "./hrtfs/elev20/L20e130a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e130a.dat" right
[ 22, 47 ] = ascii (fp) : "./hrtfs/elev20/L20e125a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e125a.dat" right
[ 22, 48 ] = ascii (fp) : "./hrtfs/elev20/L20e120a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e120a.dat" right
[ 22, 49 ] = ascii (fp) : "./hrtfs/elev20/L20e115a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e115a.dat" right
[ 22, 50 ] = ascii (fp) : "./hrtfs/elev20/L20e110a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e110a.dat" right
[ 22, 51 ] = ascii (fp) : "./hrtfs/elev20/L20e105a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e105a.dat" right
[ 22, 52 ] = ascii (fp) : "./hrtfs/elev20/L20e100a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e100a.dat" right
[ 22, 53 ] = ascii (fp) : "./hrtfs/elev20/L20e095a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e095a.dat" right
[ 22, 54 ] = ascii (fp) : "./hrtfs/elev20/L20e090a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e090a.dat" right
[ 22, 55 ] = ascii (fp) : "./hrtfs/elev20/L20e085a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e085a.dat" right
[ 22, 56 ] = ascii (fp) : "./hrtfs/elev20/L20e080a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e080a.dat" right
[ 22, 57 ] = ascii (fp) : "./hrtfs/elev20/L20e075a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e075a.dat" right
[ 22, 58 ] = ascii (fp) : "./hrtfs/elev20/L20e070a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e070a.dat" right
[ 22, 59 ] = ascii (fp) : "./hrtfs/elev20/L20e065a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e065a.dat" right
[ 22, 60 ] = ascii (fp) : "./hrtfs/elev20/L20e060a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e060a.dat" right
[ 22, 61 ] = ascii (fp) : "./hrtfs/elev20/L20e055a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e055a.dat" right
[ 22, 62 ] = ascii (fp) : "./hrtfs/elev20/L20e050a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e050a.dat" right
[ 22, 63 ] = ascii (fp) : "./hrtfs/elev20/L20e045a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e045a.dat" right
[ 22, 64 ] = ascii (fp) : "./hrtfs/elev20/L20e040a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e040a.dat" right
[ 22, 65 ] = ascii (fp) : "./hrtfs/elev20/L20e035a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e035a.dat" right
[ 22, 66 ] = ascii (fp) : "./hrtfs/elev20/L20e030a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e030a.dat" right
[ 22, 67 ] = ascii (fp) : "./hrtfs/elev20/L20e025a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e025a.dat" right
[ 22, 68 ] = ascii (fp) : "./hrtfs/elev20/L20e020a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e020a.dat" right
[ 22, 69 ] = ascii (fp) : "./hrtfs/elev20/L20e015a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e015a.dat" right
[ 22, 70 ] = ascii (fp) : "./hrtfs/elev20/L20e010a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e010a.dat" right
[ 22, 71 ] = ascii (fp) : "./hrtfs/elev20/L20e005a.dat" left
           + ascii (fp) : "./hrtfs/elev20/R20e005a.dat" right

[ 23,  0 ] = ascii (fp) : "./hrtfs/elev25/L25e000a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e000a.dat" right
[ 23,  1 ] = ascii (fp) : "./hrtfs/elev25/L25e355a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e355a.dat" right
[ 23,  2 ] = ascii (fp) : "./hrtfs/elev25/L25e350a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e350a.dat" right
[ 23,  3 ] = ascii (fp) : "./hrtfs/elev25/L25e345a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e345a.dat" right
[ 23,  4 ] = ascii (fp) : "./hrtfs/elev25/L25e340a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e340a.dat" right
[ 23,  5 ] = ascii (fp) : "./hrtfs/elev25/L25e335a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e335a.dat" right
[ 23,  6 ] = ascii (fp) : "./hrtfs/elev25/L25e330a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e330a.dat" right
[ 23,  7 ] = ascii (fp) : "./hrtfs/elev25/L25e325a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e325a.dat" right
[ 23,  8 ] = ascii (fp) : "./hrtfs/elev25/L25e320a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e320a.dat" right
[ 23,  9 ] = ascii (fp) : "./hrtfs/elev25/L25e315a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e315a.dat" right
[ 23, 10 ] = ascii (fp) : "./hrtfs/elev25/L25e310a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e310a.dat" right
[ 23, 11 ] = ascii (fp) : "./hrtfs/elev25/L25e305a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e305a.dat" right
[ 23, 12 ] = ascii (fp) : "./hrtfs/elev25/L25e300a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e300a.dat" right
[ 23, 13 ] = ascii (fp) : "./hrtfs/elev25/L25e295a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e295a.dat" right
[ 23, 14 ] = ascii (fp) : "./hrtfs/elev25/L25e290a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e290a.dat" right
[ 23, 15 ] = ascii (fp) : "./hrtfs/elev25/L25e285a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e285a.dat" right
[ 23, 16 ] = ascii (fp) : "./hrtfs/elev25/L25e280a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e280a.dat" right
[ 23, 17 ] = ascii (fp) : "./hrtfs/elev25/L25e275a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e275a.dat" right
[ 23, 18 ] = ascii (fp) : "./hrtfs/elev25/L25e270a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e270a.dat" right
[ 23, 19 ] = ascii (fp) : "./hrtfs/elev25/L25e265a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e265a.dat" right
[ 23, 20 ] = ascii (fp) : "./hrtfs/elev25/L25e260a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e260a.dat" right
[ 23, 21 ] = ascii (fp) : "./hrtfs/elev25/L25e255a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e255a.dat" right
[ 23, 22 ] = ascii (fp) : "./hrtfs/elev25/L25e250a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e250a.dat" right
[ 23, 23 ] = ascii (fp) : "./hrtfs/elev25/L25e245a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e245a.dat" right
[ 23, 24 ] = ascii (fp) : "./hrtfs/elev25/L25e240a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e240a.dat" right
[ 23, 25 ] = ascii (fp) : "./hrtfs/elev25/L25e235a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e235a.dat" right
[ 23, 26 ] = ascii (fp) : "./hrtfs/elev25/L25e230a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e230a.dat" right
[ 23, 27 ] = ascii (fp) : "./hrtfs/elev25/L25e225a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e225a.dat" right
[ 23, 28 ] = ascii (fp) : "./hrtfs/elev25/L25e220a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e220a.dat" right
[ 23, 29 ] = ascii (fp) : "./hrtfs/elev25/L25e215a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e215a.dat" right
[ 23, 30 ] = ascii (fp) : "./hrtfs/elev25/L25e210a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e210a.dat" right
[ 23, 31 ] = ascii (fp) : "./hrtfs/elev25/L25e205a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e205a.dat" right
[ 23, 32 ] = ascii (fp) : "./hrtfs/elev25/L25e200a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e200a.dat" right
[ 23, 33 ] = ascii (fp) : "./hrtfs/elev25/L25e195a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e195a.dat" right
[ 23, 34 ] = ascii (fp) : "./hrtfs/elev25/L25e190a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e190a.dat" right
[ 23, 35 ] = ascii (fp) : "./hrtfs/elev25/L25e185a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e185a.dat" right
[ 23, 36 ] = ascii (fp) : "./hrtfs/elev25/L25e180a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e180a.dat" right
[ 23, 37 ] = ascii (fp) : "./hrtfs/elev25/L25e175a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e175a.dat" right
[ 23, 38 ] = ascii (fp) : "./hrtfs/elev25/L25e170a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e170a.dat" right
[ 23, 39 ] = ascii (fp) : "./hrtfs/elev25/L25e165a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e165a.dat" right
[ 23, 40 ] = ascii (fp) : "./hrtfs/elev25/L25e160a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e160a.dat" right
[ 23, 41 ] = ascii (fp) : "./hrtfs/elev25/L25e155a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e155a.dat" right
[ 23, 42 ] = ascii (fp) : "./hrtfs/elev25/L25e150a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e150a.dat" right
[ 23, 43 ] = ascii (fp) : "./hrtfs/elev25/L25e145a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e145a.dat" right
[ 23, 44 ] = ascii (fp) : "./hrtfs/elev25/L25e140a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e140a.dat" right
[ 23, 45 ] = ascii (fp) : "./hrtfs/elev25/L25e135a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e135a.dat" right
[ 23, 46 ] = ascii (fp) : "./hrtfs/elev25/L25e130a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e130a.dat" right
[ 23, 47 ] = ascii (fp) : "./hrtfs/elev25/L25e125a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e125a.dat" right
[ 23, 48 ] = ascii (fp) : "./hrtfs/elev25/L25e120a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e120a.dat" right
[ 23, 49 ] = ascii (fp) : "./hrtfs/elev25/L25e115a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e115a.dat" right
[ 23, 50 ] = ascii (fp) : "./hrtfs/elev25/L25e110a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e110a.dat" right
[ 23, 51 ] = ascii (fp) : "./hrtfs/elev25/L25e105a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e105a.dat" right
[ 23, 52 ] = ascii (fp) : "./hrtfs/elev25/L25e100a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e100a.dat" right
[ 23, 53 ] = ascii (fp) : "./hrtfs/elev25/L25e095a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e095a.dat" right
[ 23, 54 ] = ascii (fp) : "./hrtfs/elev25/L25e090a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e090a.dat" right
[ 23, 55 ] = ascii (fp) : "./hrtfs/elev25/L25e085a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e085a.dat" right
[ 23, 56 ] = ascii (fp) : "./hrtfs/elev25/L25e080a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e080a.dat" right
[ 23, 57 ] = ascii (fp) : "./hrtfs/elev25/L25e075a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e075a.dat" right
[ 23, 58 ] = ascii (fp) : "./hrtfs/elev25/L25e070a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e070a.dat" right
[ 23, 59 ] = ascii (fp) : "./hrtfs/elev25/L25e065a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e065a.dat" right
[ 23, 60 ] = ascii (fp) : "./hrtfs/elev25/L25e060a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e060a.dat" right
[ 23, 61 ] = ascii (fp) : "./hrtfs/elev25/L25e055a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e055a.dat" right
[ 23, 62 ] = ascii (fp) : "./hrtfs/elev25/L25e050a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e050a.dat" right
[ 23, 63 ] = ascii (fp) : "./hrtfs/elev25/L25e045a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e045a.dat" right
[ 23, 64 ] = ascii (fp) : "./hrtfs/elev25/L25e040a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e040a.dat" right
[ 23, 65 ] = ascii (fp) : "./hrtfs/elev25/L25e035a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e035a.dat" right
[ 23, 66 ] = ascii (fp) : "./hrtfs/elev25/L25e030a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e030a.dat" right
[ 23, 67 ] = ascii (fp) : "./hrtfs/elev25/L25e025a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e025a.dat" right
[ 23, 68 ] = ascii (fp) : "./hrtfs/elev25/L25e020a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e020a.dat" right
[ 23, 69 ] = ascii (fp) : "./hrtfs/elev25/L25e015a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e015a.dat" right
[ 23, 70 ] = ascii (fp) : "./hrtfs/elev25/L25e010a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e010a.dat" right
[ 23, 71 ] = ascii (fp) : "./hrtfs/elev25/L25e005a.dat" left
           + ascii (fp) : "./hrtfs/elev25/R25e005a.dat" right

[ 24,  0 ] = ascii (fp) : "./hrtfs/elev30/L30e000a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e000a.dat" right
[ 24,  1 ] = ascii (fp) : "./hrtfs/elev30/L30e355a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e355a.dat" right
[ 24,  2 ] = ascii (fp) : "./hrtfs/elev30/L30e350a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e350a.dat" right
[ 24,  3 ] = ascii (fp) : "./hrtfs/elev30/L30e345a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e345a.dat" right
[ 24,  4 ] = ascii (fp) : "./hrtfs/elev30/L30e340a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e340a.dat" right
[ 24,  5 ] = ascii (fp) : "./hrtfs/elev30/L30e335a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e335a.dat" right
[ 24,  6 ] = ascii (fp) : "./hrtfs/elev30/L30e330a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e330a.dat" right
[ 24,  7 ] = ascii (fp) : "./hrtfs/elev30/L30e325a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e325a.dat" right
[ 24,  8 ] = ascii (fp) : "./hrtfs/elev30/L30e320a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e320a.dat" right
[ 24,  9 ] = ascii (fp) : "./hrtfs/elev30/L30e315a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e315a.dat" right
[ 24, 10 ] = ascii (fp) : "./hrtfs/elev30/L30e310a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e310a.dat" right
[ 24, 11 ] = ascii (fp) : "./hrtfs/elev30/L30e305a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e305a.dat" right
[ 24, 12 ] = ascii (fp) : "./hrtfs/elev30/L30e300a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e300a.dat" right
[ 24, 13 ] = ascii (fp) : "./hrtfs/elev30/L30e295a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e295a.dat" right
[ 24, 14 ] = ascii (fp) : "./hrtfs/elev30/L30e290a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e290a.dat" right
[ 24, 15 ] = ascii (fp) : "./hrtfs/elev30/L30e285a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e285a.dat" right
[ 24, 16 ] = ascii (fp) : "./hrtfs/elev30/L30e280a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e280a.dat" right
[ 24, 17 ] = ascii (fp) : "./hrtfs/elev30/L30e275a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e275a.dat" right
[ 24, 18 ] = ascii (fp) : "./hrtfs/elev30/L30e270a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e270a.dat" right
[ 24, 19 ] = ascii (fp) : "./hrtfs/elev30/L30e265a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e265a.dat" right
[ 24, 20 ] = ascii (fp) : "./hrtfs/elev30/L30e260a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e260a.dat" right
[ 24, 21 ] = ascii (fp) : "./hrtfs/elev30/L30e255a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e255a.dat" right
[ 24, 22 ] = ascii (fp) : "./hrtfs/elev30/L30e250a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e250a.dat" right
[ 24, 23 ] = ascii (fp) : "./hrtfs/elev30/L30e245a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e245a.dat" right
[ 24, 24 ] = ascii (fp) : "./hrtfs/elev30/L30e240a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e240a.dat" right
[ 24, 25 ] = ascii (fp) : "./hrtfs/elev30/L30e235a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e235a.dat" right
[ 24, 26 ] = ascii (fp) : "./hrtfs/elev30/L30e230a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e230a.dat" right
[ 24, 27 ] = ascii (fp) : "./hrtfs/elev30/L30e225a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e225a.dat" right
[ 24, 28 ] = ascii (fp) : "./hrtfs/elev30/L30e220a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e220a.dat" right
[ 24, 29 ] = ascii (fp) : "./hrtfs/elev30/L30e215a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e215a.dat" right
[ 24, 30 ] = ascii (fp) : "./hrtfs/elev30/L30e210a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e210a.dat" right
[ 24, 31 ] = ascii (fp) : "./hrtfs/elev30/L30e205a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e205a.dat" right
[ 24, 32 ] = ascii (fp) : "./hrtfs/elev30/L30e200a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e200a.dat" right
[ 24, 33 ] = ascii (fp) : "./hrtfs/elev30/L30e195a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e195a.dat" right
[ 24, 34 ] = ascii (fp) : "./hrtfs/elev30/L30e190a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e190a.dat" right
[ 24, 35 ] = ascii (fp) : "./hrtfs/elev30/L30e185a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e185a.dat" right
[ 24, 36 ] = ascii (fp) : "./hrtfs/elev30/L30e180a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e180a.dat" right
[ 24, 37 ] = ascii (fp) : "./hrtfs/elev30/L30e175a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e175a.dat" right
[ 24, 38 ] = ascii (fp) : "./hrtfs/elev30/L30e170a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e170a.dat" right
[ 24, 39 ] = ascii (fp) : "./hrtfs/elev30/L30e165a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e165a.dat" right
[ 24, 40 ] = ascii (fp) : "./hrtfs/elev30/L30e160a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e160a.dat" right
[ 24, 41 ] = ascii (fp) : "./hrtfs/elev30/L30e155a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e155a.dat" right
[ 24, 42 ] = ascii (fp) : "./hrtfs/elev30/L30e150a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e150a.dat" right
[ 24, 43 ] = ascii (fp) : "./hrtfs/elev30/L30e145a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e145a.dat" right
[ 24, 44 ] = ascii (fp) : "./hrtfs/elev30/L30e140a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e140a.dat" right
[ 24, 45 ] = ascii (fp) : "./hrtfs/elev30/L30e135a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e135a.dat" right
[ 24, 46 ] = ascii (fp) : "./hrtfs/elev30/L30e130a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e130a.dat" right
[ 24, 47 ] = ascii (fp) : "./hrtfs/elev30/L30e125a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e125a.dat" right
[ 24, 48 ] = ascii (fp) : "./hrtfs/elev30/L30e120a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e120a.dat" right
[ 24, 49 ] = ascii (fp) : "./hrtfs/elev30/L30e115a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e115a.dat" right
[ 24, 50 ] = ascii (fp) : "./hrtfs/elev30/L30e110a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e110a.dat" right
[ 24, 51 ] = ascii (fp) : "./hrtfs/elev30/L30e105a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e105a.dat" right
[ 24, 52 ] = ascii (fp) : "./hrtfs/elev30/L30e100a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e100a.dat" right
[ 24, 53 ] = ascii (fp) : "./hrtfs/elev30/L30e095a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e095a.dat" right
[ 24, 54 ] = ascii (fp) : "./hrtfs/elev30/L30e090a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e090a.dat" right
[ 24, 55 ] = ascii (fp) : "./hrtfs/elev30/L30e085a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e085a.dat" right
[ 24, 56 ] = ascii (fp) : "./hrtfs/elev30/L30e080a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e080a.dat" right
[ 24, 57 ] = ascii (fp) : "./hrtfs/elev30/L30e075a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e075a.dat" right
[ 24, 58 ] = ascii (fp) : "./hrtfs/elev30/L30e070a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e070a.dat" right
[ 24, 59 ] = ascii (fp) : "./hrtfs/elev30/L30e065a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e065a.dat" right
[ 24, 60 ] = ascii (fp) : "./hrtfs/elev30/L30e060a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e060a.dat" right
[ 24, 61 ] = ascii (fp) : "./hrtfs/elev30/L30e055a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e055a.dat" right
[ 24, 62 ] = ascii (fp) : "./hrtfs/elev30/L30e050a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e050a.dat" right
[ 24, 63 ] = ascii (fp) : "./hrtfs/elev30/L30e045a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e045a.dat" right
[ 24, 64 ] = ascii (fp) : "./hrtfs/elev30/L30e040a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e040a.dat" right
[ 24, 65 ] = ascii (fp) : "./hrtfs/elev30/L30e035a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e035a.dat" right
[ 24, 66 ] = ascii (fp) : "./hrtfs/elev30/L30e030a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e030a.dat" right
[ 24, 67 ] = ascii (fp) : "./hrtfs/elev30/L30e025a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e025a.dat" right
[ 24, 68 ] = ascii (fp) : "./hrtfs/elev30/L30e020a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e020a.dat" right
[ 24, 69 ] = ascii (fp) : "./hrtfs/elev30/L30e015a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e015a.dat" right
[ 24, 70 ] = ascii (fp) : "./hrtfs/elev30/L30e010a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e010a.dat" right
[ 24, 71 ] = ascii (fp) : "./hrtfs/elev30/L30e005a.dat" left
           + ascii (fp) : "./hrtfs/elev30/R30e005a.dat" right

[ 25,  0 ] = ascii (fp) : "./hrtfs/elev35/L35e000a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e000a.dat" right
[ 25,  1 ] = ascii (fp) : "./hrtfs/elev35/L35e355a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e355a.dat" right
[ 25,  2 ] = ascii (fp) : "./hrtfs/elev35/L35e350a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e350a.dat" right
[ 25,  3 ] = ascii (fp) : "./hrtfs/elev35/L35e345a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e345a.dat" right
[ 25,  4 ] = ascii (fp) : "./hrtfs/elev35/L35e340a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e340a.dat" right
[ 25,  5 ] = ascii (fp) : "./hrtfs/elev35/L35e335a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e335a.dat" right
[ 25,  6 ] = ascii (fp) : "./hrtfs/elev35/L35e330a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e330a.dat" right
[ 25,  7 ] = ascii (fp) : "./hrtfs/elev35/L35e325a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e325a.dat" right
[ 25,  8 ] = ascii (fp) : "./hrtfs/elev35/L35e320a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e320a.dat" right
[ 25,  9 ] = ascii (fp) : "./hrtfs/elev35/L35e315a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e315a.dat" right
[ 25, 10 ] = ascii (fp) : "./hrtfs/elev35/L35e310a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e310a.dat" right
[ 25, 11 ] = ascii (fp) : "./hrtfs/elev35/L35e305a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e305a.dat" right
[ 25, 12 ] = ascii (fp) : "./hrtfs/elev35/L35e300a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e300a.dat" right
[ 25, 13 ] = ascii (fp) : "./hrtfs/elev35/L35e295a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e295a.dat" right
[ 25, 14 ] = ascii (fp) : "./hrtfs/elev35/L35e290a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e290a.dat" right
[ 25, 15 ] = ascii (fp) : "./hrtfs/elev35/L35e285a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e285a.dat" right
[ 25, 16 ] = ascii (fp) : "./hrtfs/elev35/L35e280a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e280a.dat" right
[ 25, 17 ] = ascii (fp) : "./hrtfs/elev35/L35e275a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e275a.dat" right
[ 25, 18 ] = ascii (fp) : "./hrtfs/elev35/L35e270a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e270a.dat" right
[ 25, 19 ] = ascii (fp) : "./hrtfs/elev35/L35e265a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e265a.dat" right
[ 25, 20 ] = ascii (fp) : "./hrtfs/elev35/L35e260a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e260a.dat" right
[ 25, 21 ] = ascii (fp) : "./hrtfs/elev35/L35e255a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e255a.dat" right
[ 25, 22 ] = ascii (fp) : "./hrtfs/elev35/L35e250a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e250a.dat" right
[ 25, 23 ] = ascii (fp) : "./hrtfs/elev35/L35e245a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e245a.dat" right
[ 25, 24 ] = ascii (fp) : "./hrtfs/elev35/L35e240a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e240a.dat" right
[ 25, 25 ] = ascii (fp) : "./hrtfs/elev35/L35e235a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e235a.dat" right
[ 25, 26 ] = ascii (fp) : "./hrtfs/elev35/L35e230a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e230a.dat" right
[ 25, 27 ] = ascii (fp) : "./hrtfs/elev35/L35e225a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e225a.dat" right
[ 25, 28 ] = ascii (fp) : "./hrtfs/elev35/L35e220a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e220a.dat" right
[ 25, 29 ] = ascii (fp) : "./hrtfs/elev35/L35e215a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e215a.dat" right
[ 25, 30 ] = ascii (fp) : "./hrtfs/elev35/L35e210a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e210a.dat" right
[ 25, 31 ] = ascii (fp) : "./hrtfs/elev35/L35e205a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e205a.dat" right
[ 25, 32 ] = ascii (fp) : "./hrtfs/elev35/L35e200a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e200a.dat" right
[ 25, 33 ] = ascii (fp) : "./hrtfs/elev35/L35e195a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e195a.dat" right
[ 25, 34 ] = ascii (fp) : "./hrtfs/elev35/L35e190a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e190a.dat" right
[ 25, 35 ] = ascii (fp) : "./hrtfs/elev35/L35e185a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e185a.dat" right
[ 25, 36 ] = ascii (fp) : "./hrtfs/elev35/L35e180a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e180a.dat" right
[ 25, 37 ] = ascii (fp) : "./hrtfs/elev35/L35e175a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e175a.dat" right
[ 25, 38 ] = ascii (fp) : "./hrtfs/elev35/L35e170a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e170a.dat" right
[ 25, 39 ] = ascii (fp) : "./hrtfs/elev35/L35e165a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e165a.dat" right
[ 25, 40 ] = ascii (fp) : "./hrtfs/elev35/L35e160a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e160a.dat" right
[ 25, 41 ] = ascii (fp) : "./hrtfs/elev35/L35e155a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e155a.dat" right
[ 25, 42 ] = ascii (fp) : "./hrtfs/elev35/L35e150a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e150a.dat" right
[ 25, 43 ] = ascii (fp) : "./hrtfs/elev35/L35e145a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e145a.dat" right
[ 25, 44 ] = ascii (fp) : "./hrtfs/elev35/L35e140a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e140a.dat" right
[ 25, 45 ] = ascii (fp) : "./hrtfs/elev35/L35e135a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e135a.dat" right
[ 25, 46 ] = ascii (fp) : "./hrtfs/elev35/L35e130a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e130a.dat" right
[ 25, 47 ] = ascii (fp) : "./hrtfs/elev35/L35e125a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e125a.dat" right
[ 25, 48 ] = ascii (fp) : "./hrtfs/elev35/L35e120a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e120a.dat" right
[ 25, 49 ] = ascii (fp) : "./hrtfs/elev35/L35e115a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e115a.dat" right
[ 25, 50 ] = ascii (fp) : "./hrtfs/elev35/L35e110a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e110a.dat" right
[ 25, 51 ] = ascii (fp) : "./hrtfs/elev35/L35e105a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e105a.dat" right
[ 25, 52 ] = ascii (fp) : "./hrtfs/elev35/L35e100a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e100a.dat" right
[ 25, 53 ] = ascii (fp) : "./hrtfs/elev35/L35e095a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e095a.dat" right
[ 25, 54 ] = ascii (fp) : "./hrtfs/elev35/L35e090a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e090a.dat" right
[ 25, 55 ] = ascii (fp) : "./hrtfs/elev35/L35e085a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e085a.dat" right
[ 25, 56 ] = ascii (fp) : "./hrtfs/elev35/L35e080a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e080a.dat" right
[ 25, 57 ] = ascii (fp) : "./hrtfs/elev35/L35e075a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e075a.dat" right
[ 25, 58 ] = ascii (fp) : "./hrtfs/elev35/L35e070a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e070a.dat" right
[ 25, 59 ] = ascii (fp) : "./hrtfs/elev35/L35e065a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e065a.dat" right
[ 25, 60 ] = ascii (fp) : "./hrtfs/elev35/L35e060a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e060a.dat" right
[ 25, 61 ] = ascii (fp) : "./hrtfs/elev35/L35e055a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e055a.dat" right
[ 25, 62 ] = ascii (fp) : "./hrtfs/elev35/L35e050a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e050a.dat" right
[ 25, 63 ] = ascii (fp) : "./hrtfs/elev35/L35e045a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e045a.dat" right
[ 25, 64 ] = ascii (fp) : "./hrtfs/elev35/L35e040a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e040a.dat" right
[ 25, 65 ] = ascii (fp) : "./hrtfs/elev35/L35e035a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e035a.dat" right
[ 25, 66 ] = ascii (fp) : "./hrtfs/elev35/L35e030a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e030a.dat" right
[ 25, 67 ] = ascii (fp) : "./hrtfs/elev35/L35e025a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e025a.dat" right
[ 25, 68 ] = ascii (fp) : "./hrtfs/elev35/L35e020a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e020a.dat" right
[ 25, 69 ] = ascii (fp) : "./hrtfs/elev35/L35e015a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e015a.dat" right
[ 25, 70 ] = ascii (fp) : "./hrtfs/elev35/L35e010a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e010a.dat" right
[ 25, 71 ] = ascii (fp) : "./hrtfs/elev35/L35e005a.dat" left
           + ascii (fp) : "./hrtfs/elev35/R35e005a.dat" right

[ 26,  0 ] = ascii (fp) : "./hrtfs/elev40/L40e000a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e000a.dat" right
[ 26,  1 ] = ascii (fp) : "./hrtfs/elev40/L40e355a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e355a.dat" right
[ 26,  2 ] = ascii (fp) : "./hrtfs/elev40/L40e350a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e350a.dat" right
[ 26,  3 ] = ascii (fp) : "./hrtfs/elev40/L40e345a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e345a.dat" right
[ 26,  4 ] = ascii (fp) : "./hrtfs/elev40/L40e340a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e340a.dat" right
[ 26,  5 ] = ascii (fp) : "./hrtfs/elev40/L40e335a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e335a.dat" right
[ 26,  6 ] = ascii (fp) : "./hrtfs/elev40/L40e330a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e330a.dat" right
[ 26,  7 ] = ascii (fp) : "./hrtfs/elev40/L40e325a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e325a.dat" right
[ 26,  8 ] = ascii (fp) : "./hrtfs/elev40/L40e320a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e320a.dat" right
[ 26,  9 ] = ascii (fp) : "./hrtfs/elev40/L40e315a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e315a.dat" right
[ 26, 10 ] = ascii (fp) : "./hrtfs/elev40/L40e310a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e310a.dat" right
[ 26, 11 ] = ascii (fp) : "./hrtfs/elev40/L40e305a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e305a.dat" right
[ 26, 12 ] = ascii (fp) : "./hrtfs/elev40/L40e300a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e300a.dat" right
[ 26, 13 ] = ascii (fp) : "./hrtfs/elev40/L40e295a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e295a.dat" right
[ 26, 14 ] = ascii (fp) : "./hrtfs/elev40/L40e290a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e290a.dat" right
[ 26, 15 ] = ascii (fp) : "./hrtfs/elev40/L40e285a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e285a.dat" right
[ 26, 16 ] = ascii (fp) : "./hrtfs/elev40/L40e280a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e280a.dat" right
[ 26, 17 ] = ascii (fp) : "./hrtfs/elev40/L40e275a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e275a.dat" right
[ 26, 18 ] = ascii (fp) : "./hrtfs/elev40/L40e270a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e270a.dat" right
[ 26, 19 ] = ascii (fp) : "./hrtfs/elev40/L40e265a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e265a.dat" right
[ 26, 20 ] = ascii (fp) : "./hrtfs/elev40/L40e260a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e260a.dat" right
[ 26, 21 ] = ascii (fp) : "./hrtfs/elev40/L40e255a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e255a.dat" right
[ 26, 22 ] = ascii (fp) : "./hrtfs/elev40/L40e250a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e250a.dat" right
[ 26, 23 ] = ascii (fp) : "./hrtfs/elev40/L40e245a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e245a.dat" right
[ 26, 24 ] = ascii (fp) : "./hrtfs/elev40/L40e240a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e240a.dat" right
[ 26, 25 ] = ascii (fp) : "./hrtfs/elev40/L40e235a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e235a.dat" right
[ 26, 26 ] = ascii (fp) : "./hrtfs/elev40/L40e230a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e230a.dat" right
[ 26, 27 ] = ascii (fp) : "./hrtfs/elev40/L40e225a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e225a.dat" right
[ 26, 28 ] = ascii (fp) : "./hrtfs/elev40/L40e220a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e220a.dat" right
[ 26, 29 ] = ascii (fp) : "./hrtfs/elev40/L40e215a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e215a.dat" right
[ 26, 30 ] = ascii (fp) : "./hrtfs/elev40/L40e210a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e210a.dat" right
[ 26, 31 ] = ascii (fp) : "./hrtfs/elev40/L40e205a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e205a.dat" right
[ 26, 32 ] = ascii (fp) : "./hrtfs/elev40/L40e200a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e200a.dat" right
[ 26, 33 ] = ascii (fp) : "./hrtfs/elev40/L40e195a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e195a.dat" right
[ 26, 34 ] = ascii (fp) : "./hrtfs/elev40/L40e190a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e190a.dat" right
[ 26, 35 ] = ascii (fp) : "./hrtfs/elev40/L40e185a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e185a.dat" right
[ 26, 36 ] = ascii (fp) : "./hrtfs/elev40/L40e180a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e180a.dat" right
[ 26, 37 ] = ascii (fp) : "./hrtfs/elev40/L40e175a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e175a.dat" right
[ 26, 38 ] = ascii (fp) : "./hrtfs/elev40/L40e170a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e170a.dat" right
[ 26, 39 ] = ascii (fp) : "./hrtfs/elev40/L40e165a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e165a.dat" right
[ 26, 40 ] = ascii (fp) : "./hrtfs/elev40/L40e160a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e160a.dat" right
[ 26, 41 ] = ascii (fp) : "./hrtfs/elev40/L40e155a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e155a.dat" right
[ 26, 42 ] = ascii (fp) : "./hrtfs/elev40/L40e150a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e150a.dat" right
[ 26, 43 ] = ascii (fp) : "./hrtfs/elev40/L40e145a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e145a.dat" right
[ 26, 44 ] = ascii (fp) : "./hrtfs/elev40/L40e140a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e140a.dat" right
[ 26, 45 ] = ascii (fp) : "./hrtfs/elev40/L40e135a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e135a.dat" right
[ 26, 46 ] = ascii (fp) : "./hrtfs/elev40/L40e130a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e130a.dat" right
[ 26, 47 ] = ascii (fp) : "./hrtfs/elev40/L40e125a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e125a.dat" right
[ 26, 48 ] = ascii (fp) : "./hrtfs/elev40/L40e120a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e120a.dat" right
[ 26, 49 ] = ascii (fp) : "./hrtfs/elev40/L40e115a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e115a.dat" right
[ 26, 50 ] = ascii (fp) : "./hrtfs/elev40/L40e110a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e110a.dat" right
[ 26, 51 ] = ascii (fp) : "./hrtfs/elev40/L40e105a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e105a.dat" right
[ 26, 52 ] = ascii (fp) : "./hrtfs/elev40/L40e100a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e100a.dat" right
[ 26, 53 ] = ascii (fp) : "./hrtfs/elev40/L40e095a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e095a.dat" right
[ 26, 54 ] = ascii (fp) : "./hrtfs/elev40/L40e090a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e090a.dat" right
[ 26, 55 ] = ascii (fp) : "./hrtfs/elev40/L40e085a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e085a.dat" right
[ 26, 56 ] = ascii (fp) : "./hrtfs/elev40/L40e080a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e080a.dat" right
[ 26, 57 ] = ascii (fp) : "./hrtfs/elev40/L40e075a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e075a.dat" right
[ 26, 58 ] = ascii (fp) : "./hrtfs/elev40/L40e070a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e070a.dat" right
[ 26, 59 ] = ascii (fp) : "./hrtfs/elev40/L40e065a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e065a.dat" right
[ 26, 60 ] = ascii (fp) : "./hrtfs/elev40/L40e060a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e060a.dat" right
[ 26, 61 ] = ascii (fp) : "./hrtfs/elev40/L40e055a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e055a.dat" right
[ 26, 62 ] = ascii (fp) : "./hrtfs/elev40/L40e050a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e050a.dat" right
[ 26, 63 ] = ascii (fp) : "./hrtfs/elev40/L40e045a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e045a.dat" right
[ 26, 64 ] = ascii (fp) : "./hrtfs/elev40/L40e040a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e040a.dat" right
[ 26, 65 ] = ascii (fp) : "./hrtfs/elev40/L40e035a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e035a.dat" right
[ 26, 66 ] = ascii (fp) : "./hrtfs/elev40/L40e030a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e030a.dat" right
[ 26, 67 ] = ascii (fp) : "./hrtfs/elev40/L40e025a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e025a.dat" right
[ 26, 68 ] = ascii (fp) : "./hrtfs/elev40/L40e020a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e020a.dat" right
[ 26, 69 ] = ascii (fp) : "./hrtfs/elev40/L40e015a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e015a.dat" right
[ 26, 70 ] = ascii (fp) : "./hrtfs/elev40/L40e010a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e010a.dat" right
[ 26, 71 ] = ascii (fp) : "./hrtfs/elev40/L40e005a.dat" left
           + ascii (fp) : "./hrtfs/elev40/R40e005a.dat" right

[ 27,  0 ] = ascii (fp) : "./hrtfs/elev45/L45e000a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e000a.dat" right
[ 27,  1 ] = ascii (fp) : "./hrtfs/elev45/L45e355a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e355a.dat" right
[ 27,  2 ] = ascii (fp) : "./hrtfs/elev45/L45e350a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e350a.dat" right
[ 27,  3 ] = ascii (fp) : "./hrtfs/elev45/L45e345a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e345a.dat" right
[ 27,  4 ] = ascii (fp) : "./hrtfs/elev45/L45e340a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e340a.dat" right
[ 27,  5 ] = ascii (fp) : "./hrtfs/elev45/L45e335a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e335a.dat" right
[ 27,  6 ] = ascii (fp) : "./hrtfs/elev45/L45e330a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e330a.dat" right
[ 27,  7 ] = ascii (fp) : "./hrtfs/elev45/L45e325a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e325a.dat" right
[ 27,  8 ] = ascii (fp) : "./hrtfs/elev45/L45e320a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e320a.dat" right
[ 27,  9 ] = ascii (fp) : "./hrtfs/elev45/L45e315a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e315a.dat" right
[ 27, 10 ] = ascii (fp) : "./hrtfs/elev45/L45e310a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e310a.dat" right
[ 27, 11 ] = ascii (fp) : "./hrtfs/elev45/L45e305a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e305a.dat" right
[ 27, 12 ] = ascii (fp) : "./hrtfs/elev45/L45e300a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e300a.dat" right
[ 27, 13 ] = ascii (fp) : "./hrtfs/elev45/L45e295a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e295a.dat" right
[ 27, 14 ] = ascii (fp) : "./hrtfs/elev45/L45e290a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e290a.dat" right
[ 27, 15 ] = ascii (fp) : "./hrtfs/elev45/L45e285a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e285a.dat" right
[ 27, 16 ] = ascii (fp) : "./hrtfs/elev45/L45e280a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e280a.dat" right
[ 27, 17 ] = ascii (fp) : "./hrtfs/elev45/L45e275a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e275a.dat" right
[ 27, 18 ] = ascii (fp) : "./hrtfs/elev45/L45e270a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e270a.dat" right
[ 27, 19 ] = ascii (fp) : "./hrtfs/elev45/L45e265a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e265a.dat" right
[ 27, 20 ] = ascii (fp) : "./hrtfs/elev45/L45e260a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e260a.dat" right
[ 27, 21 ] = ascii (fp) : "./hrtfs/elev45/L45e255a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e255a.dat" right
[ 27, 22 ] = ascii (fp) : "./hrtfs/elev45/L45e250a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e250a.dat" right
[ 27, 23 ] = ascii (fp) : "./hrtfs/elev45/L45e245a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e245a.dat" right
[ 27, 24 ] = ascii (fp) : "./hrtfs/elev45/L45e240a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e240a.dat" right
[ 27, 25 ] = ascii (fp) : "./hrtfs/elev45/L45e235a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e235a.dat" right
[ 27, 26 ] = ascii (fp) : "./hrtfs/elev45/L45e230a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e230a.dat" right
[ 27, 27 ] = ascii (fp) : "./hrtfs/elev45/L45e225a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e225a.dat" right
[ 27, 28 ] = ascii (fp) : "./hrtfs/elev45/L45e220a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e220a.dat" right
[ 27, 29 ] = ascii (fp) : "./hrtfs/elev45/L45e215a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e215a.dat" right
[ 27, 30 ] = ascii (fp) : "./hrtfs/elev45/L45e210a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e210a.dat" right
[ 27, 31 ] = ascii (fp) : "./hrtfs/elev45/L45e205a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e205a.dat" right
[ 27, 32 ] = ascii (fp) : "./hrtfs/elev45/L45e200a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e200a.dat" right
[ 27, 33 ] = ascii (fp) : "./hrtfs/elev45/L45e195a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e195a.dat" right
[ 27, 34 ] = ascii (fp) : "./hrtfs/elev45/L45e190a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e190a.dat" right
[ 27, 35 ] = ascii (fp) : "./hrtfs/elev45/L45e185a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e185a.dat" right
[ 27, 36 ] = ascii (fp) : "./hrtfs/elev45/L45e180a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e180a.dat" right
[ 27, 37 ] = ascii (fp) : "./hrtfs/elev45/L45e175a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e175a.dat" right
[ 27, 38 ] = ascii (fp) : "./hrtfs/elev45/L45e170a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e170a.dat" right
[ 27, 39 ] = ascii (fp) : "./hrtfs/elev45/L45e165a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e165a.dat" right
[ 27, 40 ] = ascii (fp) : "./hrtfs/elev45/L45e160a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e160a.dat" right
[ 27, 41 ] = ascii (fp) : "./hrtfs/elev45/L45e155a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e155a.dat" right
[ 27, 42 ] = ascii (fp) : "./hrtfs/elev45/L45e150a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e150a.dat" right
[ 27, 43 ] = ascii (fp) : "./hrtfs/elev45/L45e145a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e145a.dat" right
[ 27, 44 ] = ascii (fp) : "./hrtfs/elev45/L45e140a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e140a.dat" right
[ 27, 45 ] = ascii (fp) : "./hrtfs/elev45/L45e135a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e135a.dat" right
[ 27, 46 ] = ascii (fp) : "./hrtfs/elev45/L45e130a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e130a.dat" right
[ 27, 47 ] = ascii (fp) : "./hrtfs/elev45/L45e125a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e125a.dat" right
[ 27, 48 ] = ascii (fp) : "./hrtfs/elev45/L45e120a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e120a.dat" right
[ 27, 49 ] = ascii (fp) : "./hrtfs/elev45/L45e115a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e115a.dat" right
[ 27, 50 ] = ascii (fp) : "./hrtfs/elev45/L45e110a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e110a.dat" right
[ 27, 51 ] = ascii (fp) : "./hrtfs/elev45/L45e105a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e105a.dat" right
[ 27, 52 ] = ascii (fp) : "./hrtfs/elev45/L45e100a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e100a.dat" right
[ 27, 53 ] = ascii (fp) : "./hrtfs/elev45/L45e095a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e095a.dat" right
[ 27, 54 ] = ascii (fp) : "./hrtfs/elev45/L45e090a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e090a.dat" right
[ 27, 55 ] = ascii (fp) : "./hrtfs/elev45/L45e085a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e085a.dat" right
[ 27, 56 ] = ascii (fp) : "./hrtfs/elev45/L45e080a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e080a.dat" right
[ 27, 57 ] = ascii (fp) : "./hrtfs/elev45/L45e075a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e075a.dat" right
[ 27, 58 ] = ascii (fp) : "./hrtfs/elev45/L45e070a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e070a.dat" right
[ 27, 59 ] = ascii (fp) : "./hrtfs/elev45/L45e065a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e065a.dat" right
[ 27, 60 ] = ascii (fp) : "./hrtfs/elev45/L45e060a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e060a.dat" right
[ 27, 61 ] = ascii (fp) : "./hrtfs/elev45/L45e055a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e055a.dat" right
[ 27, 62 ] = ascii (fp) : "./hrtfs/elev45/L45e050a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e050a.dat" right
[ 27, 63 ] = ascii (fp) : "./hrtfs/elev45/L45e045a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e045a.dat" right
[ 27, 64 ] = ascii (fp) : "./hrtfs/elev45/L45e040a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e040a.dat" right
[ 27, 65 ] = ascii (fp) : "./hrtfs/elev45/L45e035a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e035a.dat" right
[ 27, 66 ] = ascii (fp) : "./hrtfs/elev45/L45e030a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e030a.dat" right
[ 27, 67 ] = ascii (fp) : "./hrtfs/elev45/L45e025a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e025a.dat" right
[ 27, 68 ] = ascii (fp) : "./hrtfs/elev45/L45e020a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e020a.dat" right
[ 27, 69 ] = ascii (fp) : "./hrtfs/elev45/L45e015a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e015a.dat" right
[ 27, 70 ] = ascii (fp) : "./hrtfs/elev45/L45e010a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e010a.dat" right
[ 27, 71 ] = ascii (fp) : "./hrtfs/elev45/L45e005a.dat" left
           + ascii (fp) : "./hrtfs/elev45/R45e005a.dat" right

[ 28,  0 ] = ascii (fp) : "./hrtfs/elev50/L50e000a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e000a.dat" right
[ 28,  1 ] = ascii (fp) : "./hrtfs/elev50/L50e355a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e355a.dat" right
[ 28,  2 ] = ascii (fp) : "./hrtfs/elev50/L50e350a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e350a.dat" right
[ 28,  3 ] = ascii (fp) : "./hrtfs/elev50/L50e345a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e345a.dat" right
[ 28,  4 ] = ascii (fp) : "./hrtfs/elev50/L50e340a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e340a.dat" right
[ 28,  5 ] = ascii (fp) : "./hrtfs/elev50/L50e335a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e335a.dat" right
[ 28,  6 ] = ascii (fp) : "./hrtfs/elev50/L50e330a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e330a.dat" right
[ 28,  7 ] = ascii (fp) : "./hrtfs/elev50/L50e325a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e325a.dat" right
[ 28,  8 ] = ascii (fp) : "./hrtfs/elev50/L50e320a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e320a.dat" right
[ 28,  9 ] = ascii (fp) : "./hrtfs/elev50/L50e315a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e315a.dat" right
[ 28, 10 ] = ascii (fp) : "./hrtfs/elev50/L50e310a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e310a.dat" right
[ 28, 11 ] = ascii (fp) : "./hrtfs/elev50/L50e305a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e305a.dat" right
[ 28, 12 ] = ascii (fp) : "./hrtfs/elev50/L50e300a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e300a.dat" right
[ 28, 13 ] = ascii (fp) : "./hrtfs/elev50/L50e295a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e295a.dat" right
[ 28, 14 ] = ascii (fp) : "./hrtfs/elev50/L50e290a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e290a.dat" right
[ 28, 15 ] = ascii (fp) : "./hrtfs/elev50/L50e285a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e285a.dat" right
[ 28, 16 ] = ascii (fp) : "./hrtfs/elev50/L50e280a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e280a.dat" right
[ 28, 17 ] = ascii (fp) : "./hrtfs/elev50/L50e275a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e275a.dat" right
[ 28, 18 ] = ascii (fp) : "./hrtfs/elev50/L50e270a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e270a.dat" right
[ 28, 19 ] = ascii (fp) : "./hrtfs/elev50/L50e265a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e265a.dat" right
[ 28, 20 ] = ascii (fp) : "./hrtfs/elev50/L50e260a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e260a.dat" right
[ 28, 21 ] = ascii (fp) : "./hrtfs/elev50/L50e255a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e255a.dat" right
[ 28, 22 ] = ascii (fp) : "./hrtfs/elev50/L50e250a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e250a.dat" right
[ 28, 23 ] = ascii (fp) : "./hrtfs/elev50/L50e245a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e245a.dat" right
[ 28, 24 ] = ascii (fp) : "./hrtfs/elev50/L50e240a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e240a.dat" right
[ 28, 25 ] = ascii (fp) : "./hrtfs/elev50/L50e235a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e235a.dat" right
[ 28, 26 ] = ascii (fp) : "./hrtfs/elev50/L50e230a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e230a.dat" right
[ 28, 27 ] = ascii (fp) : "./hrtfs/elev50/L50e225a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e225a.dat" right
[ 28, 28 ] = ascii (fp) : "./hrtfs/elev50/L50e220a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e220a.dat" right
[ 28, 29 ] = ascii (fp) : "./hrtfs/elev50/L50e215a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e215a.dat" right
[ 28, 30 ] = ascii (fp) : "./hrtfs/elev50/L50e210a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e210a.dat" right
[ 28, 31 ] = ascii (fp) : "./hrtfs/elev50/L50e205a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e205a.dat" right
[ 28, 32 ] = ascii (fp) : "./hrtfs/elev50/L50e200a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e200a.dat" right
[ 28, 33 ] = ascii (fp) : "./hrtfs/elev50/L50e195a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e195a.dat" right
[ 28, 34 ] = ascii (fp) : "./hrtfs/elev50/L50e190a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e190a.dat" right
[ 28, 35 ] = ascii (fp) : "./hrtfs/elev50/L50e185a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e185a.dat" right
[ 28, 36 ] = ascii (fp) : "./hrtfs/elev50/L50e180a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e180a.dat" right
[ 28, 37 ] = ascii (fp) : "./hrtfs/elev50/L50e175a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e175a.dat" right
[ 28, 38 ] = ascii (fp) : "./hrtfs/elev50/L50e170a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e170a.dat" right
[ 28, 39 ] = ascii (fp) : "./hrtfs/elev50/L50e165a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e165a.dat" right
[ 28, 40 ] = ascii (fp) : "./hrtfs/elev50/L50e160a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e160a.dat" right
[ 28, 41 ] = ascii (fp) : "./hrtfs/elev50/L50e155a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e155a.dat" right
[ 28, 42 ] = ascii (fp) : "./hrtfs/elev50/L50e150a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e150a.dat" right
[ 28, 43 ] = ascii (fp) : "./hrtfs/elev50/L50e145a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e145a.dat" right
[ 28, 44 ] = ascii (fp) : "./hrtfs/elev50/L50e140a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e140a.dat" right
[ 28, 45 ] = ascii (fp) : "./hrtfs/elev50/L50e135a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e135a.dat" right
[ 28, 46 ] = ascii (fp) : "./hrtfs/elev50/L50e130a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e130a.dat" right
[ 28, 47 ] = ascii (fp) : "./hrtfs/elev50/L50e125a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e125a.dat" right
[ 28, 48 ] = ascii (fp) : "./hrtfs/elev50/L50e120a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e120a.dat" right
[ 28, 49 ] = ascii (fp) : "./hrtfs/elev50/L50e115a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e115a.dat" right
[ 28, 50 ] = ascii (fp) : "./hrtfs/elev50/L50e110a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e110a.dat" right
[ 28, 51 ] = ascii (fp) : "./hrtfs/elev50/L50e105a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e105a.dat" right
[ 28, 52 ] = ascii (fp) : "./hrtfs/elev50/L50e100a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e100a.dat" right
[ 28, 53 ] = ascii (fp) : "./hrtfs/elev50/L50e095a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e095a.dat" right
[ 28, 54 ] = ascii (fp) : "./hrtfs/elev50/L50e090a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e090a.dat" right
[ 28, 55 ] = ascii (fp) : "./hrtfs/elev50/L50e085a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e085a.dat" right
[ 28, 56 ] = ascii (fp) : "./hrtfs/elev50/L50e080a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e080a.dat" right
[ 28, 57 ] = ascii (fp) : "./hrtfs/elev50/L50e075a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e075a.dat" right
[ 28, 58 ] = ascii (fp) : "./hrtfs/elev50/L50e070a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e070a.dat" right
[ 28, 59 ] = ascii (fp) : "./hrtfs/elev50/L50e065a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e065a.dat" right
[ 28, 60 ] = ascii (fp) : "./hrtfs/elev50/L50e060a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e060a.dat" right
[ 28, 61 ] = ascii (fp) : "./hrtfs/elev50/L50e055a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e055a.dat" right
[ 28, 62 ] = ascii (fp) : "./hrtfs/elev50/L50e050a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e050a.dat" right
[ 28, 63 ] = ascii (fp) : "./hrtfs/elev50/L50e045a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e045a.dat" right
[ 28, 64 ] = ascii (fp) : "./hrtfs/elev50/L50e040a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e040a.dat" right
[ 28, 65 ] = ascii (fp) : "./hrtfs/elev50/L50e035a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e035a.dat" right
[ 28, 66 ] = ascii (fp) : "./hrtfs/elev50/L50e030a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e030a.dat" right
[ 28, 67 ] = ascii (fp) : "./hrtfs/elev50/L50e025a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e025a.dat" right
[ 28, 68 ] = ascii (fp) : "./hrtfs/elev50/L50e020a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e020a.dat" right
[ 28, 69 ] = ascii (fp) : "./hrtfs/elev50/L50e015a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e015a.dat" right
[ 28, 70 ] = ascii (fp) : "./hrtfs/elev50/L50e010a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e010a.dat" right
[ 28, 71 ] = ascii (fp) : "./hrtfs/elev50/L50e005a.dat" left
           + ascii (fp) : "./hrtfs/elev50/R50e005a.dat" right

[ 29,  0 ] = ascii (fp) : "./hrtfs/elev55/L55e000a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e000a.dat" right
[ 29,  1 ] = ascii (fp) : "./hrtfs/elev55/L55e355a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e355a.dat" right
[ 29,  2 ] = ascii (fp) : "./hrtfs/elev55/L55e350a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e350a.dat" right
[ 29,  3 ] = ascii (fp) : "./hrtfs/elev55/L55e345a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e345a.dat" right
[ 29,  4 ] = ascii (fp) : "./hrtfs/elev55/L55e340a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e340a.dat" right
[ 29,  5 ] = ascii (fp) : "./hrtfs/elev55/L55e335a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e335a.dat" right
[ 29,  6 ] = ascii (fp) : "./hrtfs/elev55/L55e330a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e330a.dat" right
[ 29,  7 ] = ascii (fp) : "./hrtfs/elev55/L55e325a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e325a.dat" right
[ 29,  8 ] = ascii (fp) : "./hrtfs/elev55/L55e320a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e320a.dat" right
[ 29,  9 ] = ascii (fp) : "./hrtfs/elev55/L55e315a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e315a.dat" right
[ 29, 10 ] = ascii (fp) : "./hrtfs/elev55/L55e310a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e310a.dat" right
[ 29, 11 ] = ascii (fp) : "./hrtfs/elev55/L55e305a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e305a.dat" right
[ 29, 12 ] = ascii (fp) : "./hrtfs/elev55/L55e300a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e300a.dat" right
[ 29, 13 ] = ascii (fp) : "./hrtfs/elev55/L55e295a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e295a.dat" right
[ 29, 14 ] = ascii (fp) : "./hrtfs/elev55/L55e290a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e290a.dat" right
[ 29, 15 ] = ascii (fp) : "./hrtfs/elev55/L55e285a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e285a.dat" right
[ 29, 16 ] = ascii (fp) : "./hrtfs/elev55/L55e280a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e280a.dat" right
[ 29, 17 ] = ascii (fp) : "./hrtfs/elev55/L55e275a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e275a.dat" right
[ 29, 18 ] = ascii (fp) : "./hrtfs/elev55/L55e270a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e270a.dat" right
[ 29, 19 ] = ascii (fp) : "./hrtfs/elev55/L55e265a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e265a.dat" right
[ 29, 20 ] = ascii (fp) : "./hrtfs/elev55/L55e260a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e260a.dat" right
[ 29, 21 ] = ascii (fp) : "./hrtfs/elev55/L55e255a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e255a.dat" right
[ 29, 22 ] = ascii (fp) : "./hrtfs/elev55/L55e250a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e250a.dat" right
[ 29, 23 ] = ascii (fp) : "./hrtfs/elev55/L55e245a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e245a.dat" right
[ 29, 24 ] = ascii (fp) : "./hrtfs/elev55/L55e240a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e240a.dat" right
[ 29, 25 ] = ascii (fp) : "./hrtfs/elev55/L55e235a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e235a.dat" right
[ 29, 26 ] = ascii (fp) : "./hrtfs/elev55/L55e230a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e230a.dat" right
[ 29, 27 ] = ascii (fp) : "./hrtfs/elev55/L55e225a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e225a.dat" right
[ 29, 28 ] = ascii (fp) : "./hrtfs/elev55/L55e220a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e220a.dat" right
[ 29, 29 ] = ascii (fp) : "./hrtfs/elev55/L55e215a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e215a.dat" right
[ 29, 30 ] = ascii (fp) : "./hrtfs/elev55/L55e210a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e210a.dat" right
[ 29, 31 ] = ascii (fp) : "./hrtfs/elev55/L55e205a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e205a.dat" right
[ 29, 32 ] = ascii (fp) : "./hrtfs/elev55/L55e200a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e200a.dat" right
[ 29, 33 ] = ascii (fp) : "./hrtfs/elev55/L55e195a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e195a.dat" right
[ 29, 34 ] = ascii (fp) : "./hrtfs/elev55/L55e190a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e190a.dat" right
[ 29, 35 ] = ascii (fp) : "./hrtfs/elev55/L55e185a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e185a.dat" right
[ 29, 36 ] = ascii (fp) : "./hrtfs/elev55/L55e180a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e180a.dat" right
[ 29, 37 ] = ascii (fp) : "./hrtfs/elev55/L55e175a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e175a.dat" right
[ 29, 38 ] = ascii (fp) : "./hrtfs/elev55/L55e170a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e170a.dat" right
[ 29, 39 ] = ascii (fp) : "./hrtfs/elev55/L55e165a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e165a.dat" right
[ 29, 40 ] = ascii (fp) : "./hrtfs/elev55/L55e160a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e160a.dat" right
[ 29, 41 ] = ascii (fp) : "./hrtfs/elev55/L55e155a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e155a.dat" right
[ 29, 42 ] = ascii (fp) : "./hrtfs/elev55/L55e150a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e150a.dat" right
[ 29, 43 ] = ascii (fp) : "./hrtfs/elev55/L55e145a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e145a.dat" right
[ 29, 44 ] = ascii (fp) : "./hrtfs/elev55/L55e140a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e140a.dat" right
[ 29, 45 ] = ascii (fp) : "./hrtfs/elev55/L55e135a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e135a.dat" right
[ 29, 46 ] = ascii (fp) : "./hrtfs/elev55/L55e130a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e130a.dat" right
[ 29, 47 ] = ascii (fp) : "./hrtfs/elev55/L55e125a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e125a.dat" right
[ 29, 48 ] = ascii (fp) : "./hrtfs/elev55/L55e120a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e120a.dat" right
[ 29, 49 ] = ascii (fp) : "./hrtfs/elev55/L55e115a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e115a.dat" right
[ 29, 50 ] = ascii (fp) : "./hrtfs/elev55/L55e110a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e110a.dat" right
[ 29, 51 ] = ascii (fp) : "./hrtfs/elev55/L55e105a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e105a.dat" right
[ 29, 52 ] = ascii (fp) : "./hrtfs/elev55/L55e100a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e100a.dat" right
[ 29, 53 ] = ascii (fp) : "./hrtfs/elev55/L55e095a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e095a.dat" right
[ 29, 54 ] = ascii (fp) : "./hrtfs/elev55/L55e090a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e090a.dat" right
[ 29, 55 ] = ascii (fp) : "./hrtfs/elev55/L55e085a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e085a.dat" right
[ 29, 56 ] = ascii (fp) : "./hrtfs/elev55/L55e080a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e080a.dat" right
[ 29, 57 ] = ascii (fp) : "./hrtfs/elev55/L55e075a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e075a.dat" right
[ 29, 58 ] = ascii (fp) : "./hrtfs/elev55/L55e070a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e070a.dat" right
[ 29, 59 ] = ascii (fp) : "./hrtfs/elev55/L55e065a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e065a.dat" right
[ 29, 60 ] = ascii (fp) : "./hrtfs/elev55/L55e060a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e060a.dat" right
[ 29, 61 ] = ascii (fp) : "./hrtfs/elev55/L55e055a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e055a.dat" right
[ 29, 62 ] = ascii (fp) : "./hrtfs/elev55/L55e050a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e050a.dat" right
[ 29, 63 ] = ascii (fp) : "./hrtfs/elev55/L55e045a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e045a.dat" right
[ 29, 64 ] = ascii (fp) : "./hrtfs/elev55/L55e040a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e040a.dat" right
[ 29, 65 ] = ascii (fp) : "./hrtfs/elev55/L55e035a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e035a.dat" right
[ 29, 66 ] = ascii (fp) : "./hrtfs/elev55/L55e030a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e030a.dat" right
[ 29, 67 ] = ascii (fp) : "./hrtfs/elev55/L55e025a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e025a.dat" right
[ 29, 68 ] = ascii (fp) : "./hrtfs/elev55/L55e020a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e020a.dat" right
[ 29, 69 ] = ascii (fp) : "./hrtfs/elev55/L55e015a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e015a.dat" right
[ 29, 70 ] = ascii (fp) : "./hrtfs/elev55/L55e010a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e010a.dat" right
[ 29, 71 ] = ascii (fp) : "./hrtfs/elev55/L55e005a.dat" left
           + ascii (fp) : "./hrtfs/elev55/R55e005a.dat" right

[ 30,  0 ] = ascii (fp) : "./hrtfs/elev60/L60e000a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e000a.dat" right
[ 30,  1 ] = ascii (fp) : "./hrtfs/elev60/L60e355a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e355a.dat" right
[ 30,  2 ] = ascii (fp) : "./hrtfs/elev60/L60e350a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e350a.dat" right
[ 30,  3 ] = ascii (fp) : "./hrtfs/elev60/L60e345a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e345a.dat" right
[ 30,  4 ] = ascii (fp) : "./hrtfs/elev60/L60e340a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e340a.dat" right
[ 30,  5 ] = ascii (fp) : "./hrtfs/elev60/L60e335a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e335a.dat" right
[ 30,  6 ] = ascii (fp) : "./hrtfs/elev60/L60e330a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e330a.dat" right
[ 30,  7 ] = ascii (fp) : "./hrtfs/elev60/L60e325a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e325a.dat" right
[ 30,  8 ] = ascii (fp) : "./hrtfs/elev60/L60e320a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e320a.dat" right
[ 30,  9 ] = ascii (fp) : "./hrtfs/elev60/L60e315a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e315a.dat" right
[ 30, 10 ] = ascii (fp) : "./hrtfs/elev60/L60e310a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e310a.dat" right
[ 30, 11 ] = ascii (fp) : "./hrtfs/elev60/L60e305a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e305a.dat" right
[ 30, 12 ] = ascii (fp) : "./hrtfs/elev60/L60e300a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e300a.dat" right
[ 30, 13 ] = ascii (fp) : "./hrtfs/elev60/L60e295a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e295a.dat" right
[ 30, 14 ] = ascii (fp) : "./hrtfs/elev60/L60e290a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e290a.dat" right
[ 30, 15 ] = ascii (fp) : "./hrtfs/elev60/L60e285a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e285a.dat" right
[ 30, 16 ] = ascii (fp) : "./hrtfs/elev60/L60e280a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e280a.dat" right
[ 30, 17 ] = ascii (fp) : "./hrtfs/elev60/L60e275a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e275a.dat" right
[ 30, 18 ] = ascii (fp) : "./hrtfs/elev60/L60e270a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e270a.dat" right
[ 30, 19 ] = ascii (fp) : "./hrtfs/elev60/L60e265a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e265a.dat" right
[ 30, 20 ] = ascii (fp) : "./hrtfs/elev60/L60e260a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e260a.dat" right
[ 30, 21 ] = ascii (fp) : "./hrtfs/elev60/L60e255a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e255a.dat" right
[ 30, 22 ] = ascii (fp) : "./hrtfs/elev60/L60e250a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e250a.dat" right
[ 30, 23 ] = ascii (fp) : "./hrtfs/elev60/L60e245a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e245a.dat" right
[ 30, 24 ] = ascii (fp) : "./hrtfs/elev60/L60e240a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e240a.dat" right
[ 30, 25 ] = ascii (fp) : "./hrtfs/elev60/L60e235a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e235a.dat" right
[ 30, 26 ] = ascii (fp) : "./hrtfs/elev60/L60e230a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e230a.dat" right
[ 30, 27 ] = ascii (fp) : "./hrtfs/elev60/L60e225a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e225a.dat" right
[ 30, 28 ] = ascii (fp) : "./hrtfs/elev60/L60e220a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e220a.dat" right
[ 30, 29 ] = ascii (fp) : "./hrtfs/elev60/L60e215a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e215a.dat" right
[ 30, 30 ] = ascii (fp) : "./hrtfs/elev60/L60e210a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e210a.dat" right
[ 30, 31 ] = ascii (fp) : "./hrtfs/elev60/L60e205a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e205a.dat" right
[ 30, 32 ] = ascii (fp) : "./hrtfs/elev60/L60e200a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e200a.dat" right
[ 30, 33 ] = ascii (fp) : "./hrtfs/elev60/L60e195a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e195a.dat" right
[ 30, 34 ] = ascii (fp) : "./hrtfs/elev60/L60e190a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e190a.dat" right
[ 30, 35 ] = ascii (fp) : "./hrtfs/elev60/L60e185a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e185a.dat" right
[ 30, 36 ] = ascii (fp) : "./hrtfs/elev60/L60e180a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e180a.dat" right
[ 30, 37 ] = ascii (fp) : "./hrtfs/elev60/L60e175a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e175a.dat" right
[ 30, 38 ] = ascii (fp) : "./hrtfs/elev60/L60e170a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e170a.dat" right
[ 30, 39 ] = ascii (fp) : "./hrtfs/elev60/L60e165a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e165a.dat" right
[ 30, 40 ] = ascii (fp) : "./hrtfs/elev60/L60e160a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e160a.dat" right
[ 30, 41 ] = ascii (fp) : "./hrtfs/elev60/L60e155a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e155a.dat" right
[ 30, 42 ] = ascii (fp) : "./hrtfs/elev60/L60e150a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e150a.dat" right
[ 30, 43 ] = ascii (fp) : "./hrtfs/elev60/L60e145a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e145a.dat" right
[ 30, 44 ] = ascii (fp) : "./hrtfs/elev60/L60e140a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e140a.dat" right
[ 30, 45 ] = ascii (fp) : "./hrtfs/elev60/L60e135a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e135a.dat" right
[ 30, 46 ] = ascii (fp) : "./hrtfs/elev60/L60e130a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e130a.dat" right
[ 30, 47 ] = ascii (fp) : "./hrtfs/elev60/L60e125a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e125a.dat" right
[ 30, 48 ] = ascii (fp) : "./hrtfs/elev60/L60e120a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e120a.dat" right
[ 30, 49 ] = ascii (fp) : "./hrtfs/elev60/L60e115a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e115a.dat" right
[ 30, 50 ] = ascii (fp) : "./hrtfs/elev60/L60e110a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e110a.dat" right
[ 30, 51 ] = ascii (fp) : "./hrtfs/elev60/L60e105a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e105a.dat" right
[ 30, 52 ] = ascii (fp) : "./hrtfs/elev60/L60e100a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e100a.dat" right
[ 30, 53 ] = ascii (fp) : "./hrtfs/elev60/L60e095a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e095a.dat" right
[ 30, 54 ] = ascii (fp) : "./hrtfs/elev60/L60e090a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e090a.dat" right
[ 30, 55 ] = ascii (fp) : "./hrtfs/elev60/L60e085a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e085a.dat" right
[ 30, 56 ] = ascii (fp) : "./hrtfs/elev60/L60e080a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e080a.dat" right
[ 30, 57 ] = ascii (fp) : "./hrtfs/elev60/L60e075a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e075a.dat" right
[ 30, 58 ] = ascii (fp) : "./hrtfs/elev60/L60e070a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e070a.dat" right
[ 30, 59 ] = ascii (fp) : "./hrtfs/elev60/L60e065a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e065a.dat" right
[ 30, 60 ] = ascii (fp) : "./hrtfs/elev60/L60e060a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e060a.dat" right
[ 30, 61 ] = ascii (fp) : "./hrtfs/elev60/L60e055a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e055a.dat" right
[ 30, 62 ] = ascii (fp) : "./hrtfs/elev60/L60e050a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e050a.dat" right
[ 30, 63 ] = ascii (fp) : "./hrtfs/elev60/L60e045a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e045a.dat" right
[ 30, 64 ] = ascii (fp) : "./hrtfs/elev60/L60e040a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e040a.dat" right
[ 30, 65 ] = ascii (fp) : "./hrtfs/elev60/L60e035a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e035a.dat" right
[ 30, 66 ] = ascii (fp) : "./hrtfs/elev60/L60e030a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e030a.dat" right
[ 30, 67 ] = ascii (fp) : "./hrtfs/elev60/L60e025a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e025a.dat" right
[ 30, 68 ] = ascii (fp) : "./hrtfs/elev60/L60e020a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e020a.dat" right
[ 30, 69 ] = ascii (fp) : "./hrtfs/elev60/L60e015a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e015a.dat" right
[ 30, 70 ] = ascii (fp) : "./hrtfs/elev60/L60e010a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e010a.dat" right
[ 30, 71 ] = ascii (fp) : "./hrtfs/elev60/L60e005a.dat" left
           + ascii (fp) : "./hrtfs/elev60/R60e005a.dat" right

[ 31,  0 ] = ascii (fp) : "./hrtfs/elev65/L65e000a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e000a.dat" right
[ 31,  1 ] = ascii (fp) : "./hrtfs/elev65/L65e355a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e355a.dat" right
[ 31,  2 ] = ascii (fp) : "./hrtfs/elev65/L65e350a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e350a.dat" right
[ 31,  3 ] = ascii (fp) : "./hrtfs/elev65/L65e345a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e345a.dat" right
[ 31,  4 ] = ascii (fp) : "./hrtfs/elev65/L65e340a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e340a.dat" right
[ 31,  5 ] = ascii (fp) : "./hrtfs/elev65/L65e335a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e335a.dat" right
[ 31,  6 ] = ascii (fp) : "./hrtfs/elev65/L65e330a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e330a.dat" right
[ 31,  7 ] = ascii (fp) : "./hrtfs/elev65/L65e325a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e325a.dat" right
[ 31,  8 ] = ascii (fp) : "./hrtfs/elev65/L65e320a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e320a.dat" right
[ 31,  9 ] = ascii (fp) : "./hrtfs/elev65/L65e315a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e315a.dat" right
[ 31, 10 ] = ascii (fp) : "./hrtfs/elev65/L65e310a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e310a.dat" right
[ 31, 11 ] = ascii (fp) : "./hrtfs/elev65/L65e305a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e305a.dat" right
[ 31, 12 ] = ascii (fp) : "./hrtfs/elev65/L65e300a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e300a.dat" right
[ 31, 13 ] = ascii (fp) : "./hrtfs/elev65/L65e295a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e295a.dat" right
[ 31, 14 ] = ascii (fp) : "./hrtfs/elev65/L65e290a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e290a.dat" right
[ 31, 15 ] = ascii (fp) : "./hrtfs/elev65/L65e285a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e285a.dat" right
[ 31, 16 ] = ascii (fp) : "./hrtfs/elev65/L65e280a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e280a.dat" right
[ 31, 17 ] = ascii (fp) : "./hrtfs/elev65/L65e275a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e275a.dat" right
[ 31, 18 ] = ascii (fp) : "./hrtfs/elev65/L65e270a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e270a.dat" right
[ 31, 19 ] = ascii (fp) : "./hrtfs/elev65/L65e265a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e265a.dat" right
[ 31, 20 ] = ascii (fp) : "./hrtfs/elev65/L65e260a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e260a.dat" right
[ 31, 21 ] = ascii (fp) : "./hrtfs/elev65/L65e255a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e255a.dat" right
[ 31, 22 ] = ascii (fp) : "./hrtfs/elev65/L65e250a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e250a.dat" right
[ 31, 23 ] = ascii (fp) : "./hrtfs/elev65/L65e245a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e245a.dat" right
[ 31, 24 ] = ascii (fp) : "./hrtfs/elev65/L65e240a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e240a.dat" right
[ 31, 25 ] = ascii (fp) : "./hrtfs/elev65/L65e235a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e235a.dat" right
[ 31, 26 ] = ascii (fp) : "./hrtfs/elev65/L65e230a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e230a.dat" right
[ 31, 27 ] = ascii (fp) : "./hrtfs/elev65/L65e225a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e225a.dat" right
[ 31, 28 ] = ascii (fp) : "./hrtfs/elev65/L65e220a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e220a.dat" right
[ 31, 29 ] = ascii (fp) : "./hrtfs/elev65/L65e215a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e215a.dat" right
[ 31, 30 ] = ascii (fp) : "./hrtfs/elev65/L65e210a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e210a.dat" right
[ 31, 31 ] = ascii (fp) : "./hrtfs/elev65/L65e205a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e205a.dat" right
[ 31, 32 ] = ascii (fp) : "./hrtfs/elev65/L65e200a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e200a.dat" right
[ 31, 33 ] = ascii (fp) : "./hrtfs/elev65/L65e195a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e195a.dat" right
[ 31, 34 ] = ascii (fp) : "./hrtfs/elev65/L65e190a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e190a.dat" right
[ 31, 35 ] = ascii (fp) : "./hrtfs/elev65/L65e185a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e185a.dat" right
[ 31, 36 ] = ascii (fp) : "./hrtfs/elev65/L65e180a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e180a.dat" right
[ 31, 37 ] = ascii (fp) : "./hrtfs/elev65/L65e175a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e175a.dat" right
[ 31, 38 ] = ascii (fp) : "./hrtfs/elev65/L65e170a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e170a.dat" right
[ 31, 39 ] = ascii (fp) : "./hrtfs/elev65/L65e165a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e165a.dat" right
[ 31, 40 ] = ascii (fp) : "./hrtfs/elev65/L65e160a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e160a.dat" right
[ 31, 41 ] = ascii (fp) : "./hrtfs/elev65/L65e155a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e155a.dat" right
[ 31, 42 ] = ascii (fp) : "./hrtfs/elev65/L65e150a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e150a.dat" right
[ 31, 43 ] = ascii (fp) : "./hrtfs/elev65/L65e145a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e145a.dat" right
[ 31, 44 ] = ascii (fp) : "./hrtfs/elev65/L65e140a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e140a.dat" right
[ 31, 45 ] = ascii (fp) : "./hrtfs/elev65/L65e135a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e135a.dat" right
[ 31, 46 ] = ascii (fp) : "./hrtfs/elev65/L65e130a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e130a.dat" right
[ 31, 47 ] = ascii (fp) : "./hrtfs/elev65/L65e125a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e125a.dat" right
[ 31, 48 ] = ascii (fp) : "./hrtfs/elev65/L65e120a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e120a.dat" right
[ 31, 49 ] = ascii (fp) : "./hrtfs/elev65/L65e115a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e115a.dat" right
[ 31, 50 ] = ascii (fp) : "./hrtfs/elev65/L65e110a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e110a.dat" right
[ 31, 51 ] = ascii (fp) : "./hrtfs/elev65/L65e105a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e105a.dat" right
[ 31, 52 ] = ascii (fp) : "./hrtfs/elev65/L65e100a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e100a.dat" right
[ 31, 53 ] = ascii (fp) : "./hrtfs/elev65/L65e095a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e095a.dat" right
[ 31, 54 ] = ascii (fp) : "./hrtfs/elev65/L65e090a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e090a.dat" right
[ 31, 55 ] = ascii (fp) : "./hrtfs/elev65/L65e085a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e085a.dat" right
[ 31, 56 ] = ascii (fp) : "./hrtfs/elev65/L65e080a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e080a.dat" right
[ 31, 57 ] = ascii (fp) : "./hrtfs/elev65/L65e075a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e075a.dat" right
[ 31, 58 ] = ascii (fp) : "./hrtfs/elev65/L65e070a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e070a.dat" right
[ 31, 59 ] = ascii (fp) : "./hrtfs/elev65/L65e065a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e065a.dat" right
[ 31, 60 ] = ascii (fp) : "./hrtfs/elev65/L65e060a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e060a.dat" right
[ 31, 61 ] = ascii (fp) : "./hrtfs/elev65/L65e055a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e055a.dat" right
[ 31, 62 ] = ascii (fp) : "./hrtfs/elev65/L65e050a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e050a.dat" right
[ 31, 63 ] = ascii (fp) : "./hrtfs/elev65/L65e045a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e045a.dat" right
[ 31, 64 ] = ascii (fp) : "./hrtfs/elev65/L65e040a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e040a.dat" right
[ 31, 65 ] = ascii (fp) : "./hrtfs/elev65/L65e035a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e035a.dat" right
[ 31, 66 ] = ascii (fp) : "./hrtfs/elev65/L65e030a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e030a.dat" right
[ 31, 67 ] = ascii (fp) : "./hrtfs/elev65/L65e025a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e025a.dat" right
[ 31, 68 ] = ascii (fp) : "./hrtfs/elev65/L65e020a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e020a.dat" right
[ 31, 69 ] = ascii (fp) : "./hrtfs/elev65/L65e015a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e015a.dat" right
[ 31, 70 ] = ascii (fp) : "./hrtfs/elev65/L65e010a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e010a.dat" right
[ 31, 71 ] = ascii (fp) : "./hrtfs/elev65/L65e005a.dat" left
           + ascii (fp) : "./hrtfs/elev65/R65e005a.dat" right

[ 32,  0 ] = ascii (fp) : "./hrtfs/elev70/L70e000a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e000a.dat" right
[ 32,  1 ] = ascii (fp) : "./hrtfs/elev70/L70e355a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e355a.dat" right
[ 32,  2 ] = ascii (fp) : "./hrtfs/elev70/L70e350a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e350a.dat" right
[ 32,  3 ] = ascii (fp) : "./hrtfs/elev70/L70e345a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e345a.dat" right
[ 32,  4 ] = ascii (fp) : "./hrtfs/elev70/L70e340a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e340a.dat" right
[ 32,  5 ] = ascii (fp) : "./hrtfs/elev70/L70e335a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e335a.dat" right
[ 32,  6 ] = ascii (fp) : "./hrtfs/elev70/L70e330a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e330a.dat" right
[ 32,  7 ] = ascii (fp) : "./hrtfs/elev70/L70e325a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e325a.dat" right
[ 32,  8 ] = ascii (fp) : "./hrtfs/elev70/L70e320a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e320a.dat" right
[ 32,  9 ] = ascii (fp) : "./hrtfs/elev70/L70e315a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e315a.dat" right
[ 32, 10 ] = ascii (fp) : "./hrtfs/elev70/L70e310a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e310a.dat" right
[ 32, 11 ] = ascii (fp) : "./hrtfs/elev70/L70e305a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e305a.dat" right
[ 32, 12 ] = ascii (fp) : "./hrtfs/elev70/L70e300a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e300a.dat" right
[ 32, 13 ] = ascii (fp) : "./hrtfs/elev70/L70e295a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e295a.dat" right
[ 32, 14 ] = ascii (fp) : "./hrtfs/elev70/L70e290a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e290a.dat" right
[ 32, 15 ] = ascii (fp) : "./hrtfs/elev70/L70e285a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e285a.dat" right
[ 32, 16 ] = ascii (fp) : "./hrtfs/elev70/L70e280a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e280a.dat" right
[ 32, 17 ] = ascii (fp) : "./hrtfs/elev70/L70e275a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e275a.dat" right
[ 32, 18 ] = ascii (fp) : "./hrtfs/elev70/L70e270a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e270a.dat" right
[ 32, 19 ] = ascii (fp) : "./hrtfs/elev70/L70e265a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e265a.dat" right
[ 32, 20 ] = ascii (fp) : "./hrtfs/elev70/L70e260a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e260a.dat" right
[ 32, 21 ] = ascii (fp) : "./hrtfs/elev70/L70e255a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e255a.dat" right
[ 32, 22 ] = ascii (fp) : "./hrtfs/elev70/L70e250a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e250a.dat" right
[ 32, 23 ] = ascii (fp) : "./hrtfs/elev70/L70e245a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e245a.dat" right
[ 32, 24 ] = ascii (fp) : "./hrtfs/elev70/L70e240a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e240a.dat" right
[ 32, 25 ] = ascii (fp) : "./hrtfs/elev70/L70e235a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e235a.dat" right
[ 32, 26 ] = ascii (fp) : "./hrtfs/elev70/L70e230a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e230a.dat" right
[ 32, 27 ] = ascii (fp) : "./hrtfs/elev70/L70e225a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e225a.dat" right
[ 32, 28 ] = ascii (fp) : "./hrtfs/elev70/L70e220a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e220a.dat" right
[ 32, 29 ] = ascii (fp) : "./hrtfs/elev70/L70e215a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e215a.dat" right
[ 32, 30 ] = ascii (fp) : "./hrtfs/elev70/L70e210a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e210a.dat" right
[ 32, 31 ] = ascii (fp) : "./hrtfs/elev70/L70e205a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e205a.dat" right
[ 32, 32 ] = ascii (fp) : "./hrtfs/elev70/L70e200a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e200a.dat" right
[ 32, 33 ] = ascii (fp) : "./hrtfs/elev70/L70e195a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e195a.dat" right
[ 32, 34 ] = ascii (fp) : "./hrtfs/elev70/L70e190a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e190a.dat" right
[ 32, 35 ] = ascii (fp) : "./hrtfs/elev70/L70e185a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e185a.dat" right
[ 32, 36 ] = ascii (fp) : "./hrtfs/elev70/L70e180a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e180a.dat" right
[ 32, 37 ] = ascii (fp) : "./hrtfs/elev70/L70e175a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e175a.dat" right
[ 32, 38 ] = ascii (fp) : "./hrtfs/elev70/L70e170a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e170a.dat" right
[ 32, 39 ] = ascii (fp) : "./hrtfs/elev70/L70e165a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e165a.dat" right
[ 32, 40 ] = ascii (fp) : "./hrtfs/elev70/L70e160a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e160a.dat" right
[ 32, 41 ] = ascii (fp) : "./hrtfs/elev70/L70e155a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e155a.dat" right
[ 32, 42 ] = ascii (fp) : "./hrtfs/elev70/L70e150a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e150a.dat" right
[ 32, 43 ] = ascii (fp) : "./hrtfs/elev70/L70e145a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e145a.dat" right
[ 32, 44 ] = ascii (fp) : "./hrtfs/elev70/L70e140a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e140a.dat" right
[ 32, 45 ] = ascii (fp) : "./hrtfs/elev70/L70e135a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e135a.dat" right
[ 32, 46 ] = ascii (fp) : "./hrtfs/elev70/L70e130a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e130a.dat" right
[ 32, 47 ] = ascii (fp) : "./hrtfs/elev70/L70e125a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e125a.dat" right
[ 32, 48 ] = ascii (fp) : "./hrtfs/elev70/L70e120a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e120a.dat" right
[ 32, 49 ] = ascii (fp) : "./hrtfs/elev70/L70e115a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e115a.dat" right
[ 32, 50 ] = ascii (fp) : "./hrtfs/elev70/L70e110a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e110a.dat" right
[ 32, 51 ] = ascii (fp) : "./hrtfs/elev70/L70e105a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e105a.dat" right
[ 32, 52 ] = ascii (fp) : "./hrtfs/elev70/L70e100a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e100a.dat" right
[ 32, 53 ] = ascii (fp) : "./hrtfs/elev70/L70e095a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e095a.dat" right
[ 32, 54 ] = ascii (fp) : "./hrtfs/elev70/L70e090a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e090a.dat" right
[ 32, 55 ] = ascii (fp) : "./hrtfs/elev70/L70e085a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e085a.dat" right
[ 32, 56 ] = ascii (fp) : "./hrtfs/elev70/L70e080a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e080a.dat" right
[ 32, 57 ] = ascii (fp) : "./hrtfs/elev70/L70e075a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e075a.dat" right
[ 32, 58 ] = ascii (fp) : "./hrtfs/elev70/L70e070a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e070a.dat" right
[ 32, 59 ] = ascii (fp) : "./hrtfs/elev70/L70e065a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e065a.dat" right
[ 32, 60 ] = ascii (fp) : "./hrtfs/elev70/L70e060a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e060a.dat" right
[ 32, 61 ] = ascii (fp) : "./hrtfs/elev70/L70e055a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e055a.dat" right
[ 32, 62 ] = ascii (fp) : "./hrtfs/elev70/L70e050a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e050a.dat" right
[ 32, 63 ] = ascii (fp) : "./hrtfs/elev70/L70e045a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e045a.dat" right
[ 32, 64 ] = ascii (fp) : "./hrtfs/elev70/L70e040a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e040a.dat" right
[ 32, 65 ] = ascii (fp) : "./hrtfs/elev70/L70e035a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e035a.dat" right
[ 32, 66 ] = ascii (fp) : "./hrtfs/elev70/L70e030a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e030a.dat" right
[ 32, 67 ] = ascii (fp) : "./hrtfs/elev70/L70e025a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e025a.dat" right
[ 32, 68 ] = ascii (fp) : "./hrtfs/elev70/L70e020a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e020a.dat" right
[ 32, 69 ] = ascii (fp) : "./hrtfs/elev70/L70e015a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e015a.dat" right
[ 32, 70 ] = ascii (fp) : "./hrtfs/elev70/L70e010a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e010a.dat" right
[ 32, 71 ] = ascii (fp) : "./hrtfs/elev70/L70e005a.dat" left
           + ascii (fp) : "./hrtfs/elev70/R70e005a.dat" right

[ 33,  0 ] = ascii (fp) : "./hrtfs/elev75/L75e000a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e000a.dat" right
[ 33,  1 ] = ascii (fp) : "./hrtfs/elev75/L75e355a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e355a.dat" right
[ 33,  2 ] = ascii (fp) : "./hrtfs/elev75/L75e350a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e350a.dat" right
[ 33,  3 ] = ascii (fp) : "./hrtfs/elev75/L75e345a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e345a.dat" right
[ 33,  4 ] = ascii (fp) : "./hrtfs/elev75/L75e340a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e340a.dat" right
[ 33,  5 ] = ascii (fp) : "./hrtfs/elev75/L75e335a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e335a.dat" right
[ 33,  6 ] = ascii (fp) : "./hrtfs/elev75/L75e330a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e330a.dat" right
[ 33,  7 ] = ascii (fp) : "./hrtfs/elev75/L75e325a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e325a.dat" right
[ 33,  8 ] = ascii (fp) : "./hrtfs/elev75/L75e320a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e320a.dat" right
[ 33,  9 ] = ascii (fp) : "./hrtfs/elev75/L75e315a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e315a.dat" right
[ 33, 10 ] = ascii (fp) : "./hrtfs/elev75/L75e310a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e310a.dat" right
[ 33, 11 ] = ascii (fp) : "./hrtfs/elev75/L75e305a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e305a.dat" right
[ 33, 12 ] = ascii (fp) : "./hrtfs/elev75/L75e300a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e300a.dat" right
[ 33, 13 ] = ascii (fp) : "./hrtfs/elev75/L75e295a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e295a.dat" right
[ 33, 14 ] = ascii (fp) : "./hrtfs/elev75/L75e290a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e290a.dat" right
[ 33, 15 ] = ascii (fp) : "./hrtfs/elev75/L75e285a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e285a.dat" right
[ 33, 16 ] = ascii (fp) : "./hrtfs/elev75/L75e280a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e280a.dat" right
[ 33, 17 ] = ascii (fp) : "./hrtfs/elev75/L75e275a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e275a.dat" right
[ 33, 18 ] = ascii (fp) : "./hrtfs/elev75/L75e270a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e270a.dat" right
[ 33, 19 ] = ascii (fp) : "./hrtfs/elev75/L75e265a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e265a.dat" right
[ 33, 20 ] = ascii (fp) : "./hrtfs/elev75/L75e260a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e260a.dat" right
[ 33, 21 ] = ascii (fp) : "./hrtfs/elev75/L75e255a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e255a.dat" right
[ 33, 22 ] = ascii (fp) : "./hrtfs/elev75/L75e250a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e250a.dat" right
[ 33, 23 ] = ascii (fp) : "./hrtfs/elev75/L75e245a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e245a.dat" right
[ 33, 24 ] = ascii (fp) : "./hrtfs/elev75/L75e240a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e240a.dat" right
[ 33, 25 ] = ascii (fp) : "./hrtfs/elev75/L75e235a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e235a.dat" right
[ 33, 26 ] = ascii (fp) : "./hrtfs/elev75/L75e230a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e230a.dat" right
[ 33, 27 ] = ascii (fp) : "./hrtfs/elev75/L75e225a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e225a.dat" right
[ 33, 28 ] = ascii (fp) : "./hrtfs/elev75/L75e220a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e220a.dat" right
[ 33, 29 ] = ascii (fp) : "./hrtfs/elev75/L75e215a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e215a.dat" right
[ 33, 30 ] = ascii (fp) : "./hrtfs/elev75/L75e210a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e210a.dat" right
[ 33, 31 ] = ascii (fp) : "./hrtfs/elev75/L75e205a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e205a.dat" right
[ 33, 32 ] = ascii (fp) : "./hrtfs/elev75/L75e200a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e200a.dat" right
[ 33, 33 ] = ascii (fp) : "./hrtfs/elev75/L75e195a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e195a.dat" right
[ 33, 34 ] = ascii (fp) : "./hrtfs/elev75/L75e190a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e190a.dat" right
[ 33, 35 ] = ascii (fp) : "./hrtfs/elev75/L75e185a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e185a.dat" right
[ 33, 36 ] = ascii (fp) : "./hrtfs/elev75/L75e180a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e180a.dat" right
[ 33, 37 ] = ascii (fp) : "./hrtfs/elev75/L75e175a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e175a.dat" right
[ 33, 38 ] = ascii (fp) : "./hrtfs/elev75/L75e170a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e170a.dat" right
[ 33, 39 ] = ascii (fp) : "./hrtfs/elev75/L75e165a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e165a.dat" right
[ 33, 40 ] = ascii (fp) : "./hrtfs/elev75/L75e160a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e160a.dat" right
[ 33, 41 ] = ascii (fp) : "./hrtfs/elev75/L75e155a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e155a.dat" right
[ 33, 42 ] = ascii (fp) : "./hrtfs/elev75/L75e150a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e150a.dat" right
[ 33, 43 ] = ascii (fp) : "./hrtfs/elev75/L75e145a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e145a.dat" right
[ 33, 44 ] = ascii (fp) : "./hrtfs/elev75/L75e140a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e140a.dat" right
[ 33, 45 ] = ascii (fp) : "./hrtfs/elev75/L75e135a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e135a.dat" right
[ 33, 46 ] = ascii (fp) : "./hrtfs/elev75/L75e130a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e130a.dat" right
[ 33, 47 ] = ascii (fp) : "./hrtfs/elev75/L75e125a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e125a.dat" right
[ 33, 48 ] = ascii (fp) : "./hrtfs/elev75/L75e120a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e120a.dat" right
[ 33, 49 ] = ascii (fp) : "./hrtfs/elev75/L75e115a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e115a.dat" right
[ 33, 50 ] = ascii (fp) : "./hrtfs/elev75/L75e110a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e110a.dat" right
[ 33, 51 ] = ascii (fp) : "./hrtfs/elev75/L75e105a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e105a.dat" right
[ 33, 52 ] = ascii (fp) : "./hrtfs/elev75/L75e100a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e100a.dat" right
[ 33, 53 ] = ascii (fp) : "./hrtfs/elev75/L75e095a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e095a.dat" right
[ 33, 54 ] = ascii (fp) : "./hrtfs/elev75/L75e090a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e090a.dat" right
[ 33, 55 ] = ascii (fp) : "./hrtfs/elev75/L75e085a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e085a.dat" right
[ 33, 56 ] = ascii (fp) : "./hrtfs/elev75/L75e080a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e080a.dat" right
[ 33, 57 ] = ascii (fp) : "./hrtfs/elev75/L75e075a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e075a.dat" right
[ 33, 58 ] = ascii (fp) : "./hrtfs/elev75/L75e070a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e070a.dat" right
[ 33, 59 ] = ascii (fp) : "./hrtfs/elev75/L75e065a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e065a.dat" right
[ 33, 60 ] = ascii (fp) : "./hrtfs/elev75/L75e060a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e060a.dat" right
[ 33, 61 ] = ascii (fp) : "./hrtfs/elev75/L75e055a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e055a.dat" right
[ 33, 62 ] = ascii (fp) : "./hrtfs/elev75/L75e050a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e050a.dat" right
[ 33, 63 ] = ascii (fp) : "./hrtfs/elev75/L75e045a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e045a.dat" right
[ 33, 64 ] = ascii (fp) : "./hrtfs/elev75/L75e040a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e040a.dat" right
[ 33, 65 ] = ascii (fp) : "./hrtfs/elev75/L75e035a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e035a.dat" right
[ 33, 66 ] = ascii (fp) : "./hrtfs/elev75/L75e030a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e030a.dat" right
[ 33, 67 ] = ascii (fp) : "./hrtfs/elev75/L75e025a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e025a.dat" right
[ 33, 68 ] = ascii (fp) : "./hrtfs/elev75/L75e020a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e020a.dat" right
[ 33, 69 ] = ascii (fp) : "./hrtfs/elev75/L75e015a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e015a.dat" right
[ 33, 70 ] = ascii (fp) : "./hrtfs/elev75/L75e010a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e010a.dat" right
[ 33, 71 ] = ascii (fp) : "./hrtfs/elev75/L75e005a.dat" left
           + ascii (fp) : "./hrtfs/elev75/R75e005a.dat" right

[ 34,  0 ] = ascii (fp) : "./hrtfs/elev80/L80e000a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e000a.dat" right
[ 34,  1 ] = ascii (fp) : "./hrtfs/elev80/L80e355a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e355a.dat" right
[ 34,  2 ] = ascii (fp) : "./hrtfs/elev80/L80e350a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e350a.dat" right
[ 34,  3 ] = ascii (fp) : "./hrtfs/elev80/L80e345a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e345a.dat" right
[ 34,  4 ] = ascii (fp) : "./hrtfs/elev80/L80e340a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e340a.dat" right
[ 34,  5 ] = ascii (fp) : "./hrtfs/elev80/L80e335a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e335a.dat" right
[ 34,  6 ] = ascii (fp) : "./hrtfs/elev80/L80e330a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e330a.dat" right
[ 34,  7 ] = ascii (fp) : "./hrtfs/elev80/L80e325a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e325a.dat" right
[ 34,  8 ] = ascii (fp) : "./hrtfs/elev80/L80e320a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e320a.dat" right
[ 34,  9 ] = ascii (fp) : "./hrtfs/elev80/L80e315a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e315a.dat" right
[ 34, 10 ] = ascii (fp) : "./hrtfs/elev80/L80e310a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e310a.dat" right
[ 34, 11 ] = ascii (fp) : "./hrtfs/elev80/L80e305a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e305a.dat" right
[ 34, 12 ] = ascii (fp) : "./hrtfs/elev80/L80e300a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e300a.dat" right
[ 34, 13 ] = ascii (fp) : "./hrtfs/elev80/L80e295a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e295a.dat" right
[ 34, 14 ] = ascii (fp) : "./hrtfs/elev80/L80e290a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e290a.dat" right
[ 34, 15 ] = ascii (fp) : "./hrtfs/elev80/L80e285a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e285a.dat" right
[ 34, 16 ] = ascii (fp) : "./hrtfs/elev80/L80e280a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e280a.dat" right
[ 34, 17 ] = ascii (fp) : "./hrtfs/elev80/L80e275a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e275a.dat" right
[ 34, 18 ] = ascii (fp) : "./hrtfs/elev80/L80e270a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e270a.dat" right
[ 34, 19 ] = ascii (fp) : "./hrtfs/elev80/L80e265a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e265a.dat" right
[ 34, 20 ] = ascii (fp) : "./hrtfs/elev80/L80e260a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e260a.dat" right
[ 34, 21 ] = ascii (fp) : "./hrtfs/elev80/L80e255a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e255a.dat" right
[ 34, 22 ] = ascii (fp) : "./hrtfs/elev80/L80e250a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e250a.dat" right
[ 34, 23 ] = ascii (fp) : "./hrtfs/elev80/L80e245a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e245a.dat" right
[ 34, 24 ] = ascii (fp) : "./hrtfs/elev80/L80e240a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e240a.dat" right
[ 34, 25 ] = ascii (fp) : "./hrtfs/elev80/L80e235a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e235a.dat" right
[ 34, 26 ] = ascii (fp) : "./hrtfs/elev80/L80e230a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e230a.dat" right
[ 34, 27 ] = ascii (fp) : "./hrtfs/elev80/L80e225a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e225a.dat" right
[ 34, 28 ] = ascii (fp) : "./hrtfs/elev80/L80e220a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e220a.dat" right
[ 34, 29 ] = ascii (fp) : "./hrtfs/elev80/L80e215a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e215a.dat" right
[ 34, 30 ] = ascii (fp) : "./hrtfs/elev80/L80e210a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e210a.dat" right
[ 34, 31 ] = ascii (fp) : "./hrtfs/elev80/L80e205a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e205a.dat" right
[ 34, 32 ] = ascii (fp) : "./hrtfs/elev80/L80e200a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e200a.dat" right
[ 34, 33 ] = ascii (fp) : "./hrtfs/elev80/L80e195a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e195a.dat" right
[ 34, 34 ] = ascii (fp) : "./hrtfs/elev80/L80e190a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e190a.dat" right
[ 34, 35 ] = ascii (fp) : "./hrtfs/elev80/L80e185a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e185a.dat" right
[ 34, 36 ] = ascii (fp) : "./hrtfs/elev80/L80e180a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e180a.dat" right
[ 34, 37 ] = ascii (fp) : "./hrtfs/elev80/L80e175a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e175a.dat" right
[ 34, 38 ] = ascii (fp) : "./hrtfs/elev80/L80e170a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e170a.dat" right
[ 34, 39 ] = ascii (fp) : "./hrtfs/elev80/L80e165a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e165a.dat" right
[ 34, 40 ] = ascii (fp) : "./hrtfs/elev80/L80e160a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e160a.dat" right
[ 34, 41 ] = ascii (fp) : "./hrtfs/elev80/L80e155a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e155a.dat" right
[ 34, 42 ] = ascii (fp) : "./hrtfs/elev80/L80e150a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e150a.dat" right
[ 34, 43 ] = ascii (fp) : "./hrtfs/elev80/L80e145a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e145a.dat" right
[ 34, 44 ] = ascii (fp) : "./hrtfs/elev80/L80e140a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e140a.dat" right
[ 34, 45 ] = ascii (fp) : "./hrtfs/elev80/L80e135a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e135a.dat" right
[ 34, 46 ] = ascii (fp) : "./hrtfs/elev80/L80e130a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e130a.dat" right
[ 34, 47 ] = ascii (fp) : "./hrtfs/elev80/L80e125a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e125a.dat" right
[ 34, 48 ] = ascii (fp) : "./hrtfs/elev80/L80e120a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e120a.dat" right
[ 34, 49 ] = ascii (fp) : "./hrtfs/elev80/L80e115a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e115a.dat" right
[ 34, 50 ] = ascii (fp) : "./hrtfs/elev80/L80e110a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e110a.dat" right
[ 34, 51 ] = ascii (fp) : "./hrtfs/elev80/L80e105a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e105a.dat" right
[ 34, 52 ] = ascii (fp) : "./hrtfs/elev80/L80e100a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e100a.dat" right
[ 34, 53 ] = ascii (fp) : "./hrtfs/elev80/L80e095a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e095a.dat" right
[ 34, 54 ] = ascii (fp) : "./hrtfs/elev80/L80e090a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e090a.dat" right
[ 34, 55 ] = ascii (fp) : "./hrtfs/elev80/L80e085a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e085a.dat" right
[ 34, 56 ] = ascii (fp) : "./hrtfs/elev80/L80e080a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e080a.dat" right
[ 34, 57 ] = ascii (fp) : "./hrtfs/elev80/L80e075a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e075a.dat" right
[ 34, 58 ] = ascii (fp) : "./hrtfs/elev80/L80e070a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e070a.dat" right
[ 34, 59 ] = ascii (fp) : "./hrtfs/elev80/L80e065a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e065a.dat" right
[ 34, 60 ] = ascii (fp) : "./hrtfs/elev80/L80e060a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e060a.dat" right
[ 34, 61 ] = ascii (fp) : "./hrtfs/elev80/L80e055a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e055a.dat" right
[ 34, 62 ] = ascii (fp) : "./hrtfs/elev80/L80e050a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e050a.dat" right
[ 34, 63 ] = ascii (fp) : "./hrtfs/elev80/L80e045a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e045a.dat" right
[ 34, 64 ] = ascii (fp) : "./hrtfs/elev80/L80e040a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e040a.dat" right
[ 34, 65 ] = ascii (fp) : "./hrtfs/elev80/L80e035a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e035a.dat" right
[ 34, 66 ] = ascii (fp) : "./hrtfs/elev80/L80e030a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e030a.dat" right
[ 34, 67 ] = ascii (fp) : "./hrtfs/elev80/L80e025a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e025a.dat" right
[ 34, 68 ] = ascii (fp) : "./hrtfs/elev80/L80e020a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e020a.dat" right
[ 34, 69 ] = ascii (fp) : "./hrtfs/elev80/L80e015a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e015a.dat" right
[ 34, 70 ] = ascii (fp) : "./hrtfs/elev80/L80e010a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e010a.dat" right
[ 34, 71 ] = ascii (fp) : "./hrtfs/elev80/L80e005a.dat" left
           + ascii (fp) : "./hrtfs/elev80/R80e005a.dat" right

[ 35,  0 ] = ascii (fp) : "./hrtfs/elev85/L85e000a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e000a.dat" right
[ 35,  1 ] = ascii (fp) : "./hrtfs/elev85/L85e355a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e355a.dat" right
[ 35,  2 ] = ascii (fp) : "./hrtfs/elev85/L85e350a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e350a.dat" right
[ 35,  3 ] = ascii (fp) : "./hrtfs/elev85/L85e345a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e345a.dat" right
[ 35,  4 ] = ascii (fp) : "./hrtfs/elev85/L85e340a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e340a.dat" right
[ 35,  5 ] = ascii (fp) : "./hrtfs/elev85/L85e335a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e335a.dat" right
[ 35,  6 ] = ascii (fp) : "./hrtfs/elev85/L85e330a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e330a.dat" right
[ 35,  7 ] = ascii (fp) : "./hrtfs/elev85/L85e325a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e325a.dat" right
[ 35,  8 ] = ascii (fp) : "./hrtfs/elev85/L85e320a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e320a.dat" right
[ 35,  9 ] = ascii (fp) : "./hrtfs/elev85/L85e315a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e315a.dat" right
[ 35, 10 ] = ascii (fp) : "./hrtfs/elev85/L85e310a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e310a.dat" right
[ 35, 11 ] = ascii (fp) : "./hrtfs/elev85/L85e305a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e305a.dat" right
[ 35, 12 ] = ascii (fp) : "./hrtfs/elev85/L85e300a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e300a.dat" right
[ 35, 13 ] = ascii (fp) : "./hrtfs/elev85/L85e295a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e295a.dat" right
[ 35, 14 ] = ascii (fp) : "./hrtfs/elev85/L85e290a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e290a.dat" right
[ 35, 15 ] = ascii (fp) : "./hrtfs/elev85/L85e285a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e285a.dat" right
[ 35, 16 ] = ascii (fp) : "./hrtfs/elev85/L85e280a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e280a.dat" right
[ 35, 17 ] = ascii (fp) : "./hrtfs/elev85/L85e275a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e275a.dat" right
[ 35, 18 ] = ascii (fp) : "./hrtfs/elev85/L85e270a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e270a.dat" right
[ 35, 19 ] = ascii (fp) : "./hrtfs/elev85/L85e265a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e265a.dat" right
[ 35, 20 ] = ascii (fp) : "./hrtfs/elev85/L85e260a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e260a.dat" right
[ 35, 21 ] = ascii (fp) : "./hrtfs/elev85/L85e255a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e255a.dat" right
[ 35, 22 ] = ascii (fp) : "./hrtfs/elev85/L85e250a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e250a.dat" right
[ 35, 23 ] = ascii (fp) : "./hrtfs/elev85/L85e245a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e245a.dat" right
[ 35, 24 ] = ascii (fp) : "./hrtfs/elev85/L85e240a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e240a.dat" right
[ 35, 25 ] = ascii (fp) : "./hrtfs/elev85/L85e235a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e235a.dat" right
[ 35, 26 ] = ascii (fp) : "./hrtfs/elev85/L85e230a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e230a.dat" right
[ 35, 27 ] = ascii (fp) : "./hrtfs/elev85/L85e225a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e225a.dat" right
[ 35, 28 ] = ascii (fp) : "./hrtfs/elev85/L85e220a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e220a.dat" right
[ 35, 29 ] = ascii (fp) : "./hrtfs/elev85/L85e215a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e215a.dat" right
[ 35, 30 ] = ascii (fp) : "./hrtfs/elev85/L85e210a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e210a.dat" right
[ 35, 31 ] = ascii (fp) : "./hrtfs/elev85/L85e205a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e205a.dat" right
[ 35, 32 ] = ascii (fp) : "./hrtfs/elev85/L85e200a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e200a.dat" right
[ 35, 33 ] = ascii (fp) : "./hrtfs/elev85/L85e195a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e195a.dat" right
[ 35, 34 ] = ascii (fp) : "./hrtfs/elev85/L85e190a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e190a.dat" right
[ 35, 35 ] = ascii (fp) : "./hrtfs/elev85/L85e185a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e185a.dat" right
[ 35, 36 ] = ascii (fp) : "./hrtfs/elev85/L85e180a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e180a.dat" right
[ 35, 37 ] = ascii (fp) : "./hrtfs/elev85/L85e175a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e175a.dat" right
[ 35, 38 ] = ascii (fp) : "./hrtfs/elev85/L85e170a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e170a.dat" right
[ 35, 39 ] = ascii (fp) : "./hrtfs/elev85/L85e165a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e165a.dat" right
[ 35, 40 ] = ascii (fp) : "./hrtfs/elev85/L85e160a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e160a.dat" right
[ 35, 41 ] = ascii (fp) : "./hrtfs/elev85/L85e155a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e155a.dat" right
[ 35, 42 ] = ascii (fp) : "./hrtfs/elev85/L85e150a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e150a.dat" right
[ 35, 43 ] = ascii (fp) : "./hrtfs/elev85/L85e145a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e145a.dat" right
[ 35, 44 ] = ascii (fp) : "./hrtfs/elev85/L85e140a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e140a.dat" right
[ 35, 45 ] = ascii (fp) : "./hrtfs/elev85/L85e135a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e135a.dat" right
[ 35, 46 ] = ascii (fp) : "./hrtfs/elev85/L85e130a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e130a.dat" right
[ 35, 47 ] = ascii (fp) : "./hrtfs/elev85/L85e125a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e125a.dat" right
[ 35, 48 ] = ascii (fp) : "./hrtfs/elev85/L85e120a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e120a.dat" right
[ 35, 49 ] = ascii (fp) : "./hrtfs/elev85/L85e115a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e115a.dat" right
[ 35, 50 ] = ascii (fp) : "./hrtfs/elev85/L85e110a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e110a.dat" right
[ 35, 51 ] = ascii (fp) : "./hrtfs/elev85/L85e105a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e105a.dat" right
[ 35, 52 ] = ascii (fp) : "./hrtfs/elev85/L85e100a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e100a.dat" right
[ 35, 53 ] = ascii (fp) : "./hrtfs/elev85/L85e095a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e095a.dat" right
[ 35, 54 ] = ascii (fp) : "./hrtfs/elev85/L85e090a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e090a.dat" right
[ 35, 55 ] = ascii (fp) : "./hrtfs/elev85/L85e085a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e085a.dat" right
[ 35, 56 ] = ascii (fp) : "./hrtfs/elev85/L85e080a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e080a.dat" right
[ 35, 57 ] = ascii (fp) : "./hrtfs/elev85/L85e075a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e075a.dat" right
[ 35, 58 ] = ascii (fp) : "./hrtfs/elev85/L85e070a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e070a.dat" right
[ 35, 59 ] = ascii (fp) : "./hrtfs/elev85/L85e065a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e065a.dat" right
[ 35, 60 ] = ascii (fp) : "./hrtfs/elev85/L85e060a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e060a.dat" right
[ 35, 61 ] = ascii (fp) : "./hrtfs/elev85/L85e055a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e055a.dat" right
[ 35, 62 ] = ascii (fp) : "./hrtfs/elev85/L85e050a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e050a.dat" right
[ 35, 63 ] = ascii (fp) : "./hrtfs/elev85/L85e045a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e045a.dat" right
[ 35, 64 ] = ascii (fp) : "./hrtfs/elev85/L85e040a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e040a.dat" right
[ 35, 65 ] = ascii (fp) : "./hrtfs/elev85/L85e035a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e035a.dat" right
[ 35, 66 ] = ascii (fp) : "./hrtfs/elev85/L85e030a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e030a.dat" right
[ 35, 67 ] = ascii (fp) : "./hrtfs/elev85/L85e025a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e025a.dat" right
[ 35, 68 ] = ascii (fp) : "./hrtfs/elev85/L85e020a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e020a.dat" right
[ 35, 69 ] = ascii (fp) : "./hrtfs/elev85/L85e015a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e015a.dat" right
[ 35, 70 ] = ascii (fp) : "./hrtfs/elev85/L85e010a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e010a.dat" right
[ 35, 71 ] = ascii (fp) : "./hrtfs/elev85/L85e005a.dat" left
           + ascii (fp) : "./hrtfs/elev85/R85e005a.dat" right

[ 36,  0 ] = ascii (fp) : "./hrtfs/elev90/L90e000a.dat" left
           + ascii (fp) : "./hrtfs/elev90/R90e000a.dat" right


